//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Application\CAppBase.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CAppBase.h"

namespace App
{
	CAppBase::CAppBase()
	{
		m_pTimer							= NULL;

		m_pWindow							= NULL;
		m_pWinData							= NULL;
		m_pDirectX							= NULL;
		m_pDXStates							= NULL;
		m_pDX11Data							= NULL;
		m_pInput							= NULL;

		m_pPostQuad							= NULL;

		m_pAudioEngine						= NULL;

		m_bInFocus							= true;
		m_bMinimized						= false;
		m_bQuitting							= false;

		m_bInitialized						= false;
	}

	CAppBase::CAppBase(const CAppBase& other)
	{
	}

	CAppBase::~CAppBase()
	{
	}

	int CAppBase::Initialize()
	{
		int retCode = Util::Err::NONE;

		// Allocate memory.
		m_pTimer							= new Util::CTimer();

		m_pWindow							= new CWinInit();
		m_pDirectX							= new CDirectX11();
		m_pDXStates							= new CDXStates();
		m_pInput							= new CInput();

		m_pPostQuad							= new GFX::CPostScreenQuad();

		m_pAudioEngine						= new Audio::CXAudio2Engine();
		
		m_pWinData = m_pWindow->WindowData();

		if((retCode = Setup()) != Util::Err::NONE)
			return retCode;
			
		// Window initialization.
		if((retCode = m_pWindow->Create(DisplayWinProc, (long)this)) != Util::Err::NONE)
			return retCode;

		// DirectX 11 initialization.
		if((retCode = m_pDirectX->Initialize(m_pWindow->WindowData())) != Util::Err::NONE)
			return retCode;

		m_pDX11Data = m_pDirectX->DXData();

		// DirectX 11 states initialization.
		m_pDXStates->DirectXData(m_pDX11Data);
		if((retCode = m_pDXStates->Initialize()) != Util::Err::NONE)
			return retCode;

		// Input initialization.
		m_pInput->hWnd(m_pWinData->hWnd);

		if((retCode = m_pInput->Initialize()) != Util::Err::NONE)
			return retCode;
			
		CControls::Instance().Initialize(m_pInput, m_pTimer);
		
		// Post screen quad initialization.
		m_pPostQuad->RenderData(&m_renderData);
		m_pPostQuad->DirectXData(m_pDX11Data);
		m_pPostQuad->SetScreenSize(&m_pWinData->width, &m_pWinData->height);

		// Setup texture list.
		GFX::CTextureList::Instance().DirectXData(m_pDX11Data);

		if((retCode = AdjustToResize()) != Util::Err::NONE)
			return retCode;

		// Audio engine.
		if((retCode = m_pAudioEngine->Initialize()) != Util::Err::NONE)
			return retCode;

		Audio::CAudioList::Instance().SetEngine(m_pAudioEngine);

		//
		// Setup application singletons the need to wait for application initialization.
		//

		// Setup the screenshot singleton.
		Util::CScreenshot::Instance().DirectXData(m_pDX11Data);

		// Setup the system info singleton.
		Util::CSysInfo::Instance().CalculateMemoryInfo();
		Util::CSysInfo::Instance().InitCPUCalculations();
			
		return retCode;
	}

	// Post quad initialization is in a seperate method for if the user wants to make any changes.
	int CAppBase::InitializePostQuad()
	{
		int retCode = Util::Err::NONE;

		if((retCode = m_pPostQuad->Initialize()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	int CAppBase::Run()
	{
		int retCode = Util::Err::NONE;

		MSG msg;

		// Start of the game loop.
		while(true)
		{
			if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if(msg.message == WM_QUIT)
				{
					retCode = msg.wParam;
					m_bQuitting = true;
					break;
				}

				if((retCode = MessageHandler(msg)) != Util::Err::NONE)
					PostQuitMessage(retCode);

				if((retCode = m_pInput->MessageHandler(msg)) != Util::Err::NONE)
					PostQuitMessage(retCode);

				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			else
			{
				m_bInFocus = !m_bMinimized;
				
				if(m_bInFocus)
				{
					if((retCode = m_pInput->Update()) != Util::Err::NONE)
						return retCode;

					CControls::Instance().Update();

					if((retCode = Update()) != Util::Err::NONE)
						return retCode;
				}

				if(!m_bMinimized)
				{
					// Reset render data render counts.
					m_renderData.geoRenderCount = 0;
					m_renderData.polyRenderCount = 0;
					m_renderData.vertRenderCount = 0;

					// Only render if the game isn't minimized.
					if((retCode = Render()) != Util::Err::NONE)
						return retCode;
				}
				else // Otherwise sleep 5 miliseconds per frame.
					Sleep(5);

				if(m_bInFocus)
				{
					if((retCode = m_pInput->PostUpdate()) != Util::Err::NONE)
						return retCode;
				}
			}
		}

		return retCode;
	}

	// Method for calling pre-resize events specific to this object.
	int CAppBase::PreResize()
	{
		int retCode = Util::Err::NONE;
		
		if(m_bInitialized)
		{
			if((retCode = m_pPostQuad->PreResize()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}
	
	// Method for calling post-resize events specific to this object.
	int CAppBase::PostResize()
	{
		int retCode = Util::Err::NONE;
		
		if(m_bInitialized)
		{
			if((retCode = m_pPostQuad->PostResize()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}

	void CAppBase::Release()
	{
		//SAFE_RELEASE(m_pAudioEngine); --- Released in the audio list singleton.

		SAFE_RELEASE_DELETE(m_pPostQuad);

		SAFE_DELETE(m_pInput);
		SAFE_RELEASE_DELETE(m_pDXStates);
		SAFE_RELEASE_DELETE(m_pDirectX);
		SAFE_RELEASE_DELETE(m_pWindow);

		SAFE_DELETE(m_pTimer);
	}

	//
	// Window procedure methods.
	//

	LRESULT CALLBACK CAppBase::DisplayWinProc(HWND hWnd, u32 msg, WPARAM wParam, LPARAM lParam)
	{
		if(msg == WM_CREATE)
			SetWindowLongPtr(hWnd, GWL_USERDATA, (LONG_PTR)((CREATESTRUCT FAR*)lParam)->lpCreateParams);

		CAppBase *Destination = (CAppBase*)((LONG_PTR)GetWindowLongPtr(hWnd, GWL_USERDATA));

		if(Destination) return Destination->WinProc(hWnd, msg, wParam, lParam);

		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	LRESULT CALLBACK CAppBase::WinProc(HWND hWnd, u32 msg, WPARAM wParam, LPARAM lParam)
	{
		int retCode = Util::Err::NONE;

		switch(msg)
		{
			case WM_CREATE:
				break;

			case WM_CLOSE:
				PostQuitMessage(0);
				break;

			case WM_DESTROY:
				PostQuitMessage(0);
				break;

			case WM_SIZE:
				m_bMinimized = false;

				// Handle window size event.
				switch(wParam)
				{
					case SIZE_MINIMIZED:
						m_bMinimized = true;
						break;
					case SIZE_MAXIMIZED:
						m_pWindow->UpdateRects();

						if((retCode = AdjustToResize()) != Util::Err::NONE)
							PostQuitMessage(retCode);

						break;
					case SIZE_RESTORED:
						if((retCode = AdjustToResize()) != Util::Err::NONE)
							PostQuitMessage(retCode);

						break;
					default:
						
						// If a window is resize with a mouse drag, wait for the left button to be released before resize.
						if(m_pInput->MouseUp(0))
						{
							m_pWindow->UpdateRects();

							if((retCode = AdjustToResize()) != Util::Err::NONE)
								PostQuitMessage(retCode);

							m_pWindow->UpdateRects();
						}

						break;
				}

				break;

			default:
				break;
		}

		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	// Method for adjusting everything that is affected by a resized window.
	int CAppBase::AdjustToResize()
	{
		int retCode = Util::Err::NONE;

		// Make sure the app was initialize prior to resize.
		if(!m_bInitialized || m_bQuitting)
			return retCode;

		if((retCode = PreResize()) != Util::Err::NONE)
			return (retCode);
			
		if((retCode = m_pDirectX->ResizeWindow()) != Util::Err::NONE)
			return (retCode);

		if((retCode = m_pDirectX->FindScreenMode()) != Util::Err::NONE)
			PostQuitMessage(retCode);
			
		if((retCode = PostResize()) != Util::Err::NONE)
			return (retCode);	

		return retCode;
	}
};