//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Application\CDX11States.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CDXStates.h"

namespace App
{
	CDXStates::CDXStates()
	{
		m_pBlendNone						= NULL;
		m_pBlendDefault						= NULL;
		m_pBlendUI							= NULL;
		m_pBlendParticle					= NULL;

		m_pRasterizerStateDefault			= NULL;
		m_pRasterizerStateCullNone			= NULL;
		m_pRasterizerStateWireframe			= NULL;
		
		m_pSamplerStateAnisotropic			= NULL;
		m_pSamplerStateLinear				= NULL;
		m_pSamplerStatePoint				= NULL;
		m_pSamplerStateAnisotropicClamped	= NULL;
		m_pSamplerStateLinearClamped		= NULL;
		m_pSamplerStatePointClamped			= NULL;
	}

	CDXStates::CDXStates(const CDXStates& other)
	{
	}

	CDXStates::~CDXStates()
	{
	}

	// Initialize all states, and set the default states to active.
	int CDXStates::Initialize()
	{
		int retCode = Util::Err::NONE;

		// Initialize states.
		if((retCode = InitializeBlendStates()) != Util::Err::NONE)
			return retCode;

		if((retCode = InitializeRasterizerStates()) != Util::Err::NONE)
			return retCode;

		if((retCode = InitializeSamplerStates()) != Util::Err::NONE)
			return retCode;
			
		// Set and bind initial states.
		
		if((retCode = SetBlend(BlendState::Default)) != Util::Err::NONE)
			return retCode;
		
		if((retCode = SetRasterizer(RasterizerState::Default)) != Util::Err::NONE)
			return retCode;

		if((retCode = BindSamplers()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for initializing blend states.
	int CDXStates::InitializeBlendStates()
	{
		int retCode = Util::Err::NONE;

		D3D11_BLEND_DESC bd;
		ZeroMemory(&bd, sizeof(bd));

		bd.AlphaToCoverageEnable						= false;
		bd.IndependentBlendEnable						= true;
		
		// Set all render target blend states to default values.
		for(int i = 0; i < 8; i++)
		{
			bd.RenderTarget[i].BlendEnable					= false;
			bd.RenderTarget[i].BlendOp						= D3D11_BLEND_OP_ADD;
			bd.RenderTarget[i].SrcBlend						= D3D11_BLEND_SRC_ALPHA;
			bd.RenderTarget[i].DestBlend					= D3D11_BLEND_INV_SRC_ALPHA;
			bd.RenderTarget[i].BlendOpAlpha					= D3D11_BLEND_OP_ADD;
			bd.RenderTarget[i].SrcBlendAlpha				= D3D11_BLEND_SRC_ALPHA;
			bd.RenderTarget[i].DestBlendAlpha				= D3D11_BLEND_INV_SRC_ALPHA;
			bd.RenderTarget[i].RenderTargetWriteMask		= D3D11_COLOR_WRITE_ENABLE_ALL;
		}
		
		// No blending.
		if(FAILED(m_pDev->CreateBlendState(&bd, &m_pBlendNone)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_BLEND_STATE;
		}
		
		bd.RenderTarget[0].BlendEnable					= true;
		
		// Default alpha blending.
		if(FAILED(m_pDev->CreateBlendState(&bd, &m_pBlendDefault)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_BLEND_STATE;
		}
		
		bd.AlphaToCoverageEnable						= true;
		bd.RenderTarget[0].BlendEnable					= true;
		bd.IndependentBlendEnable						= true;
		
		// UI alpha blending.
		if(FAILED(m_pDev->CreateBlendState(&bd, &m_pBlendUI)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_BLEND_STATE;
		}
		
		bd.AlphaToCoverageEnable						= false;
		bd.RenderTarget[0].SrcBlend						= D3D11_BLEND_SRC_COLOR;
		bd.RenderTarget[0].DestBlend					= D3D11_BLEND_INV_SRC_COLOR;
		
		// Default additive blending.
		if(FAILED(m_pDev->CreateBlendState(&bd, &m_pBlendParticle)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_BLEND_STATE;
		}

		return retCode;
	}
	
	// Function for initializing rasterizer states.
	int CDXStates::InitializeRasterizerStates()
	{
		int retCode = Util::Err::NONE;

		D3D11_RASTERIZER_DESC rd;
		ZeroMemory(&rd, sizeof(rd));
		
		rd.AntialiasedLineEnable			= false;
		rd.MultisampleEnable				= false;
		rd.CullMode							= D3D11_CULL_BACK;
		rd.FillMode							= D3D11_FILL_SOLID;
		rd.FrontCounterClockwise			= false;
		rd.DepthClipEnable					= true;
		rd.ScissorEnable					= false;
		rd.DepthBias						= 0;
		rd.DepthBiasClamp					= 0.0f;
		rd.SlopeScaledDepthBias				= 0.0f;
		
		// Default rasterizer.
		if(FAILED(m_pDev->CreateRasterizerState(&rd, &m_pRasterizerStateDefault)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_RASTERIZER_STATE;
		}

		rd.CullMode							= D3D11_CULL_NONE;

		// Disabled culling rasterizer.
		if(FAILED(m_pDev->CreateRasterizerState(&rd, &m_pRasterizerStateCullNone)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_RASTERIZER_STATE;
		}
		
		rd.FillMode							= D3D11_FILL_WIREFRAME;

		// Wireframe rasterizer.
		if(FAILED(m_pDev->CreateRasterizerState(&rd, &m_pRasterizerStateWireframe)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_RASTERIZER_STATE;
		}

		return retCode;
	}
	
	// Function for initializing sampler states.
	int CDXStates::InitializeSamplerStates()
	{
		int retCode = Util::Err::NONE;

		// Setup the sampler state.
		D3D11_SAMPLER_DESC sd;
		ZeroMemory(&sd, sizeof(sd));

		sd.Filter							= D3D11_FILTER_ANISOTROPIC;
		sd.AddressU							= D3D11_TEXTURE_ADDRESS_WRAP;
		sd.AddressV							= D3D11_TEXTURE_ADDRESS_WRAP;
		sd.AddressW							= D3D11_TEXTURE_ADDRESS_WRAP;
		sd.MipLODBias						= 0.0f;
		sd.MaxAnisotropy					= 16;
		sd.ComparisonFunc					= D3D11_COMPARISON_ALWAYS;
		sd.BorderColor[0]					= 0.0f;
		sd.BorderColor[1]					= 0.0f;
		sd.BorderColor[2]					= 0.0f;
		sd.BorderColor[3]					= 0.0f;
		sd.MinLOD							= 0.0f;
		sd.MaxLOD							= D3D11_FLOAT32_MAX;
		
		// Wrapped anisotropic.
		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStateAnisotropic)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}
		
		// Wrapped linear.
		sd.Filter							= D3D11_FILTER_MIN_MAG_MIP_LINEAR;

		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStateLinear)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}

		// Wrapped point.
		sd.Filter							= D3D11_FILTER_MIN_MAG_MIP_POINT;

		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStatePoint)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}

		sd.AddressU							= D3D11_TEXTURE_ADDRESS_CLAMP;
		sd.AddressV							= D3D11_TEXTURE_ADDRESS_CLAMP;
		sd.AddressW							= D3D11_TEXTURE_ADDRESS_CLAMP;
		
		// Clamped anisotropic.
		sd.Filter							= D3D11_FILTER_ANISOTROPIC;

		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStateAnisotropicClamped)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}

		// Clamped linear.
		
		sd.Filter							= D3D11_FILTER_MIN_MAG_MIP_LINEAR;

		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStateLinearClamped)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}

		// Clamped point.
		sd.Filter							= D3D11_FILTER_MIN_MAG_MIP_POINT;

		if(FAILED(m_pDev->CreateSamplerState(&sd, &m_pSamplerStatePointClamped)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_SAMPLER_STATE;
		}

		return retCode;
	}

	// Sets a blend state to the directx context.
	int CDXStates::SetBlend(const BlendState& state)
	{
		int retCode = Util::Err::NONE;

		switch(state)
		{
			case App::BlendState::None:
				m_pDevCon->OMSetBlendState(m_pBlendNone, 0, 0xFFFFFF);
				break;
			case App::BlendState::Default:
				m_pDevCon->OMSetBlendState(m_pBlendDefault, 0, 0xFFFFFF);
				break;
			case App::BlendState::UI:
				m_pDevCon->OMSetBlendState(m_pBlendUI, 0, 0xFFFFFF);
				break;
			case App::BlendState::Particle:
				m_pDevCon->OMSetBlendState(m_pBlendParticle, 0, 0xFFFFFF);
				break;
			default:
				break;
		}

		return retCode;
	}
	
	// Sets a rasterizer state to the directx context.
	int CDXStates::SetRasterizer(const RasterizerState& state)
	{
		int retCode = Util::Err::NONE;
		switch(state)
		{
			case App::RasterizerState::Default:
				m_pDevCon->RSSetState(m_pRasterizerStateDefault);
				break;
			case App::RasterizerState::CullNone:
				m_pDevCon->RSSetState(m_pRasterizerStateCullNone);
				break;
			case App::RasterizerState::Wireframe:
				m_pDevCon->RSSetState(m_pRasterizerStateWireframe);
				break;
			default:
				break;
		}

		return retCode;
	}
	
	// Sets a sampler state to the directx context.
	int CDXStates::BindSamplers()
	{
		int retCode = Util::Err::NONE;
		
		ID3D11SamplerState* samplers[6];
		
		samplers[0] = m_pSamplerStateAnisotropic;
		samplers[1] = m_pSamplerStateAnisotropicClamped;
		samplers[2] = m_pSamplerStateLinear;
		samplers[3] = m_pSamplerStateLinearClamped;
		samplers[4] = m_pSamplerStatePoint;
		samplers[5] = m_pSamplerStatePointClamped;

		m_pDevCon->PSSetSamplers(0, 6, samplers);

		return retCode;
	}

	// Retrieves a blend state.
	ID3D11BlendState* CDXStates::GetBlend(const BlendState& state) const
	{
		switch(state)
		{
			case App::BlendState::None:
				return m_pBlendNone;
			case App::BlendState::Default:
				return m_pBlendDefault;
			case App::BlendState::UI:
				return m_pBlendUI;
			case App::BlendState::Particle:
				return m_pBlendParticle;
			default:
				return m_pBlendDefault;
		}
	}
	
	// Retrieves a rasterizer state.
	ID3D11RasterizerState* CDXStates::GetRasterizer(const RasterizerState& state) const
	{
		switch(state)
		{
			case App::RasterizerState::Default:
				return m_pRasterizerStateDefault;
			case App::RasterizerState::CullNone:
				return m_pRasterizerStateCullNone;
			case App::RasterizerState::Wireframe:
				return m_pRasterizerStateWireframe;
			default:
				return m_pRasterizerStateDefault;
		}
	}
	
	// Retrieves a sampler state.
	ID3D11SamplerState* CDXStates::GetSampler(const SamplerState& state) const
	{
		switch(state)
		{
			case App::SamplerState::Anisotropic:
				return m_pSamplerStateAnisotropic;
			case App::SamplerState::AnisotropicClamped:
				return m_pSamplerStateAnisotropicClamped;
			case App::SamplerState::Linear:
				return m_pSamplerStateLinear;
			case App::SamplerState::LinearClamped:
				return m_pSamplerStateLinearClamped;
			case App::SamplerState::Point:
				return m_pSamplerStatePoint;
			case App::SamplerState::PointClamped:
				return m_pSamplerStatePointClamped;
			default:
				return m_pSamplerStateAnisotropic;
		}
	}

	void CDXStates::Release()
	{
		SAFE_RELEASE(m_pSamplerStatePointClamped);
		SAFE_RELEASE(m_pSamplerStateLinearClamped);
		SAFE_RELEASE(m_pSamplerStateAnisotropicClamped);
		SAFE_RELEASE(m_pSamplerStatePoint);
		SAFE_RELEASE(m_pSamplerStateLinear);
		SAFE_RELEASE(m_pSamplerStateAnisotropic);

		SAFE_RELEASE(m_pRasterizerStateWireframe);
		SAFE_RELEASE(m_pRasterizerStateCullNone);
		SAFE_RELEASE(m_pRasterizerStateDefault);
		
		SAFE_RELEASE(m_pBlendParticle);
		SAFE_RELEASE(m_pBlendUI);
		SAFE_RELEASE(m_pBlendDefault);
		SAFE_RELEASE(m_pBlendNone);
	}
};