//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Application\CInput.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CInput.h"

namespace App
{
	CInput::CInput()
	{
		m_pRawInput							= NULL;
		m_hWnd								= NULL;

		memset(m_bKeyDown, false, sizeof(bool) * MAX_KEYS);
		memset(m_bKeyUp, false, sizeof(bool) * MAX_KEYS);
		memset(m_bKey, false, sizeof(bool) * MAX_KEYS);

		memset(m_bMouseDown, false, sizeof(bool) * MAX_MOUSE_BUTTONS);
		memset(m_bMouseUp, false, sizeof(bool) * MAX_MOUSE_BUTTONS);
		memset(m_bMouse, false, sizeof(bool) * MAX_MOUSE_BUTTONS);
		memset(m_bMouseDClick, false, sizeof(bool) * MAX_MOUSE_BUTTONS);
	}

	CInput::CInput(const CInput& other)
	{
	}

	CInput::~CInput()
	{
	}

	// Method for initialize raw input devices.
	int CInput::Initialize()
	{
		int retCode = Util::Err::NONE;

		RAWINPUTDEVICE Rid[2];

		// Mouse.
		Rid[0].usUsagePage					= 0x01;
		Rid[0].usUsage						= 0x02;
		Rid[0].dwFlags						= RIDEV_INPUTSINK;
		Rid[0].hwndTarget					= m_hWnd;

		// Keyboard.
		Rid[1].usUsagePage					= 0x01;
		Rid[1].usUsage						= 0x06;
		Rid[1].dwFlags						= RIDEV_INPUTSINK;
		Rid[1].hwndTarget					= m_hWnd;

		if(!RegisterRawInputDevices(Rid, 2, sizeof(Rid[0])))
		{
			return Util::Err::ERR_RAW_INPUT_DEVICE_REGISTRATION;
		}

		return retCode;
	}

	// Method for updating mouse/keyboard input. (Place at top of logic loop).
	int CInput::Update()
	{
		int retCode = Util::Err::NONE;

		POINT mousePt;
		GetCursorPos(&mousePt);
		ScreenToClient(m_hWnd, &mousePt);
		m_mouseData.pos.x = (float)mousePt.x;
		m_mouseData.pos.y = (float)mousePt.y;

		UpdateMouse(0, (GetKeyState(VK_LBUTTON) & 0x8000) ? true : false);
		UpdateMouse(1, (GetKeyState(VK_RBUTTON) & 0x8000) ? true : false);
		UpdateMouse(2, (GetKeyState(VK_MBUTTON) & 0x8000) ? true : false);

		return retCode;
	}
	
	// Method for mouse/keyboard input post-update. (Place at bottom of logic loop).
	int CInput::PostUpdate()
	{
		int retCode = Util::Err::NONE;

		m_pRawInput = NULL;
		UpdateMouseWheel(0);
		UpdateMouseData();

		memset(m_bKeyDown, false, sizeof(bool) * MAX_KEYS);
		memset(m_bKeyUp, false, sizeof(bool) * MAX_KEYS);

		memset(m_bMouseDown, false, sizeof(bool) * MAX_MOUSE_BUTTONS);
		memset(m_bMouseUp, false, sizeof(bool) * MAX_MOUSE_BUTTONS);

		memset(m_bMouseDClick, false, sizeof(bool) * MAX_MOUSE_BUTTONS);

		return retCode;
	}

	// Method for extracting input data from a window message structure.
	int CInput::MessageHandler(MSG msg)
	{
		int retCode = Util::Err::NONE;

		switch(msg.message)
		{
			case WM_LBUTTONDBLCLK:
				UpdateMouseDClick(0);
				break;
			case WM_RBUTTONDBLCLK:
				UpdateMouseDClick(1);
				break;
			case WM_MBUTTONDBLCLK:
				UpdateMouseDClick(2);
				break;
			case WM_INPUT:
			{
				// Update raw user input data.
				u32 size = 0;
				GetRawInputData((HRAWINPUT)msg.lParam, RID_INPUT, NULL, &size, sizeof(RAWINPUTHEADER));
				if(size == 0)
				{
					m_pRawInput = NULL;
					break;
				}

				u8* lpb = new u8[size];
				GetRawInputData((HRAWINPUT)msg.lParam, RID_INPUT, lpb, &size, sizeof(RAWINPUTHEADER));
				
				m_pRawInput = (RAWINPUT*)lpb;
				if(m_pRawInput->header.dwType == RIM_TYPEKEYBOARD)
				{
					UpdateKey((u32)m_pRawInput->data.keyboard.VKey, (m_pRawInput->data.keyboard.Flags & 1) == 1 ? false : true);
				}
				else if(m_pRawInput->header.dwType == RIM_TYPEMOUSE)
				{
					if(m_pRawInput->data.mouse.ulRawButtons & 0x0400)
					{
						UpdateMouseWheel(((float)m_pRawInput->data.mouse.usButtonData - 0x7fff) / 0x7fff);
					}

					UpdateMouseData();
				}

				delete [] lpb;
				break;
			}
			default:
				break;
		}

		return retCode;
	}

	// Method for updating mouse position data.
	void CInput::UpdateMouseData()
	{
		bool resetXY = true;

		if(m_pRawInput)
		{
			if(m_pRawInput->header.dwType == RIM_TYPEMOUSE)
			{
				if(m_pRawInput->data.mouse.usFlags == 0)
				{
					m_mouseData.x = (float)m_pRawInput->data.mouse.lLastX;
					m_mouseData.y = (float)m_pRawInput->data.mouse.lLastY;
					resetXY = false;
				}
			}
		}

		if(resetXY)
		{
			// Reset mouse pos(x, y) on null raw input.
			m_mouseData.x = m_mouseData.xPrev = 0;
			m_mouseData.y = m_mouseData.yPrev = 0;
		}
	}
};