//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Application\CInputTyping.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CInputTyping.h"

namespace App
{
	CInputTyping::CInputTyping()
	{
		m_text								= "";

		m_bCaps								= false;
		m_bNumLock							= false;

		m_typingFlag						= TYPING_FLAG::TYPING_EVERYTHING;

		m_pInput							= NULL;
		m_pTimer							= NULL;

		memset(m_keyTimers, 0, sizeof(float) * MAX_KEYS);
	}

	CInputTyping::CInputTyping(const CInputTyping& other)
	{
	}

	CInputTyping::~CInputTyping()
	{
	}

	int CInputTyping::Initialize(CInput*& input, Util::CTimer* timer)
	{
		int retCode = Util::Err::NONE;

		m_pInput = input;
		m_pTimer = timer;

		return retCode;
	}

	// Method for updating a string based on user input key strokes.
	bool CInputTyping::Update(int& caretIndex, std::string* outputText)
	{
		std::string text = "";
		int lastCaret = caretIndex;
		bool textChange = false;
		
		// Update the keyboard state.
		m_bCaps = ((GetKeyState(VK_CAPITAL) & 0x0001) != 0);
		m_bNumLock = ((GetKeyState(VK_NUMLOCK) & 0x0001) != 0);

		TestNumeric(text);
		TestAlpha(text);
		TestSpecial(text);

		std::string newStr = m_text;

		HandleBackAndDelete(newStr, caretIndex);

		// Update the member text, and the caret position.
		newStr.insert(caretIndex, text);

		caretIndex += (int)text.length();
		HandleArrowKeys(newStr, caretIndex);
		caretIndex = Math::Clamp(caretIndex, 0, newStr.length());

		if(m_text != newStr)
		{
			textChange = true;
			m_text = newStr;
		}

		if(outputText)
			*outputText = m_text;

		return lastCaret != caretIndex || textChange;
	}

	// Method for testing the numeric keys.
	void CInputTyping::TestNumeric(std::string& text)
	{
		int index;

		if(m_bNumLock)
		{
			if((m_typingFlag & TYPING_INTEGER) == TYPING_INTEGER)
			{
				for(int i = 0; i < 10; i++)
				{
					index = VK_NUMPAD0 + i;
					HandleKeyStroke(text, index, index - VK_KEY_0);
				}
			}

			// Numpad specials.
			if((m_typingFlag & TYPING_NUMERIC_SIGNS) == TYPING_NUMERIC_SIGNS)
				HandleKeyStroke(text, 0x6E, '.');
		}

		if(!m_pInput->Key(VK_SHIFT))
		{
			if((m_typingFlag & TYPING_INTEGER) == TYPING_INTEGER)
			{
				for(int i = 0; i < 10; i++)
				{
					index = VK_KEY_0 + i;
					HandleKeyStroke(text, index, index);
				}
			}
			
			// Numpad specials.
			if((m_typingFlag & TYPING_NUMERIC_SIGNS) == TYPING_NUMERIC_SIGNS)
			{
				HandleKeyStroke(text, 0x6B, '+');
				HandleKeyStroke(text, 0x6D, '-');
			}
			if((m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
			{
				HandleKeyStroke(text, 0x6A, '*');
				HandleKeyStroke(text, 0x6F, '/');
			}
		}
	}

	// Method for testing letters (upper and lower case)
	void CInputTyping::TestAlpha(std::string& text)
	{
		int index;

		for(int i = 0; i < 26; i++)
		{
			index = VK_KEY_A + i;
			if(m_bCaps || m_pInput->Key(VK_SHIFT))
			{
				if((m_typingFlag & TYPING_UPPER) == TYPING_UPPER)
					HandleKeyStroke(text, index, index);
			}
			else
			{
				if((m_typingFlag & TYPING_LOWER) == TYPING_LOWER)
					HandleKeyStroke(text, index, index + 0x20);
			}
		}
	}
	
	// Method for testing special character.
	void CInputTyping::TestSpecial(std::string& text)
	{
		int index;
		char character;

		if(!m_pInput->Key(VK_SHIFT))
		{
			// Checl for lower-case special characters.
			for(index = 0xBA; index <= 0xC0; index++)
			{
				if(((index == 0xBD || index == 0xBE) && (m_typingFlag & TYPING_NUMERIC_SIGNS) == TYPING_NUMERIC_SIGNS) || 
					(m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
				{
					character = GetSpecialFromKeyStroke(index);
					HandleKeyStroke(text, index, character);
				}
			}

			if((m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
			{
				for(index = 0xDB; index <= 0xDE; index++)
				{
					character = GetSpecialFromKeyStroke(index);
					HandleKeyStroke(text, index, character);
				}
			}
		}
		if(m_pInput->Key(VK_SHIFT))
		{
			// Checl for capital special characters.
			if((m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
			{
				for(int i = 0; i < 10; i++)
				{
					index = VK_KEY_0 + i;

					character = GetSpecialFromLower(index);
					HandleKeyStroke(text, index, character);
				}
			}

			for(index = 0xBA; index <= 0xC0; index++)
			{
				if((index == 0xBB && (m_typingFlag & TYPING_NUMERIC_SIGNS) == TYPING_NUMERIC_SIGNS) || 
					(m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
				{
					character = GetSpecialFromLower(index);
					HandleKeyStroke(text, index, character);
				}
			}

			if((m_typingFlag & TYPING_SPECIAL) == TYPING_SPECIAL)
			{
				for(index = 0xDB; index <= 0xDE; index++)
				{
					character = GetSpecialFromLower(index);
					HandleKeyStroke(text, index, character);
				}
			}
		}

		// Spacebar.
		if(m_typingFlag & ~TYPING_RATIONAL)
			HandleKeyStroke(text, 0x20, ' ');
	}

	// Method for returning the special character of a keystroke.
	char CInputTyping::GetSpecialFromKeyStroke(int key)
	{
		switch(key)
		{
			case 0xBA:
				return ';';
			case 0xBB:
				return '=';
			case 0xBC:
				return ',';
			case 0xBD:
				return '-';
			case 0xBE:
				return '.';
			case 0xBF:
				return '/';
			case 0xC0:
				return '`';
			default:
				break;
		}

		switch(key)
		{
			case 0xDB:
				return '[';
			case 0xDC:
				return '\\';
			case 0xDD:
				return ']';
			case 0xDE:
				return '\'';
			default:
				break;
		}

		return '\0';
	}

	// Method for returning the special character of a lowercase version of it.
	char CInputTyping::GetSpecialFromLower(int lower)
	{
		// Test for numeric special cases.
		switch(lower)
		{
			case '0':
				return ')';
			case '1':
				return '!';
			case '2':
				return '@';
			case '3':
				return '#';
			case '4':
				return '$';
			case '5':
				return '%';
			case '6':
				return '^';
			case '7':
				return '&';
			case '8':
				return '*';
			case '9':
				return ')';
			default:
				break;
		}
		
		// Test for special cases.
		switch(lower)
		{
			case 0xBA:
				return ':';
			case 0xBB:
				return '+';
			case 0xBC:
				return '<';
			case 0xBD:
				return '_';
			case 0xBE:
				return '>';
			case 0xBF:
				return '?';
			case 0xC0:
				return '~';
			default:
				break;
		}

		switch(lower)
		{
			case 0xDB:
				return '{';
			case 0xDC:
				return '|';
			case 0xDD:
				return '}';
			case 0xDE:
				return '\"';
			default:
				break;
		}

		return '\0';
	}

	// Method for handling when a key is pressed.
	void CInputTyping::HandleKeyStroke(std::string& text, int key, char character)
	{
		if(m_pInput->KeyDown(key))
		{
			m_keyTimers[key] = 0.0f;
			text += character;
		}
		else if(m_pInput->Key(key))
		{
			m_keyTimers[key] += m_pTimer->Delta();

			if(m_keyTimers[key] >= TYPING_KEY_HELD_TIME)
			{
				text += character;
				m_keyTimers[key] -= TYPING_KEY_SENSITIVITY;
			}
		}
	}

	// Method for hadling when either backspace or delete is pressed.
	void CInputTyping::HandleBackAndDelete(std::string& text, int& caretIndex)
	{
		if(text.length() <= 0)
			return;

		if(caretIndex > 0)
		{
			// Backspace keystroke.
			if(m_pInput->KeyDown(VK_BACK))
			{
				text.erase(caretIndex - 1, 1);
				caretIndex--;
				m_keyTimers[VK_BACK] = 0.0f;
			}
			else if(m_pInput->Key(VK_BACK))
			{
				m_keyTimers[VK_BACK] += m_pTimer->Delta();

				if(m_keyTimers[VK_BACK] >= TYPING_KEY_HELD_TIME)
				{
					text.erase(caretIndex - 1, 1);
					caretIndex--;
					m_keyTimers[VK_BACK] -= TYPING_KEY_SENSITIVITY;
				}
			}
		}

		if(caretIndex <= (int)text.length())
		{
			// Delete keystroke.
			if(m_pInput->KeyDown(VK_DELETE))
			{
				text.erase(caretIndex, 1);
				if((int)text.length() < caretIndex)
					caretIndex--;

				m_keyTimers[VK_DELETE] = 0.0f;
			}
			else if(m_pInput->Key(VK_DELETE))
			{
				m_keyTimers[VK_DELETE] += m_pTimer->Delta();

				if(m_keyTimers[VK_DELETE] >= TYPING_KEY_HELD_TIME)
				{
					text.erase(caretIndex, 1);
					if((int)text.length() < caretIndex)
						caretIndex--;

					m_keyTimers[VK_DELETE] -= TYPING_KEY_SENSITIVITY;
				}
			}
		}
	}

	// Method for handling arrow keys.
	void CInputTyping::HandleArrowKeys(const std::string& text, int& caretIndex)
	{
		if(text.length() <= 0)
			return;
			
		if(caretIndex > 0)
		{
			// Backspace keystroke.
			if(m_pInput->KeyDown(VK_LEFT))
			{
				caretIndex--;
				m_keyTimers[VK_LEFT] = 0.0f;
			}
			else if(m_pInput->Key(VK_LEFT))
			{
				m_keyTimers[VK_LEFT] += m_pTimer->Delta();

				if(m_keyTimers[VK_LEFT] >= TYPING_KEY_HELD_TIME)
				{
					caretIndex--;
					m_keyTimers[VK_LEFT] -= TYPING_KEY_SENSITIVITY;
				}
			}
		}

		if(caretIndex < (int)text.length())
		{
			// Backspace keystroke.
			if(m_pInput->KeyDown(VK_RIGHT))
			{
				caretIndex++;
				m_keyTimers[VK_RIGHT] = 0.0f;
			}
			else if(m_pInput->Key(VK_RIGHT))
			{
				m_keyTimers[VK_RIGHT] += m_pTimer->Delta();

				if(m_keyTimers[VK_RIGHT] >= TYPING_KEY_HELD_TIME)
				{
					caretIndex++;
					m_keyTimers[VK_RIGHT] -= TYPING_KEY_SENSITIVITY;
				}
			}
		}
	}
};