//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CXAudio.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CXAudio2.h"

namespace Audio
{
	//---------------------------------------------------------------------------------------------
	// XAudio2 Engine
	//---------------------------------------------------------------------------------------------

	CXAudio2Engine::CXAudio2Engine()
	{
		m_volume							= 1.0f;

		m_pX3dInstance						= NULL;
		m_pX3dListener						= NULL;

		m_pXAudio2							= NULL;
		m_pMasterVoice						= NULL;

		m_pListenerTransform				= NULL;
	}

	CXAudio2Engine::CXAudio2Engine(const CXAudio2Engine& other)
	{
	}

	CXAudio2Engine::~CXAudio2Engine()
	{
	}

	// Method for initializing the xaudio2 engine and listener, along with x3daudio.
	int CXAudio2Engine::Initialize()
	{
		int retCode = Util::Err::NONE;

		if(FAILED(XAudio2Create(&m_pXAudio2, 0, XAUDIO2_DEFAULT_PROCESSOR)))
		{
			return Util::Err::ERR_XAUDIO2_FAILED_TO_CREATE_SOUND_ENGINE;
		}

		if(FAILED(m_pXAudio2->CreateMasteringVoice(&m_pMasterVoice)))
		{
			return Util::Err::ERR_XAUDIO2_FAILED_TO_CREATE_MASTER_VOICE;
		}
		
		DWORD channelMask;
		m_pMasterVoice->GetChannelMask(&channelMask);
		if(FAILED(X3DAudioInitialize(channelMask, X3DAUDIO_SPEED_OF_SOUND, m_X3dInstance)))
		{
			return Util::Err::ERR_XAUDIO2_FAILED_INITIALIZE_X3DAUDIO;
		}

		ZeroMemory(&m_X3dListener, sizeof(m_X3dListener));

		m_pX3dInstance = &m_X3dInstance;
		m_pX3dListener = &m_X3dListener;

		return retCode;
	}

	// Method for updating the listener position, and orientation, along with the master volume.
	int CXAudio2Engine::Update()
	{
		int retCode = Util::Err::NONE;

		if(m_pListenerTransform)
		{
			Math::CVector3 p = m_pListenerTransform->Position();
			Math::CVector3 f = m_pListenerTransform->Forward();
			Math::CVector3 u = m_pListenerTransform->Up();
			
			X3DAUDIO_VECTOR position = { p.x, p.y, p.z };
			X3DAUDIO_VECTOR front = { f.x, f.y, f.z };
			X3DAUDIO_VECTOR up = { u.x, u.y, u.z };

			m_X3dListener.Position = position;
			m_X3dListener.OrientFront = front;
			m_X3dListener.OrientTop = up;
		}

		m_pMasterVoice->SetVolume(m_volume);

		return retCode;
	}

	void CXAudio2Engine::Release()
	{
		SAFE_RELEASE(m_pXAudio2);
	}

	//---------------------------------------------------------------------------------------------
	// XAudio2 Source
	//---------------------------------------------------------------------------------------------

	CXAudio2Source::CXAudio2Source()
	{
		m_b3dSound							= false;
		m_bLoop								= false;
		m_volume							= 1.0f;

		m_distanceScale						= 1.0f;
		m_outerCone							= X3DAUDIO_2PI;
		m_innerCone							= X3DAUDIO_2PI;

		m_matrixSize						= 0;
		m_pMatrix							= NULL;
		
		m_pXAudioEngine						= NULL;
		m_pXAudio2							= NULL;
		m_pWaveFile							= NULL;
	}

	CXAudio2Source::CXAudio2Source(const CXAudio2Source& other)
	{
	}

	CXAudio2Source::~CXAudio2Source()
	{
	}

	// Method for initializing a new audio source with default and specified values in place.
	int CXAudio2Source::Initialize(const std::string& filename, CXAudio2Engine*& audioEngine)
	{
		int retCode = Util::Err::NONE;

		m_pXAudioEngine		= audioEngine;
		m_pXAudio2			= audioEngine->m_pXAudio2;
		m_filename			= filename;

		ZeroMemory(&m_buffer, sizeof(m_buffer));
		ZeroMemory(&m_wfx, sizeof(m_wfx));

		// Open specified file.
		m_pWaveFile = new CWaveFile();
		if((retCode = m_pWaveFile->OpenFile(filename, m_wfx, m_buffer)) != Util::Err::NONE)
			return retCode;

		m_buffer.LoopCount = m_bLoop ? XAUDIO2_LOOP_INFINITE : 0;
		
		// Setup this audio source for being used as a 3d sound.
		XAUDIO2_VOICE_DETAILS details;
		audioEngine->m_pMasterVoice->GetVoiceDetails(&details);

		u32 input		= m_wfx.Format.nChannels;
		u32 output		= details.InputChannels;
		m_matrixSize	= input * output;
		m_pMatrix		= new float[m_matrixSize];

		ZeroMemory(m_pMatrix, sizeof(float) * m_matrixSize);
		ZeroMemory(&m_emitter, sizeof(m_emitter));
		ZeroMemory(&m_dspSettings, sizeof(m_dspSettings));
		ZeroMemory(&m_cone, sizeof(m_cone));

		// Setup the audio settings, emitter, and cone.
		m_dspSettings.SrcChannelCount		= input;
		m_dspSettings.DstChannelCount		= output;
		m_dspSettings.pMatrixCoefficients	= m_pMatrix;

		m_emitter.ChannelCount				= 1;
		m_emitter.CurveDistanceScaler		= m_distanceScale;

		m_cone.OuterAngle					= m_outerCone;
		m_cone.OuterVolume					= 0.0f;
		m_cone.InnerAngle					= m_innerCone;
		m_cone.InnerVolume					= 1.0f;
		m_emitter.pCone						= &m_cone;

		return retCode;
	}

	// Method for updating this audio source's position, orientation, and other 3d data.
	int CXAudio2Source::Update()
	{
		int retCode = Util::Err::NONE;

		if(m_b3dSound)
		{
			m_emitter.CurveDistanceScaler		= m_distanceScale;

			m_cone.OuterAngle					= m_outerCone;
			m_cone.InnerAngle					= m_innerCone;

			// Set all stored audio source voices to the updated audio output matrix.
			XAUDIO2_VOICE_STATE state;
			for(u32 i = 0; i < m_sSourceVoiceStack.Length(); i++)
			{
				m_sSourceVoiceStack[i].voice->GetState(&state);
				if(state.BuffersQueued)
				{
					m_emitter.Position = { m_sSourceVoiceStack[i].position.x, m_sSourceVoiceStack[i].position.y, m_sSourceVoiceStack[i].position.z };
					m_emitter.OrientFront = { m_sSourceVoiceStack[i].front.x, m_sSourceVoiceStack[i].front.y, m_sSourceVoiceStack[i].front.z };
				
					X3DAudioCalculate(m_pXAudioEngine->m_X3dInstance, m_pXAudioEngine->m_pX3dListener, &m_emitter,
						X3DAUDIO_CALCULATE_MATRIX, &m_dspSettings);

					m_sSourceVoiceStack[i].voice->SetOutputMatrix(m_pXAudioEngine->m_pMasterVoice, m_dspSettings.SrcChannelCount, 
						m_dspSettings.DstChannelCount, m_dspSettings.pMatrixCoefficients);
				}
			}
		}

		return retCode;
	}

	// Method for playing this audio source. (Creates a new voice source if all others are busy.
	int CXAudio2Source::Play(float volume, bool loop, Math::CVector3 position, Math::CVector3 front, AUDIO_VOICE** voiceOut)
	{
		int retCode = Util::Err::NONE;

		m_bLoop = loop;
		m_volume = volume;
		
		m_buffer.LoopCount = m_bLoop ? XAUDIO2_LOOP_INFINITE : 0;

		XAUDIO2_VOICE_STATE state;
		bool played = false;

		// Check for a source voice that is ready to play to sound.
		for(u32 i = 0; i < m_sSourceVoiceStack.Length(); i++)
		{
			m_sSourceVoiceStack[i].voice->GetState(&state);
			if(state.BuffersQueued == 0)
			{
				Play(m_sSourceVoiceStack[i].voice);
				played = true;
				break;
			}
		}

		// If there is no source voice ready to play the sound, push a new source voice to the stack, and play it in the new source voice.
		if(!played)
		{
			IXAudio2SourceVoice* voice;
			if(FAILED(m_pXAudio2->CreateSourceVoice(&voice, (WAVEFORMATEX*)&m_wfx)))
			{
				return Util::Err::ERR_XAUDIO2_FAILED_TO_CREATE_SOURCE_VOICE;
			}

			AUDIO_VOICE audioVoice;

			audioVoice.position		= position;
			audioVoice.front		= front;
			audioVoice.voice		= voice;

			m_sSourceVoiceStack.Push(audioVoice);
			if(voiceOut)
				*voiceOut = &m_sSourceVoiceStack[m_sSourceVoiceStack.Length() - 1];

			Play(voice);
		}
		
		return retCode;
	}

	// Method for playing a audio source in a specified source voice.
	int CXAudio2Source::Play(IXAudio2SourceVoice* voice)
	{
		int retCode = Util::Err::NONE;

		if(FAILED(voice->SubmitSourceBuffer(&m_buffer)))
		{
			return Util::Err::ERR_XAUDIO2_FAILED_TO_SUBMIT_SOURCE_BUFFER;
		}
		
		voice->SetVolume(m_volume);

		if(FAILED(voice->Start(0)))
		{
			return Util::Err::ERR_XAUDIO2_FAILED_TO_START_SOURCE_VOICE;
		}

		return retCode;
	}

	// Method for releasing this audio source.
	void CXAudio2Source::Release()
	{
		SAFE_DELETE_ARRAY(m_pMatrix);

		for(u32 i = 0; i < m_sSourceVoiceStack.Length(); i++)
		{
			if(m_sSourceVoiceStack[i].voice)
				m_sSourceVoiceStack[i].voice->DestroyVoice();
		}

		m_sSourceVoiceStack.Release();
		SAFE_DELETE(m_pWaveFile);
	}
};