//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CGeoObject.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CGeoObject.h"

namespace GFX
{
	CGeoObject::CGeoObject()
	{
		m_bEnableIndices					= true;
		m_bEnableInstancing					= false;
		m_bCreateBoundingAABB				= false;
		m_bRenderBoundingAABB				= false;
		m_bSeperateVerticesIndices			= false;

		m_pBufferArray						= NULL;
		m_pVertices							= NULL;
		m_pInstances						= NULL;
		m_pIndices16						= NULL;
		m_pIndices32						= NULL;
		
		m_totalCount						= 0;
		m_vertexCount						= 0;
		m_vertexSize						= 0;
		m_indexCount						= 0;
		m_indexSize							= 0;
		m_instanceCount						= 0;
		m_instanceAllocation				= 0;
		m_instanceSize						= 0;

		m_vertexStride						= 0;
		m_vertexPositionOffset				= 0;
		m_instanceStride					= 0;
		m_instancePositionOffset			= 0;
		m_indexStride						= 0;

		m_primitiveTopology					= D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST;

		m_pDebugMaterial					= NULL;
		m_pDebugBounds						= NULL;

		m_pRenderData						= NULL;

		m_pVertexBuffer						= NULL;
		m_pInstanceBuffer					= NULL;
		m_pIndexBuffer						= NULL;
	}

	CGeoObject::CGeoObject(const CGeoObject& other)
	{
	}

	CGeoObject::~CGeoObject()
	{
	}

	// Method for initializing the geometrico objects data.
	// Create vertex, index, and instance buffers, if they a implied.
	// Also creates a bound volume around them to help with optimizing rendering.
	int CGeoObject::Initialize()
	{
		int retCode = Util::Err::NONE;

		if(m_vertexCount > 0)
		{
			// Calculate the size of each buffer.
			if(m_bEnableIndices && !m_bSeperateVerticesIndices)
				m_indexCount	= m_vertexCount;

			m_vertexSize		= m_vertexCount * m_vertexStride;

			if(m_bEnableIndices)
				m_indexSize		= m_indexCount * m_indexStride;
			if(m_bEnableInstancing)
				m_instanceSize	= m_instanceAllocation * m_instanceStride;

			// Create arrays for storing buffer data about this object.
			m_totalCount = m_vertexSize + m_indexSize + m_instanceSize;
			m_pBufferArray = new u8[m_totalCount];

			m_pVertices = (void*)(m_pBufferArray);
			
			if(m_bEnableIndices)
			{
				if(m_indexStride == 2)	m_pIndices16 = (u16*)(m_pBufferArray + m_vertexSize);
				else					m_pIndices32 = (u32*)(m_pBufferArray + m_vertexSize);
			}
			if(m_bEnableInstancing)
			{
				m_pInstances = (void*)(m_pBufferArray + m_vertexSize + m_indexSize);
			}

			void* vertex;
			u32 index = 0;

			if(m_bCreateBoundingAABB)
			{
				m_pDebugBounds = new CDebugCube();
				m_pDebugBounds->DirectXData(m_pDX11Data);
				m_pDebugBounds->Material(m_pDebugMaterial);
			}

			m_minBounds = FLT_MAX;
			m_maxBounds = FLT_MIN;

			Math::CVector3 vertexPos;

			// Add all of the queued vertices to the vertex buffer, and set their index.
			while(vertex = m_vertexQueue.Pop())
			{
				memcpy(((u8*)m_pVertices + (index * m_vertexStride)), ((u8*)vertex), m_vertexStride);

				if(m_bEnableIndices && !m_bSeperateVerticesIndices)
				{
					if(m_pIndices16)    m_pIndices16[index] = index;
					else				m_pIndices32[index] = index;
				}

				if(m_bCreateBoundingAABB)
				{
					vertexPos = (*(Math::CVector3*)((u8*)m_pVertices + (index * m_vertexStride) + m_vertexPositionOffset));

					m_minBounds.x = min(m_minBounds.x, vertexPos.x);
					m_minBounds.y = min(m_minBounds.y, vertexPos.y);
					m_minBounds.z = min(m_minBounds.z, vertexPos.z);

					m_maxBounds.x = max(m_maxBounds.x, vertexPos.x);
					m_maxBounds.y = max(m_maxBounds.y, vertexPos.y);
					m_maxBounds.z = max(m_maxBounds.z, vertexPos.z);
				}

				delete vertex;
				index++;
			}
			
			// If indices are defined seperate from the vertices, process them here.
			if(m_bEnableIndices && m_bSeperateVerticesIndices)
			{
				index = 0;
				u32 indice = 0;

				while(m_indexQueue.Pop(indice))
				{
					if(m_pIndices16)    m_pIndices16[index] = indice;
					else				m_pIndices32[index] = indice;

					index++;
				}
			}

			if(m_bCreateBoundingAABB)
			{
				m_pDebugBounds->CubeExtents(m_minBounds, m_maxBounds);
				if((retCode = m_pDebugBounds->Initialize()) != Util::Err::NONE)
					return retCode;

				m_pDebugBounds->m_debugCBuffer.color = Math::CColor(0, 1, 0);
			}

			D3D11_SUBRESOURCE_DATA data;
			D3D11_BUFFER_DESC bd;
			ZeroMemory(&bd, sizeof(bd));

			// Create the vertex buffer.
			bd.Usage						= D3D11_USAGE_DYNAMIC;
			bd.ByteWidth					= m_vertexSize;
			bd.BindFlags					= D3D11_BIND_VERTEX_BUFFER;
			bd.CPUAccessFlags				= D3D11_CPU_ACCESS_WRITE;
			bd.MiscFlags					= 0;
			bd.StructureByteStride			= 0;
			data.pSysMem					= m_pVertices;
			
			if(FAILED(m_pDev->CreateBuffer(&bd, &data, &m_pVertexBuffer)))
			{
				return Util::Err::ERR_GEOMETRY_FAILED_TO_CREATE_VERTEX_BUFFER;
			}

			if(m_bEnableInstancing)
			{
				// Create the instance buffer.
				bd.ByteWidth					= m_instanceSize;
				data.pSysMem					= m_pInstances;
			
				if(FAILED(m_pDev->CreateBuffer(&bd, &data, &m_pInstanceBuffer)))
				{
					return Util::Err::ERR_GEOMETRY_FAILED_TO_CREATE_INSTANCE_BUFFER;
				}
			}
			
			if(m_bEnableIndices)
			{
				// Create the index buffer.
				bd.ByteWidth					= m_indexSize;
				bd.BindFlags					= D3D11_BIND_INDEX_BUFFER;
				data.pSysMem					= (m_pIndices16 ? (void*)m_pIndices16 : (void*)m_pIndices32);
			
				if(FAILED(m_pDev->CreateBuffer(&bd, &data, &m_pIndexBuffer)))
				{
					return Util::Err::ERR_GEOMETRY_FAILED_TO_CREATE_INDEX_BUFFER;
				}
			}
		}

		if((retCode = CGraphicalObject::Initialize()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	int CGeoObject::Update()
	{
		int retCode = Util::Err::NONE;

		return retCode;
	}

	// Method for recalculating a geometric object's bounds.
	void CGeoObject::RecalculateBounds()
	{
		Math::CVector3 bounds0 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_minBounds.x, m_minBounds.y, m_minBounds.z));
		Math::CVector3 bounds1 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_maxBounds.x, m_minBounds.y, m_minBounds.z));
		Math::CVector3 bounds2 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_minBounds.x, m_maxBounds.y, m_minBounds.z));
		Math::CVector3 bounds3 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_minBounds.x, m_minBounds.y, m_maxBounds.z));
		Math::CVector3 bounds4 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_maxBounds.x, m_maxBounds.y, m_minBounds.z));
		Math::CVector3 bounds5 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_minBounds.x, m_maxBounds.y, m_maxBounds.z));
		Math::CVector3 bounds6 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_maxBounds.x, m_minBounds.y, m_maxBounds.z));
		Math::CVector3 bounds7 = m_pTransform->WorldMatrix().TransformCoord(Math::CVector3(m_maxBounds.x, m_maxBounds.y, m_maxBounds.z));

		m_minBoundsTransformed = Math::CVector3::Min(
			Math::CVector3::Min(Math::CVector3::Min(Math::CVector3::Min(bounds0, bounds1), bounds2), bounds3),
			Math::CVector3::Min(Math::CVector3::Min(Math::CVector3::Min(bounds4, bounds5), bounds6), bounds7)
		);

		m_maxBoundsTransformed = Math::CVector3::Max(
			Math::CVector3::Max(Math::CVector3::Max(Math::CVector3::Max(bounds0, bounds1), bounds2), bounds3),
			Math::CVector3::Max(Math::CVector3::Max(Math::CVector3::Max(bounds4, bounds5), bounds6), bounds7)
		);
	}

	// Method for rendering the geometic object with it's custom material.
	int CGeoObject::Render(bool count)
	{
		int retCode = Util::Err::NONE;
		
		if((retCode = Render(m_pMaterial, count)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for rendering a geometric object with an inputted custom material.
	int CGeoObject::Render(CMaterial* material, bool count)
	{
		int retCode = Util::Err::NONE;
		
		bool renderFlag = !m_bCreateBoundingAABB || m_bEnableInstancing; // CPU frustum culling is not supported with instances.
		
		if(!renderFlag && m_pTransform)
		{
			RecalculateBounds();

			renderFlag = material->Camera()->Frustum().TestAABB(Math::CAABB(m_minBoundsTransformed, 
				m_maxBoundsTransformed, m_pTransform->Position()));

			if(m_bRenderBoundingAABB)
			{
				m_pDebugBounds->Transform(m_pTransform);
				m_pDebugBounds->RecalculateBounds();
				if((retCode = m_pDebugBounds->Render()) != Util::Err::NONE)
					return retCode;
			}
		}

		// Render this object if any buffer data exists in it.
		if(renderFlag && m_totalCount > 0)
		{
			if(m_pTransform)
				material->Transform(m_pTransform);

			if((retCode = material->Render()) != Util::Err::NONE)
				return retCode;

			if(m_bEnableInstancing)
				RenderInstanced(material == m_pMaterial && count);
			else
				RenderDefault(material == m_pMaterial && count);
		}

		if((retCode = CGraphicalObject::Render()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for rendering an object with no instancing.
	int CGeoObject::RenderDefault(bool count)
	{
		int retCode = Util::Err::NONE;
		
		u32 stride = m_vertexStride;
		u32 offset = 0;

		m_pDevCon->IASetVertexBuffers(0, 1, &m_pVertexBuffer, &stride, &offset);

		if(m_bEnableIndices)
			m_pDevCon->IASetIndexBuffer(m_pIndexBuffer, m_indexStride == 2 ? DXGI_FORMAT_R16_UINT : DXGI_FORMAT_R32_UINT, 0);

		m_pDevCon->IASetPrimitiveTopology(m_primitiveTopology);
		
		if(m_bEnableIndices)
			m_pDevCon->DrawIndexed(m_indexCount, 0, 0);
		else
			m_pDevCon->Draw(m_vertexCount, 0);

		// Add the object data to render data is told to do so.
		if(count)
		{
			m_pRenderData->geoRenderCount += 1;
			m_pRenderData->polyRenderCount += m_indexCount / 3;
			m_pRenderData->vertRenderCount += m_vertexCount;
		}

		return retCode;
	}
	
	// Method for rendering an object with instancing.
	int CGeoObject::RenderInstanced(bool count)
	{
		int retCode = Util::Err::NONE;

		D3D11_MAPPED_SUBRESOURCE ms;
		m_pDevCon->Map(m_pInstanceBuffer, NULL, D3D11_MAP_WRITE_DISCARD, NULL, &ms);
		memcpy(ms.pData, m_pInstances, m_instanceCount * m_instanceStride);
		m_pDevCon->Unmap(m_pInstanceBuffer, NULL);

		u32 stride[2];
		u32 offset[2];
		ID3D11Buffer* bufferPointers[2];

		stride[0] = m_vertexStride;
		stride[1] = m_instanceStride;
		offset[0] = 0;
		offset[1] = 0;

		bufferPointers[0] = m_pVertexBuffer;
		bufferPointers[1] = m_pInstanceBuffer;

		m_pDevCon->IASetVertexBuffers(0, 2, bufferPointers, stride, offset);

		if(m_bEnableIndices)
			m_pDevCon->IASetIndexBuffer(m_pIndexBuffer, m_indexStride == 2 ? DXGI_FORMAT_R16_UINT : DXGI_FORMAT_R32_UINT, 0);

		m_pDevCon->IASetPrimitiveTopology(m_primitiveTopology);
		
		if(m_bEnableIndices)
			m_pDevCon->DrawIndexedInstanced(m_indexCount, m_instanceCount, 0, 0, 0);
		else
			m_pDevCon->DrawInstanced(m_vertexCount, m_instanceCount, 0, 0);

		// Add the object data to render data is told to do so.
		if(count)
		{
			m_pRenderData->geoRenderCount += m_instanceCount;
			m_pRenderData->polyRenderCount += (m_indexCount / 3) * m_instanceCount;
			m_pRenderData->vertRenderCount += m_vertexCount * m_instanceCount;
		}

		return retCode;
	}

	// Method for releasing the geometric object.
	void CGeoObject::Release()
	{
		SAFE_RELEASE(m_pIndexBuffer);
		SAFE_RELEASE(m_pInstanceBuffer);
		SAFE_RELEASE(m_pVertexBuffer);

		SAFE_RELEASE_DELETE(m_pDebugBounds);

		SAFE_DELETE_ARRAY(m_pBufferArray);

		m_indexQueue.Release();
		m_vertexQueue.Release();

		CGraphicalObject::Release();
	}

	//
	// Methods for adding vertices to an object.
	//

	void CGeoObject::AddVertex(void* v0)
	{
		m_vertexQueue.Push(v0);
		m_vertexCount++;
	}

	void CGeoObject::AddTriangle(void* v0, void* v1, void* v2, bool cw)
	{
		if(cw)	{ m_vertexQueue.Push(v0); m_vertexQueue.Push(v2); m_vertexQueue.Push(v1); }
		else	{ m_vertexQueue.Push(v0); m_vertexQueue.Push(v1); m_vertexQueue.Push(v2); }
		m_vertexCount += 3;
	}

	void CGeoObject::AddQuad(void* v0, void* v1, void* v2, void* v3, bool cw)
	{
		// Create two new vertices in order to keep every vertice pushed to the queue unique.
		void *v4 = new u8[m_vertexStride];
		void *v5 = new u8[m_vertexStride];

		memcpy(v4, v0, m_vertexStride);
		memcpy(v5, v2, m_vertexStride);

		if(cw)	{ AddTriangle(v0, v1, v2); AddTriangle(v4, v5, v3); }
		else	{ AddTriangle(v0, v2, v1); AddTriangle(v4, v3, v5); }
	}

	void CGeoObject::AddIndex(u32 index)
	{
		m_bSeperateVerticesIndices = true;

		m_indexQueue.Push(index);
		m_indexCount++;
	}
};