//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CLightDirectional.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CLIGHTDIRECTIONAL_H
#define CLIGHTDIRECTIONAL_H

#include "CConstantBuffer.h"
#include "..\Math\CMathVector2.h"
#include "..\Math\CMathCamera.h"
#include "CTextureRenderTo.h"
#include "CRenderingData.h"
#include "CMatShadowMap.h"
#include "CDebugSphere.h"
#include "..\Utilities\CDataQueue.h"

namespace GFX
{
	// Directional light.
	struct DIRECTIONAL_LIGHT
	{
		Math::CMatrix4x4				viewProj;
		Math::CColor					color;
		Math::CVector4					direction;
		float							strength;
		int								shadowsOn;
		float							shadowStrength;
		float							padding0[1];
	};

	// Directional light queue structure.
	struct DIRECTIONAL_LIGHT_QUEUE : public DIRECTIONAL_LIGHT
	{
		DIRECTIONAL_LIGHT** lightPtr;

		DIRECTIONAL_LIGHT_QUEUE()
		{
		}

		DIRECTIONAL_LIGHT_QUEUE(const Math::CColor& Color, const Math::CVector4& Direction, float Strength, int ShadowsOn, 
			float ShadowStrength, DIRECTIONAL_LIGHT** ptr = NULL)
		{
			color						= Color;
			direction					= Direction;
			strength					= Strength;
			shadowsOn					= ShadowsOn;
			shadowStrength				= ShadowStrength;
			lightPtr					= ptr;
		}
	};

	// CLightDirectional class
	// Class for creating directional lights that used with deferred rendering.
	class CLightDirectional : public App::CDX11Object
	{
	public:
		friend class CMatPostScreen;
		friend class CLighting;
		
	public:
		CLightDirectional();
		CLightDirectional(const CLightDirectional& other);
		virtual ~CLightDirectional();

		int									Initialize();

		int									ResetBuffer();
		int									BindBuffer();

		int									RenderMaps(void* data, int (*RenderWorldObjects)(void*, CMaterial*));
		int									Render();
		
		void								Release();

		inline void							AddDirectionalLight(DIRECTIONAL_LIGHT_QUEUE* dirLight);

		inline RENDERDATA*					RenderData() const;
		inline void							RenderData(RENDERDATA* pRenderData);

		inline CMatDebug*					DebugMaterial() const;
		inline void							DebugMaterial(CMatDebug* pMatDebug);
		inline Math::CCamera*				CurrentCamera() const;
		inline void							CurrentCamera(Math::CCamera* pCurrentCamera);

	private:
		u32									m_dirLightCount;
		std::string							m_dirLightCountStr;

		u32									m_shadowMapSize;
		
		u32									m_dirShadowMapCount;
		DIRECTIONAL_LIGHT*					m_pDirLightArray;
		Math::CCamera**						m_pDirLightPOVs;
		CTextureRenderTo*					m_pDirShadowMaps;

		CMatShadowMap*						m_pMatShadowMap;
		RENDERDATA*							m_pRenderData;

		Util::CPtrQueue						m_qDirLightQueue;
		
		CConstantBuffer*					m_pDirLightCBuffer;

		CMatDebug*							m_pMatDebug;
		Math::CCamera*						m_pCurrentCamera;
	};
	
	// Mathod for adding a directional light to the directional light queue.
	inline void CLightDirectional::AddDirectionalLight(DIRECTIONAL_LIGHT_QUEUE* dirLight)
	{
		m_qDirLightQueue.Push(dirLight);
		m_dirLightCount++;
	}

	// Render data getter and setter.

	inline RENDERDATA* CLightDirectional::RenderData() const
	{
		return m_pRenderData;
	}

	inline void CLightDirectional::RenderData(RENDERDATA* pRenderData)
	{
		m_pRenderData = pRenderData;
	}

	// Debug material getter and setter.

	inline CMatDebug* CLightDirectional::DebugMaterial() const
	{
		return m_pMatDebug;
	}

	inline void CLightDirectional::DebugMaterial(CMatDebug* pMatDebug)
	{
		m_pMatDebug = pMatDebug;
	}

	// Current camera getter and setter.

	inline Math::CCamera* CLightDirectional::CurrentCamera() const
	{
		return m_pCurrentCamera;
	}

	inline void CLightDirectional::CurrentCamera(Math::CCamera* pCurrentCamera)
	{
		m_pCurrentCamera = pCurrentCamera;
	}
};

#endif
