//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CLightPoint.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CLIGHTPOINT_H
#define CLIGHTPOINT_H

#include "CConstantBuffer.h"
#include "..\Math\CMathVector2.h"
#include "..\Math\CMathCamera.h"
#include "CTextureRenderTo.h"
#include "CRenderingData.h"
#include "CMatShadowMap.h"
#include "CDebugSphere.h"
#include "..\Utilities\CDataQueue.h"

namespace GFX
{
	// Point light.
	struct POINT_LIGHT
	{
		Math::CMatrix4x4				viewProj[6];
		Math::CColor					color;
		Math::CVector4					position;
		Math::CVector4					attenuation;
		float							range;
		float							strength;
		int								shadowsOn;
		float							shadowStrength;
	};
	
	// Point light queue structure.
	struct POINT_LIGHT_QUEUE : public POINT_LIGHT
	{
		POINT_LIGHT** lightPtr;

		POINT_LIGHT_QUEUE()
		{
		}

		POINT_LIGHT_QUEUE(const Math::CColor& Color, const Math::CVector4& Position, const Math::CVector4& Attenuation, 
			float Range, float Strength, int ShadowsOn, float ShadowStrength, POINT_LIGHT** ptr = NULL)
		{
			color						= Color;
			position					= Position;
			attenuation					= Attenuation;
			range						= Range;
			strength					= Strength;
			shadowsOn					= ShadowsOn;
			shadowStrength				= ShadowStrength;
			lightPtr					= ptr;
		}
	};
	
	// CLightPoint class
	// Class for creating point lights that are used with deferred rendering.
	class CLightPoint : public App::CDX11Object
	{
	public:
		friend class CMatPostScreen;
		friend class CLighting;
		
	public:
		CLightPoint();
		CLightPoint(const CLightPoint& other);
		virtual ~CLightPoint();

		int									Initialize();

		int									ResetBuffer();
		int									BindBuffer();

		int									RenderMaps(void* data, int (*RenderWorldObjects)(void*, CMaterial*));
		int									Render();
		
		void								Release();

		inline void							AddPointLight(POINT_LIGHT_QUEUE* pointLight);
		
		inline RENDERDATA*					RenderData() const;
		inline void							RenderData(RENDERDATA* pRenderData);

		inline CMatDebug*					DebugMaterial() const;
		inline void							DebugMaterial(CMatDebug* pMatDebug);
		inline Math::CCamera*				CurrentCamera() const;
		inline void							CurrentCamera(Math::CCamera* pCurrentCamera);

	private:
		u32									m_pointLightCount;
		std::string							m_pointLightCountStr;

		CDebugSphere*						m_pDebugSphere;

		u32									m_shadowMapSize;
		
		u32									m_pointShadowMapCount;
		POINT_LIGHT*						m_pPointLightArray;
		Math::CCamera**						m_pPointLightPOVs;
		CTextureRenderTo*					m_pPointShadowMaps;
		
		CMatShadowMap*						m_pMatShadowMap;
		RENDERDATA*							m_pRenderData;

		Util::CPtrQueue						m_qPointLightQueue;
		
		CConstantBuffer*					m_pPointLightCBuffer;

		CMatDebug*							m_pMatDebug;
		Math::CCamera*						m_pCurrentCamera;
	};

	// Mathod for adding a point light to the point light queue.
	inline void CLightPoint::AddPointLight(POINT_LIGHT_QUEUE* pointLight)
	{
		m_qPointLightQueue.Push(pointLight);
		m_pointLightCount++;
	}

	// Render data getter and setter.

	inline RENDERDATA* CLightPoint::RenderData() const
	{
		return m_pRenderData;
	}

	inline void CLightPoint::RenderData(RENDERDATA* pRenderData)
	{
		m_pRenderData = pRenderData;
	}

	// Debug material getter and setter.

	inline CMatDebug* CLightPoint::DebugMaterial() const
	{
		return m_pMatDebug;
	}

	inline void CLightPoint::DebugMaterial(CMatDebug* pMatDebug)
	{
		m_pMatDebug = pMatDebug;
	}

	// Current camera getter and setter.

	inline Math::CCamera* CLightPoint::CurrentCamera() const
	{
		return m_pCurrentCamera;
	}

	inline void CLightPoint::CurrentCamera(Math::CCamera* pCurrentCamera)
	{
		m_pCurrentCamera = pCurrentCamera;
	}
};

#endif
