//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CLightSpot.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CLIGHTSPOT_H
#define CLIGHTSPOT_H

#include "CConstantBuffer.h"
#include "..\Math\CMathVector2.h"
#include "..\Math\CMathCamera.h"
#include "CTextureRenderTo.h"
#include "CRenderingData.h"
#include "CMatShadowMap.h"
#include "CDebugSphere.h"
#include "..\Utilities\CDataQueue.h"

namespace GFX
{
	// Spotlight.
	struct SPOTLIGHT
	{
		Math::CMatrix4x4				viewProj;
		Math::CColor					color;
		Math::CVector4					direction;
		Math::CVector4					position;
		Math::CVector4					attenuation;
		float							range;
		float							strength;
		float							falloff;
		float							phi;
		float							theta;
		int								shadowsOn;
		float							shadowStrength;
		float							padding0[1];
	};
	
	// Spotlight queue structure.
	struct SPOTLIGHT_QUEUE : public SPOTLIGHT
	{
		SPOTLIGHT** lightPtr;

		SPOTLIGHT_QUEUE()
		{
		}

		SPOTLIGHT_QUEUE(const Math::CColor& Color, const Math::CVector4& Direction, const Math::CVector4& Position, const Math::CVector4& Attenuation, 
			float Range, float Strength, float Falloff, float Phi, float Theta, int ShadowsOn, float ShadowStrength, SPOTLIGHT** ptr = NULL)
		{
			color						= Color;
			direction					= Direction;
			position					= Position;
			attenuation					= Attenuation;
			range						= Range;
			strength					= Strength;
			falloff						= Falloff;
			phi							= Phi;
			theta						= Theta;
			shadowsOn					= ShadowsOn;
			shadowStrength				= ShadowStrength;
			lightPtr					= ptr;
		}
	};
		
	// CLightPoint class
	// Class for creating spotlights that are used with deferred rendering.
	class CLightSpot : public App::CDX11Object
	{
	public:
		friend class CMatPostScreen;
		friend class CLighting;
		
	public:
		CLightSpot();
		CLightSpot(const CLightSpot& other);
		virtual ~CLightSpot();

		int									Initialize();

		int									ResetBuffer();
		int									BindBuffer();

		int									RenderMaps(void* data, int (*RenderWorldObjects)(void*, CMaterial*));
		int									Render();
		
		void								Release();

		inline void							AddSpotlight(SPOTLIGHT_QUEUE* spotlight);

		inline RENDERDATA*					RenderData() const;
		inline void							RenderData(RENDERDATA* pRenderData);

		inline CMatDebug*					DebugMaterial() const;
		inline void							DebugMaterial(CMatDebug* pMatDebug);
		inline Math::CCamera*				CurrentCamera() const;
		inline void							CurrentCamera(Math::CCamera* pCurrentCamera);

	private:
		u32									m_spotlightCount;
		std::string							m_spotlightCountStr;

		CDebugSphere*						m_pDebugSphere;

		u32									m_shadowMapSize;
		
		u32									m_spotShadowMapCount;
		SPOTLIGHT*							m_pSpotlightArray;
		Math::CCamera**						m_pSpotlightPOVs;
		CTextureRenderTo*					m_pSpotShadowMaps;

		CMatShadowMap*						m_pMatShadowMap;
		RENDERDATA*							m_pRenderData;

		Util::CPtrQueue						m_qSpotlightQueue;
		CConstantBuffer*					m_pSpotlightCBuffer;

		CMatDebug*							m_pMatDebug;
		Math::CCamera*						m_pCurrentCamera;
	};
	
	// Mathod for adding a spotlight to the spotlight queue.
	inline void CLightSpot::AddSpotlight(SPOTLIGHT_QUEUE* spotLight)
	{
		m_qSpotlightQueue.Push(spotLight);
		m_spotlightCount++;
	}

	// Render data getter and setter.

	inline RENDERDATA* CLightSpot::RenderData() const
	{
		return m_pRenderData;
	}

	inline void CLightSpot::RenderData(RENDERDATA* pRenderData)
	{
		m_pRenderData = pRenderData;
	}

	// Debug material getter and setter.

	inline CMatDebug* CLightSpot::DebugMaterial() const
	{
		return m_pMatDebug;
	}

	inline void CLightSpot::DebugMaterial(CMatDebug* pMatDebug)
	{
		m_pMatDebug = pMatDebug;
	}

	// Current camera getter and setter.

	inline Math::CCamera* CLightSpot::CurrentCamera() const
	{
		return m_pCurrentCamera;
	}

	inline void CLightSpot::CurrentCamera(Math::CCamera* pCurrentCamera)
	{
		m_pCurrentCamera = pCurrentCamera;
	}
};

#endif
