//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CLighting.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CLighting.h"

namespace GFX
{
	CLighting::CLighting()
	{
		m_bDebugMode						= false;

		m_pDebugSphere						= NULL;

		m_shadowMapSize						= 256;
		
		m_pMatShadowMap						= NULL;
		m_pRenderData						= NULL;
		
		m_pLightDirectional					= new CLightDirectional();
		m_pLightPoint						= new CLightPoint();
		m_pLightSpot						= new CLightSpot();

		m_pMatDebug							= NULL;
		m_pCurrentCamera					= NULL;
	}

	CLighting::CLighting(const CLighting& other)
	{
		
	}

	CLighting::~CLighting()
	{
		
	}

	// Method for Initializing 
	int CLighting::Initialize()
	{
		int retCode = Util::Err::NONE;

		// Allocate memory.
		m_pDebugSphere						= new CDebugSphere();
		m_pMatShadowMap						= new CMatShadowMap();

		//
		// Debug elements.
		//

		// Debug sphere.
		m_pDebugSphere->DirectXData(m_pDX11Data);
		m_pDebugSphere->SetupSphere(m_pCurrentCamera, 1.0f);

		m_pDebugSphere->Material(m_pMatDebug);
		m_pDebugSphere->RenderData(m_pRenderData);

		m_pDebugSphere->Transform(Math::CTransformSharedPtr(new Math::CTransform()));
		m_pDebugSphere->Transform()->Update();

		if((retCode = m_pDebugSphere->Initialize()) != Util::Err::NONE)
			return retCode;

		//
		// Lighting and shadow mapping.
		//

		// Setup the shadow map material.
		m_pMatShadowMap->DirectXData(m_pDX11Data);
		if((retCode = m_pMatShadowMap->Initialize()) != Util::Err::NONE)
			return retCode;
		
		// Directional lights.
		m_pLightDirectional->DirectXData(m_pDX11Data);

		m_pLightDirectional->m_pMatShadowMap	= m_pMatShadowMap;
		m_pLightDirectional->m_pCurrentCamera	= m_pCurrentCamera;
		m_pLightDirectional->m_pMatDebug		= m_pMatDebug;
		m_pLightDirectional->m_shadowMapSize	= m_shadowMapSize * 4; // Quadruple the shadow map size for direction lights.

		if((retCode = m_pLightDirectional->Initialize()) != Util::Err::NONE)
			return retCode;

		// Point lights.
		m_pLightPoint->DirectXData(m_pDX11Data);

		m_pLightPoint->m_pMatShadowMap		= m_pMatShadowMap;
		m_pLightPoint->m_pCurrentCamera		= m_pCurrentCamera;
		m_pLightPoint->m_pMatDebug			= m_pMatDebug;
		m_pLightPoint->m_pDebugSphere		= m_pDebugSphere;
		m_pLightPoint->m_shadowMapSize		= m_shadowMapSize;

		if((retCode = m_pLightPoint->Initialize()) != Util::Err::NONE)
			return retCode;

		// Spot lights.
		m_pLightSpot->DirectXData(m_pDX11Data);

		m_pLightSpot->m_pMatShadowMap		= m_pMatShadowMap;
		m_pLightSpot->m_pCurrentCamera		= m_pCurrentCamera;
		m_pLightSpot->m_pMatDebug			= m_pMatDebug;
		m_pLightSpot->m_pDebugSphere		= m_pDebugSphere;
		m_pLightSpot->m_shadowMapSize		= m_shadowMapSize;

		if((retCode = m_pLightSpot->Initialize()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	int CLighting::Finalize()
	{
		int retCode = Util::Err::NONE;
		
		m_pLightDirectional->m_pCurrentCamera = m_pCurrentCamera;
		m_pLightDirectional->m_pRenderData = m_pRenderData;

		m_pLightPoint->m_pCurrentCamera = m_pCurrentCamera;
		m_pLightPoint->m_pRenderData = m_pRenderData;

		m_pLightSpot->m_pCurrentCamera = m_pCurrentCamera;
		m_pLightSpot->m_pRenderData = m_pRenderData;

		return retCode;
	}

	int CLighting::ResetBuffer()
	{
		int retCode = Util::Err::NONE;
		
		if(m_pLightDirectional) if((retCode = m_pLightDirectional->ResetBuffer()) != Util::Err::NONE) return retCode;
		if(m_pLightPoint) if((retCode = m_pLightPoint->ResetBuffer()) != Util::Err::NONE) return retCode;
		if(m_pLightSpot) if((retCode = m_pLightSpot->ResetBuffer()) != Util::Err::NONE) return retCode;

		return retCode;
	}

	int CLighting::BindBuffer()
	{
		int retCode = Util::Err::NONE;
		
		if(m_pLightDirectional) if((retCode = m_pLightDirectional->BindBuffer()) != Util::Err::NONE) return retCode;
		if(m_pLightPoint) if((retCode = m_pLightPoint->BindBuffer()) != Util::Err::NONE) return retCode;
		if(m_pLightSpot) if((retCode = m_pLightSpot->BindBuffer()) != Util::Err::NONE) return retCode;

		return retCode;
	}

	int CLighting::RenderMaps(void* data, int (*RenderWorldObjects)(void*, CMaterial*))
	{
		int retCode = Util::Err::NONE;
		
		if((retCode = m_pLightDirectional->RenderMaps(data, RenderWorldObjects)) != Util::Err::NONE)
			return retCode;

		if((retCode = m_pLightPoint->RenderMaps(data, RenderWorldObjects)) != Util::Err::NONE)
			return retCode;

		if((retCode = m_pLightSpot->RenderMaps(data, RenderWorldObjects)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}
	
	int CLighting::Render()
	{
		int retCode = Util::Err::NONE;
		
		if(m_bDebugMode)
		{
			if((retCode = m_pLightDirectional->Render()) != Util::Err::NONE)
				return retCode;

			if((retCode = m_pLightPoint->Render()) != Util::Err::NONE)
				return retCode;

			if((retCode = m_pLightSpot->Render()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}	
	
	void CLighting::Release()
	{
		SAFE_RELEASE_DELETE(m_pLightSpot);
		SAFE_RELEASE_DELETE(m_pLightPoint);
		SAFE_RELEASE_DELETE(m_pLightDirectional);

		SAFE_RELEASE_DELETE(m_pMatShadowMap);

		SAFE_RELEASE_DELETE(m_pDebugSphere);
	}
};