//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CMatMesh.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATMESH_H
#define CMATMESH_H

#include "CMaterial.h"
#include "..\Utilities\CDataStack.h"
#include <fbxsdk.h>
#include "CTextureList.h"

#pragma comment(lib, "libfbxsdk.lib")

namespace GFX
{
	// Enumerator for the supported mesh shading models.
	enum class SHADING_MODEL
	{
		UNKNOWN,
		PHONG,
		LAMBERT,
	};

	// Enumerator for the supported texture maps.
	enum MESH_MAT_PARAM
	{
		MMP_ALPHA_MAP				= 0x1,
		MMP_DIFFUSE_MAP				= 0x2,
		MMP_BUMP_MAP				= 0x4,
		MMP_SPECULAR_MAP			= 0x8,
		MMP_EMISSIVE_MAP			= 0x10,
		MMP_TEXTURE					= 0x1F,
	};

	// Mesh material color constant buffer.
	struct MESH_COLOR_BUFFER
	{
		Math::CColor	ambient;
		Math::CColor	diffuse;
		Math::CColor	specular;
		Math::CColor	emissive;
		float			transparency;
		int				textureFlags;
		Math::CVector2	padding0;
	};
	
	// Mesh material lighting constant buffer.
	struct MESH_LIGHTING_BUFFER
	{
		Math::CVector3	cameraPoint;
		float			padding0;
		float			specular;
		float			shininess;
		float			reflectivity;
		float			padding1;
	};

	// Enumerator for the type of imported parameters.
	enum EXPORT_MATERIAL_PARAM_TYPE
	{
		MPT_STRING,
		MPT_BOOL,
		MPT_iNTENGER,
		MPT_FLOAT,
		MPT_FLOAT2,
		MPT_FLOAT3,
		MPT_FLOAT4,
		MPT_TEXTURE2D,
		MPT_TEXTURECUBE,
		MPT_TEXTUREVOLUME,
	};
	
	// char arrat for the type of imported parameters.
	static const std::string ExportMaterialParamTypeNames[] = 
	{
		"string",
		"bool",
		"integer",
		"float",
		"float2",
		"float3",
		"float4",
		"texture2d",
		"texturecube",
		"texturevolume",
	};

	// CMeshMaterialParameter class.
	// Class that stores imported material parameters.
	class CMeshMaterialParameter
	{
	public:
		CMeshMaterialParameter()
		{
			memset(this, 0, sizeof(CMeshMaterialParameter));
		}

		char								name[128];
		EXPORT_MATERIAL_PARAM_TYPE			paramType;
		char								hint[128];
		bool								bInstanceParam;
		char								valString[256];
		float								valFloat[16];
		int									valInt;
		u32									flags;
	};

	// Type defines.
	typedef Util::CStack<CMeshMaterialParameter> MeshMaterialParamStack;

	// CMatMesh class
	// Material class primarily for imported meshes. Designed to allow for inheritance by custom mesh material classes.
	class CMatMesh : public CMaterial
	{
	public:
		friend class CFBXLoader;

	public:
		CMatMesh();
		CMatMesh(const CMatMesh& other);
		virtual ~CMatMesh();

		int									Render() override;
		void								Release() override;

		MESH_COLOR_BUFFER					m_colorBuffer;
		MESH_LIGHTING_BUFFER				m_lightingBuffer;

	protected:
		int									SetupShader() override;

	private:
		int									OpenTextureFiles();
		bool								CheckForParameter(const std::string& name);

		std::string							m_name;
		__int64								m_uniqueId;

		u32									m_textureFlags;
		bool								m_bTransparent;

		SHADING_MODEL						m_shadingModel;

		MeshMaterialParamStack				m_sMaterialParamStack;
	};
};

#endif