//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CMatPostScreen.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMatPostScreen.h"

namespace GFX
{
	CMatPostScreen::CMatPostScreen()
	{
		m_elementCount						= 2;
		m_bufferCount						= 3;
		m_textureCount						= POST_TEXTURE_COUNT;

		m_bUseInverse						= true;

		m_pWorldCamera						= NULL;
		m_pLighting							= NULL;
		m_pTextureProjection				= NULL;
	}

	CMatPostScreen::CMatPostScreen(const CMatPostScreen& other)
	{
	}

	CMatPostScreen::~CMatPostScreen()
	{
	}
	
	// Method for setting up the post screen material shader.
	int CMatPostScreen::SetupShader()
	{
		int retCode = Util::Err::NONE;
		
		if(m_pTextureProjection)
		{
			m_pTextureProjection->DirectXData(m_pDX11Data);
			if((retCode = m_pTextureProjection->Initialize()) != Util::Err::NONE)
				return retCode;
		}

		if(m_pLighting)
		{
			m_pLighting->DirectXData(m_pDX11Data);
			if((retCode = m_pLighting->Initialize()) != Util::Err::NONE)
				return retCode;
		}

		// Setup the shader input elements.
		m_pIed								= new D3D11_INPUT_ELEMENT_DESC[m_elementCount];
		
		m_pIed[0]	= { "POSITION",			0,		DXGI_FORMAT_R32G32B32_FLOAT,		0,		0,		D3D11_INPUT_PER_VERTEX_DATA,		0 };
		m_pIed[1]	= { "TEXCOORD",			0,		DXGI_FORMAT_R32G32_FLOAT,			0,		12,		D3D11_INPUT_PER_VERTEX_DATA,		0 };
		
		m_pVertexResName					= MAKEINTRESOURCE(HLSL_VPOSTSCREEN);
		m_pPixelResName						= MAKEINTRESOURCE(HLSL_PPOSTSCREEN);
		
		m_pLogFilename						= ".\\PostScreenShader.txt";
		
		if((retCode = m_pShader->InitializeConstantBuffer(1, sizeof(POSTSCREEN_CBUFFER), true)) != Util::Err::NONE)
			return retCode;

		if((retCode = m_pShader->InitializeConstantBuffer(2, sizeof(RENDERDATA::FOG_BUFFER), true)) != Util::Err::NONE)
			return retCode;

		// Setup shader post screen macros.
		m_pMacros = new D3D_SHADER_MACRO[6];

		m_fogTypeStr = std::to_string((int)m_pRenderData->fogType);

		int index = 0;
		
		m_pMacros[index++] = { "FOG_TYPE",  m_fogTypeStr.c_str() };

		if(m_pTextureProjection)
			m_pMacros[index++] = { "TEXTUREPROJ_COUNT", m_pTextureProjection->m_texProjCountStr.c_str() };
		
		if(m_pLighting)
		{
			m_pMacros[index++] = { "DIRLIGHT_COUNT", m_pLighting->m_pLightDirectional->m_dirLightCountStr.c_str() };
			m_pMacros[index++] = { "POINTLIGHT_COUNT", m_pLighting->m_pLightPoint->m_pointLightCountStr.c_str() };
			m_pMacros[index++] = { "SPOTLIGHT_COUNT", m_pLighting->m_pLightSpot->m_spotlightCountStr.c_str() };
		}

		m_pMacros[index++] = { NULL, NULL };

		// Set the post screen constance buffer's data.
		ZeroMemory(&m_postCBuffer, sizeof(m_postCBuffer));

		m_postCBuffer.scale					= Math::CVector3(1.0f, 1.0f, 1.0f);

		// Setup post processing quad camera.
		m_pCamera		= new Math::CCamera();
		
		m_pCamera->CameraMode(Math::CAMERA_MODE::Orthographic);
		m_pCamera->NearView(0.5f);
		m_pCamera->FarView(2.0f);
		m_pCamera->Width(10.0f);
		m_pCamera->Height(10.0f);
		m_pCamera->Transform(Math::CTransformSharedPtr(new Math::CTransform()));
		m_pCamera->Initialize(1.0f);
		m_pCamera->Update();

		return retCode;
	}

	// Method for setting the post screen material for render.
	int CMatPostScreen::Render()
	{
		int retCode = Util::Err::NONE;

		// Scale the post screen quad to the client bounds.
		m_postCBuffer.scale.x = m_pCamera->Width() * 0.5f * m_pCamera->Aspect();
		m_postCBuffer.scale.y = m_pCamera->Height() * 0.5f;
		
		if(m_pWorldCamera)
		{
			m_postCBuffer.cameraPoint = m_pWorldCamera->Transform()->Position();
			m_mtxInvCBuffer.projectionInv = m_pWorldCamera->ProjectionMatrixInv();
			m_mtxInvCBuffer.viewInv = m_pWorldCamera->ViewMatrixInv();
		}

		if((retCode = m_pShader->ResetConstantBuffer(1, &m_postCBuffer)) != Util::Err::NONE)
			return retCode;
			
		if(m_pRenderData->fogType != FOG_TYPE::Off)
			if((retCode = m_pShader->ResetConstantBuffer(2, &m_pRenderData->fogBuffer)) != Util::Err::NONE) return retCode;

		if((retCode = m_pShader->RenderShader(m_shaderType)) != Util::Err::NONE)
			return retCode;

		// Bind constant buffers to shader.
		if((retCode = m_pShader->BindCBuffer(0, m_shaderType)) != Util::Err::NONE) return retCode;
		if((retCode = m_pShader->BindCBuffer(1, m_shaderType)) != Util::Err::NONE) return retCode;
		
		if(m_pTextureProjection)
		{
			if((retCode = m_pTextureProjection->ResetBuffer()) != Util::Err::NONE) return retCode;
			if((retCode = m_pTextureProjection->BindBuffer()) != Util::Err::NONE) return retCode;
		}
		if(m_pLighting)
		{
			if((retCode = m_pLighting->ResetBuffer()) != Util::Err::NONE) return retCode;
			if((retCode = m_pLighting->BindBuffer()) != Util::Err::NONE) return retCode;
		}

		if(m_pRenderData->fogType != FOG_TYPE::Off)
			if((retCode = m_pShader->BindCBuffer(2, m_shaderType)) != Util::Err::NONE) return retCode;

		if((retCode = CMaterial::Render()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	void CMatPostScreen::Release()
	{
		SAFE_DELETE(m_pCamera);
		SAFE_DELETE_ARRAY(m_pMacros);

		CMaterial::Release();
	}
};