//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CMaterial.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMaterial.h"

namespace GFX
{
	CMaterial::CMaterial()
	{
		m_pIed								= NULL;
		m_elementCount						= 0;
		m_bufferCount						= 1;
		m_textureCount						= 0;

		m_pVertexResName					= NULL;
		m_pHullResName						= NULL;
		m_pDomainResName					= NULL;
		m_pPixelResName						= NULL;
		m_pGeometryResName					= NULL;
		m_pComputeResName					= NULL;

		m_pVertexEntryPoint					= NULL;
		m_pHullEntryPoint					= NULL;
		m_pDomainEntryPoint					= NULL;
		m_pPixelEntryPoint					= NULL;
		m_pGeometryEntryPoint				= NULL;
		m_pComputeEntryPoint				= NULL;

		m_bUseInverse						= false;

		m_shaderType						= SHADER_TYPE_NONE;

		m_pLogFilename						= NULL;

		m_pMacros							= NULL;

		m_pTexture							= NULL;
		m_pTextureArray						= NULL;

		m_pRenderData						= NULL;

		m_pShader							= NULL;
		m_pCamera							= NULL;
		m_pTransform						= NULL;
	}

	CMaterial::CMaterial(const CMaterial& other)
	{
	}

	CMaterial::~CMaterial()
	{
	}

	// Method for initializing a material. (focused on shader initialization primarily).
	int CMaterial::Initialize()
	{
		int retCode = Util::Err::NONE;

		// Allocate memory.
		m_pShader							= new CShader();
		
		if(!m_bUseInverse)
		{
			if(m_pCamera)
			{
				m_mtxCBuffer.view				= m_pCamera->ViewMatrix();
				m_mtxCBuffer.projection			= m_pCamera->ProjectionMatrix();
			}
			if(m_pTransform)
			{
				m_mtxCBuffer.world				= m_pTransform->WorldMatrix();
			}
		}
		else
		{
			// Only set the non-inverted values for the matrix inverse constant buffer, 
			// for if an external transformation matrices inverse is desired.
			if(m_pCamera)
			{
				m_mtxInvCBuffer.view			= m_pCamera->ViewMatrix();
				m_mtxInvCBuffer.projection		= m_pCamera->ProjectionMatrix();
			}
			if(m_pTransform)
			{
				m_mtxInvCBuffer.world			= m_pTransform->WorldMatrix();
			}
		}

		m_pShader->DirectXData(m_pDX11Data);

		if((retCode = m_pShader->InitializeConstantBufferArray(m_bufferCount)) != Util::Err::NONE)
			return retCode;
			
		if(!m_bUseInverse)
		{
			if((retCode = m_pShader->InitializeConstantBuffer(0, sizeof(MatrixCBuffer), true)) != Util::Err::NONE)
				return retCode;
		}
		else
		{
			if((retCode = m_pShader->InitializeConstantBuffer(0, sizeof(MatrixInvCBuffer), true)) != Util::Err::NONE)
				return retCode;
		}
			
		if((retCode = SetupShader()) != Util::Err::NONE)
			return retCode;

		if(m_pVertexResName)
		{
			// Create the vertex shader.
			if((retCode = m_pShader->CreateVertexShader(m_pVertexResName, m_pMacros, m_pVertexEntryPoint ? m_pVertexEntryPoint : "VShader", 
				m_pLogFilename, m_pIed, m_elementCount)) != Util::Err::NONE)
			{
				return retCode;
			}
		}

		if(m_pHullResName)
		{
			// Create the hull shader.
			if((retCode = m_pShader->CreateHullShader(m_pHullResName, m_pMacros, m_pHullEntryPoint ? m_pHullEntryPoint : "HShader", 
				m_pLogFilename)) != Util::Err::NONE)
			{
				return retCode;
			}
		}

		if(m_pDomainResName)
		{
			// Create the domain shader.
			if((retCode = m_pShader->CreateDomainShader(m_pDomainResName, m_pMacros, m_pDomainEntryPoint ? m_pDomainEntryPoint : "DShader", 
				m_pLogFilename)) != Util::Err::NONE)
			{
				return retCode;
			}
		}

		if(m_pPixelResName)
		{
			// Create the pixel shader.
			if((retCode = m_pShader->CreatePixelShader(m_pPixelResName, m_pMacros, m_pPixelEntryPoint ? m_pPixelEntryPoint : "PShader", 
				m_pLogFilename)) != Util::Err::NONE)
			{
				return retCode;
			}
		}
		
		if(m_pGeometryResName)
		{
			// Create the geometry shader.
			if((retCode = m_pShader->CreateGeometryShader(m_pGeometryResName, m_pMacros, m_pGeometryEntryPoint ? m_pGeometryEntryPoint : "GShader", 
				m_pLogFilename)) != Util::Err::NONE)
			{
				return retCode;
			}
		}

		if(m_pComputeResName)
		{
			// Create the geometry shader.
			if((retCode = m_pShader->CreateComputeShader(m_pComputeResName, m_pMacros, m_pComputeEntryPoint ? m_pComputeEntryPoint : "CShader", 
				m_pLogFilename)) != Util::Err::NONE)
			{
				return retCode;
			}
		}

		m_shaderType = m_pShader->GenerateShaderType();

		return retCode;
	}

	// Method for setting a material up for render.
	int CMaterial::Render()
	{
		int retCode = Util::Err::NONE;
		
		if(!m_bUseInverse)
		{
			if(m_pCamera)
			{
				m_mtxCBuffer.view				= m_pCamera->ViewMatrix();
				m_mtxCBuffer.projection			= m_pCamera->ProjectionMatrix();
			}
			if(m_pTransform)
			{
				m_mtxCBuffer.world				= m_pTransform->WorldMatrix();
			}
		}
		else
		{
			// Only set the non-inverted values for the matrix inverse constant buffer, 
			// for if an external transformation matrices inverse is desired.
			if(m_pCamera)
			{
				m_mtxInvCBuffer.view			= m_pCamera->ViewMatrix();
				m_mtxInvCBuffer.projection		= m_pCamera->ProjectionMatrix();
			}
			if(m_pTransform)
			{
				m_mtxInvCBuffer.world			= m_pTransform->WorldMatrix();
			}
		}
		
		if(!m_bUseInverse)
		{
			if((retCode = m_pShader->ResetConstantBuffer(0, &m_mtxCBuffer)) != Util::Err::NONE)
				return retCode;
		}
		else
		{
			if((retCode = m_pShader->ResetConstantBuffer(0, &m_mtxInvCBuffer)) != Util::Err::NONE)
				return retCode;
		}
			
		// Bind textures to the shader.
		if(m_pTextureArray)
		{
			for(u32 i = 0; i < m_textureCount; i++)
			{
				if(m_pTextureArray[i])
					m_pShader->SetPSResource(m_pTextureArray[i]->ppTexture(), i, 1);
			}
		}
		else if(m_pTexture)
		{
			m_pShader->SetPSResource(m_pTexture->ppTexture(), 0, 1);
		}

		return retCode;
	}

	void CMaterial::Release()
	{
		SAFE_DELETE_ARRAY(m_pIed);
		SAFE_RELEASE_DELETE(m_pShader);
	}
};