//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CParticleDevice.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CParticleDevice.h"

namespace GFX
{
	CParticleDevice::CParticleDevice()
	{
		m_bLoop								= false;

		m_duration							= 5.0f;
		m_playbackTimer						= 0.0f;
		m_playbackState						= PARTICLE_PLAYBACK_STATE::Play;
		m_billboardType						= PARTICLE_BILLBOARD_TYPE::PBT_All;

		m_lifetime							= 5.0f;
		m_spawnTime							= 1.0f;
		m_spawnTimer						= 0.0f;

		m_spawnRate							= 1.0f;

		m_initialScale						= Math::CVector3(1, 1, 1);
		
		m_softness							= 0.5f;
							
		m_textureCount						= 0;

		m_pParticleFunctionData				= NULL;
		m_ParticleFade						= NULL;
		m_ParticleScale						= NULL;

		m_pTimer							= NULL;

		m_bTextureMember					= false;
		m_ppTexture							= NULL;
	}

	CParticleDevice::CParticleDevice(const CParticleDevice& other)
	{
	}

	CParticleDevice::~CParticleDevice()
	{
	}

	// Method for initializing the particle device's geometry and textures.
	int CParticleDevice::Initialize()
	{
		int retCode = Util::Err::NONE;
		
		m_bEnableIndices					= false;
		m_bEnableInstancing					= true;

		m_vertexStride						= sizeof(PARTICLE_VERTEX);
		m_instanceStride					= sizeof(PARTICLE_INSTANCE);

		if(m_instanceAllocation <= 0)
			m_instanceAllocation = 100;
		m_instanceCount = 1;

		AddVertex(new PARTICLE_VERTEX(Math::CVector3(-0.5f, -0.5f, 0.0f)));
		//AddVertex(new PARTICLE_VERTEX(Math::CVector3( 0.5f, -0.5f, 0.0f)));
		//AddVertex(new PARTICLE_VERTEX(Math::CVector3( 0.5f,  0.5f, 0.0f)));
		//AddVertex(new PARTICLE_VERTEX(Math::CVector3(-0.5f,  0.5f, 0.0f)));

		m_primitiveTopology = D3D11_PRIMITIVE_TOPOLOGY_POINTLIST;

		m_pTransform = Math::CTransformSharedPtr(new Math::CTransform());
		m_pTransform->Update();

		// Check if texture should be created in object with filename.
		if(!m_ppTexture && m_textureFilename.length() > 0)
		{
			m_textureCount = 1;
			m_ppTexture = new CTexture*[m_textureCount];
			m_ppTexture[0] = new CTexture();
			m_ppTexture[0]->DirectXData(m_pDX11Data);
			if((retCode = m_ppTexture[0]->Initialize((char*)m_textureFilename.c_str())) != Util::Err::NONE)
				return retCode;

			m_bTextureMember = true;
		}
		
		if((retCode = CGeoObject::Initialize()) != Util::Err::NONE)
			return retCode;

		// Set the particle instance's to their default values.
		for(u32 i = 0; i < m_instanceAllocation; i++)
		{
			PARTICLE_INSTANCE* instance = &(((PARTICLE_INSTANCE*)m_pInstances)[i]);
			instance->color = Math::CColor(1);
			instance->data0 = Math::CVector4(0, 0, 0, 0);
			instance->position = Math::CVector3(0, 0.0f, 0);
			instance->scale = m_initialScale;
			instance->uvOffset = Math::CVector4(0, 0, 1, 1);
		}

		if(m_spawnRate > Math::_Epsilon)
			m_spawnTime = 1.0f / m_spawnRate;

		return retCode;
	}

	// Method for updating a particle device based on its playback state.
	int CParticleDevice::Update()
	{
		int retCode = Util::Err::NONE;

		switch(m_playbackState)
		{
			case PARTICLE_PLAYBACK_STATE::Play:
				if(m_spawnRate > Math::_Epsilon)
					Spawning();
				Simulate();
				break;
			case PARTICLE_PLAYBACK_STATE::Resume:
				m_playbackState = PARTICLE_PLAYBACK_STATE::Play;
				break;
			case PARTICLE_PLAYBACK_STATE::Pause:
				break;
			case PARTICLE_PLAYBACK_STATE::Stop:
				Simulate();
				break;
			case PARTICLE_PLAYBACK_STATE::Reset:
				Reset();
				m_playbackState = PARTICLE_PLAYBACK_STATE::Play;
				break;
			default:
				break;
		}
		
		if((retCode = CGeoObject::Update()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for rendering a particle device with textures.
	int CParticleDevice::Render()
	{
		int retCode = Util::Err::NONE;
		
		if(m_textureCount == 1)
			m_pMaterial->Texture(m_ppTexture[0]);
		else if(m_textureCount > 1)
			m_pMaterial->TextureArray(m_ppTexture);

		if((retCode = CGeoObject::Render()) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for simulating the particle devices logic.
	int CParticleDevice::Simulate()
	{
		int retCode = Util::Err::NONE;

		m_playbackTimer += m_pTimer->Delta();

		if(m_playbackTimer >= m_duration)
		{
			if(!m_bLoop)
			{
				m_playbackState = PARTICLE_PLAYBACK_STATE::Stop;
			}
		}
		
		for(u32 i = 0; i < m_instanceAllocation; i++)
		{
			PARTICLE_INSTANCE* instance = &(((PARTICLE_INSTANCE*)m_pInstances)[i]);

			if(instance->data0.y > 0)
			{
				float lifetime = (m_playbackTimer - instance->data0.x);
				float normalizedLifetime = 1.0f - lifetime / instance->data0.y;

				if(normalizedLifetime > 0.0f)
				{
					if(m_ParticleFade) instance->color = m_ParticleFade(m_pParticleFunctionData, normalizedLifetime);
					if(m_ParticleScale) instance->scale = m_initialScale * m_ParticleScale(m_pParticleFunctionData, normalizedLifetime);
					instance->position += m_velocity * m_pTimer->Delta();
				}
			}
		}

		return retCode;
	}

	// Method for spawning new particle in the specified way.
	void CParticleDevice::Spawning()
	{
		m_spawnTimer += m_pTimer->Delta();

		if(m_spawnTimer >= m_spawnTime)
		{
			if(m_instanceCount + 1 < m_instanceAllocation)
			{
				PARTICLE_INSTANCE* instance = &(((PARTICLE_INSTANCE*)m_pInstances)[m_instanceCount]);

				instance->data0 = Math::CVector4(m_playbackTimer, m_lifetime, 0, 0);

				m_instanceCount++;
			}
			else
			{
				m_instanceCount = 0;
			}

			m_spawnTimer = 0.0f;
		}
	}

	// Method for reseting the particle device.
	void CParticleDevice::Reset()
	{
		for(u32 i = 0; i < m_instanceAllocation; i++)
		{
			PARTICLE_INSTANCE* instance = &(((PARTICLE_INSTANCE*)m_pInstances)[i]);
			instance->color = Math::CColor(1);
			instance->data0 = Math::CVector4(0, 0, 0, 0);
			instance->position = Math::CVector3(0, 1.0f, 0);
			instance->scale = Math::CVector3(2, 2, 2);
			instance->uvOffset = Math::CVector4(0, 0, 1, 1);
		}

		m_playbackTimer = 0.0f;
		m_spawnTimer = 0.0f;
	}

	// Method for releasing the particle device.
	void CParticleDevice::Release()
	{
		if(m_bTextureMember)
		{
			for(u32 i = 0; i < m_textureCount; i++)
				SAFE_RELEASE_DELETE(m_ppTexture[i]);
			SAFE_DELETE_ARRAY(m_ppTexture);
		}

		CGeoObject::Release();
	}
};