//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CShader.h
//
//-------------------------------------------------------------------------------------------------

#include "CShader.h"

namespace GFX
{
	CShader::CShader()
	{
		m_pLayout							= NULL;
		m_pVS								= NULL;
		m_pHS								= NULL;
		m_pDS								= NULL;
		m_pPS								= NULL;
		m_pGS								= NULL;
		m_pCS								= NULL;
		m_pErrorMessage						= NULL;
		m_pCBuffer							= NULL;
		m_pBufferWidth						= NULL;
		
		m_bufferCount						= 0;

		m_shaderPath						= Util::ShaderPath();
	}

	CShader::CShader(const CShader& other)
	{
	}

	CShader::~CShader()
	{
		Release();
	}

	// Method for initializing a constant buffer array.
	int CShader::InitializeConstantBufferArray(u32 bufferCount)
	{
		int retCode = Util::Err::NONE;

		m_bufferCount		= bufferCount;
		m_pCBuffer			= new CConstantBuffer*[bufferCount];
		m_pBufferWidth		= new u32[bufferCount];

		for(u32 i = 0; i < bufferCount; i++)
			m_pCBuffer[i]= NULL;

		return retCode;
	}

	// Method for initializing a constant buffer at an index of the constant buffer array.
	int CShader::InitializeConstantBuffer(u32 index, u32 bufferWidth, bool dynamic)
	{
		int retCode = Util::Err::NONE;

		if(index >= m_bufferCount)
		{
			return Util::Err::ERR_SHADER_INVALID_CONSTANT_BUFFER_INDEX;
		}
		
		m_pBufferWidth[index] = bufferWidth;

		m_pCBuffer[index] = new CConstantBuffer();
		m_pCBuffer[index]->DirectXData(m_pDX11Data);
		if((retCode = m_pCBuffer[index]->Initialize(bufferWidth, dynamic)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for creating a vertex shader from a resource.
	int CShader::CreateVertexShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename, D3D11_INPUT_ELEMENT_DESC* ied, u32 elemCount)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* VS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);

		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "vs_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &VS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_VERTEX_SHADER;
			}
		}

		if(FAILED(m_pDev->CreateVertexShader(VS->GetBufferPointer(), VS->GetBufferSize(), NULL, &m_pVS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_VERTEX_SHADER;
		}

		if(FAILED(m_pDev->CreateInputLayout(ied, elemCount, VS->GetBufferPointer(), VS->GetBufferSize(), &m_pLayout)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_INPUT_LAYOUT;
		}

		SAFE_RELEASE(VS);

		return retCode;
	}
	
	// Method for creating a hull shader from a resource.
	int CShader::CreateHullShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* HS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);

		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "hs_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &HS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_HULL_SHADER;
			}
		}

		if(FAILED(m_pDev->CreateHullShader(HS->GetBufferPointer(), HS->GetBufferSize(), NULL, &m_pHS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_HULL_SHADER;
		}

		SAFE_RELEASE(HS);

		return retCode;
	}
	
	// Method for creating a domain shader from a resource.
	int CShader::CreateDomainShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* DS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);

		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "ds_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &DS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_DOMAIN_SHADER;
			}
		}

		if(FAILED(m_pDev->CreateDomainShader(DS->GetBufferPointer(), DS->GetBufferSize(), NULL, &m_pDS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_DOMAIN_SHADER;
		}

		SAFE_RELEASE(DS);

		return retCode;
	}
	
	// Method for creating a pixel shader from a resource.
	int CShader::CreatePixelShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* PS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);

		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "ps_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &PS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_PIXEL_SHADER;
			}
		}

		if(FAILED(m_pDev->CreatePixelShader(PS->GetBufferPointer(), PS->GetBufferSize(), NULL, &m_pPS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_PIXEL_SHADER;
		}

		SAFE_RELEASE(PS);

		return retCode;
	}
	
	// Method for creating a geometry shader from a resource.
	int CShader::CreateGeometryShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* GS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);
		
		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "gs_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &GS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_GEOMETRY_SHADER;
			}
		}

		if(FAILED(m_pDev->CreateGeometryShader(GS->GetBufferPointer(), GS->GetBufferSize(), NULL, &m_pGS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_GEOMETRY_SHADER;
		}

		SAFE_RELEASE(GS);

		return retCode;
	}

	
	// Method for creating a compute shader from a resource.
	int CShader::CreateComputeShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename)
	{
		int retCode = Util::Err::NONE;

		ID3D10Blob* CS;
		HRSRC src = FindResource(NULL, shaderResName, RT_RCDATA);
		
		if(FAILED(D3DCompile(LockResource(LoadResource(NULL, src)), SizeofResource(NULL, src), m_shaderPath.c_str(), macros, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "cs_5_0", 
			D3D10_SHADER_ENABLE_STRICTNESS, 0, &CS, &m_pErrorMessage)))
		{
			if(m_pErrorMessage && logFilename)
			{
				// Save error message log file if one is created.
				if((retCode = SaveErrorMessage(logFilename)) != Util::Err::NONE)
					return retCode;

				return Util::Err::ERR_SHADER_FAILED_TO_COMPILE_COMPUTE_SHADER;
			}
		}

		if(FAILED(m_pDev->CreateComputeShader(CS->GetBufferPointer(), CS->GetBufferSize(), NULL, &m_pCS)))
		{
			return Util::Err::ERR_SHADER_FAILED_TO_CREATE_COMPUTE_SHADER;
		}

		SAFE_RELEASE(CS);

		return retCode;
	}

	// Method of setting the shader(s) up for render.
	int CShader::RenderShader(const SHADER_TYPE& shaderType)
	{
		int retCode = Util::Err::NONE;

		m_pDevCon->IASetInputLayout(NULL);
		m_pDevCon->VSSetShader(NULL, 0, 0);
		m_pDevCon->HSSetShader(NULL, 0, 0);
		m_pDevCon->DSSetShader(NULL, 0, 0);
		m_pDevCon->PSSetShader(NULL, 0, 0);
		m_pDevCon->GSSetShader(NULL, 0, 0);
		m_pDevCon->CSSetShader(NULL, 0, 0);

		if(shaderType & SHADER_TYPE_VS) m_pDevCon->IASetInputLayout(m_pLayout);
		if(shaderType & SHADER_TYPE_VS) m_pDevCon->VSSetShader(m_pVS, 0, 0);
		if(shaderType & SHADER_TYPE_HS) m_pDevCon->HSSetShader(m_pHS, 0, 0);
		if(shaderType & SHADER_TYPE_DS) m_pDevCon->DSSetShader(m_pDS, 0, 0);
		if(shaderType & SHADER_TYPE_PS) m_pDevCon->PSSetShader(m_pPS, 0, 0);
		if(shaderType & SHADER_TYPE_GS) m_pDevCon->GSSetShader(m_pGS, 0, 0);
		if(shaderType & SHADER_TYPE_CS) m_pDevCon->CSSetShader(m_pCS, 0, 0);

		return retCode;
	}

	// Method for binding a constant buffer at an index of the constant buffer array to the shader(s).
	int CShader::BindCBuffer(u32 index, const SHADER_TYPE& shaderType)
	{
		int retCode = Util::Err::NONE;
		
		if(shaderType & SHADER_TYPE_VS) m_pCBuffer[index]->BindBufferVS(index);
		if(shaderType & SHADER_TYPE_HS) m_pCBuffer[index]->BindBufferHS(index);
		if(shaderType & SHADER_TYPE_DS) m_pCBuffer[index]->BindBufferDS(index);
		if(shaderType & SHADER_TYPE_PS) m_pCBuffer[index]->BindBufferPS(index);
		if(shaderType & SHADER_TYPE_GS) m_pCBuffer[index]->BindBufferGS(index);
		if(shaderType & SHADER_TYPE_CS) m_pCBuffer[index]->BindBufferCS(index);

		return retCode;
	}

	// Method for setting pixel shader resources.
	int CShader::SetPSResource(ID3D11ShaderResourceView** resource, u32 startSlot, u32 numViews)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->PSSetShaderResources(startSlot, numViews, resource);
		
		return retCode;
	}

	// Method for setting pixel shader sampler states.
	int CShader::SetPSSamplerState(ID3D11SamplerState** setState, u32 startSlot, u32 numViews)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->PSSetSamplers(startSlot, numViews, setState);

		return retCode;
	}
	
	// Method for setting vertex shader resources.
	int CShader::SetVSResource(ID3D11ShaderResourceView** resource, u32 startSlot, u32 numViews)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->VSSetShaderResources(startSlot, numViews, resource);
		
		return retCode;
	}
		
	// Method for resetting a constant buffer at an index in the constant buffer array.
	int CShader::ResetConstantBuffer(u32 index, void* updatedData)
	{
		int retCode = Util::Err::NONE;

		if(index >= m_bufferCount)
		{
			return Util::Err::ERR_SHADER_INVALID_CONSTANT_BUFFER_INDEX;
		}

		if((retCode = m_pCBuffer[index]->ResetBuffer(updatedData)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}
		
	// Method for updating a constant buffer at an index in the constant buffer array.
	int CShader::UpdateConstantBuffer(u32 index, void* updatedData)
	{
		int retCode = Util::Err::NONE;

		if(index >= m_bufferCount)
		{
			return Util::Err::ERR_SHADER_INVALID_CONSTANT_BUFFER_INDEX;
		}

		if((retCode = m_pCBuffer[index]->UpdateBuffer(updatedData)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for releasing all shader data.
	void CShader::Release()
	{
		SAFE_RELEASE(m_pCS);
		SAFE_RELEASE(m_pGS);
		SAFE_RELEASE(m_pPS);
		SAFE_RELEASE(m_pDS);
		SAFE_RELEASE(m_pHS);
		SAFE_RELEASE(m_pVS);

		for(u32 i = 0; i < m_bufferCount; i++)
			if(m_pCBuffer)
				SAFE_RELEASE_DELETE(m_pCBuffer[i]);

		SAFE_DELETE_ARRAY(m_pCBuffer);
		SAFE_DELETE_ARRAY(m_pBufferWidth);
	}

	// Method for saving shader compilation error to the inputted log file.
	int CShader::SaveErrorMessage(char* logFilename)
	{
		int retCode = Util::Err::NONE;
		std::ofstream file(logFilename);

		if(file.is_open())
		{
			file.write((char*)m_pErrorMessage->GetBufferPointer(), m_pErrorMessage->GetBufferSize());
			file.close();
		}
		else
		{
			return Util::Err::ERR_SHADER_FAILED_TO_SAVE_COMPILATION_ERROR;
		}

		SAFE_RELEASE(m_pErrorMessage);

		return retCode;
	}
};