//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CShader.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CSHADER_H
#define CSHADER_H

#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <fstream>
#include "..\Resources\EngineResources.h"
#include "..\Globals\CGlobals.h"
#include "..\Utilities\CDebugError.h"
#include "..\Utilities\CMemoryFree.h"
#include "..\Utilities\CFileUtil.h"
#include "..\Application\CDX11Object.h"
#include "CConstantBuffer.h"

#pragma comment(lib, "d3dcompiler.lib")

namespace GFX
{
	// Enumerator for the different types of shaders.
	enum SHADER_TYPE : u8
	{
		SHADER_TYPE_NONE	= 0x0,
		SHADER_TYPE_VS		= 0x1,
		SHADER_TYPE_HS		= 0x2,
		SHADER_TYPE_DS		= 0x4,
		SHADER_TYPE_PS		= 0x8,
		SHADER_TYPE_GS		= 0x10,
		SHADER_TYPE_CS		= 0x20,
		SHADER_TYPE_ALL		= 0x3F,
	};

	// CShader class.
	// Class for compiling and creating shaders from resources.
	// Used with the material base class 'CMaterial'.
	class CShader : public App::CDX11Object
	{
	public:
		CShader();
		CShader(const CShader& other);
		virtual ~CShader();

		int									InitializeConstantBufferArray(u32 bufferCount);
		int									InitializeConstantBuffer(u32 bufferIndex, u32 bufferWidth, bool dynamic);

		int									CreateVertexShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename, D3D11_INPUT_ELEMENT_DESC* ied, u32 elemCount);
		int									CreateHullShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename);
		int									CreateDomainShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename);
		int									CreatePixelShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename);
		int									CreateGeometryShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename);
		int									CreateComputeShader(char* shaderResName, D3D_SHADER_MACRO* macros, char* entrypoint, char* logFilename);

		int									RenderShader(const SHADER_TYPE& shaderType);
		int									BindCBuffer(u32 index, const SHADER_TYPE& shaderType);
		int									SetPSResource(ID3D11ShaderResourceView** resource, u32 startSlot, u32 numViews);
		int									SetPSSamplerState(ID3D11SamplerState** setState, u32 startSlot, u32 numViews);
		int									SetVSResource(ID3D11ShaderResourceView** resource, u32 startSlot, u32 numViews);
		int									ResetConstantBuffer(u32 index, void* updatedData);
		int									UpdateConstantBuffer(u32 index, void* updatedData);
		
		void								Release();

		inline SHADER_TYPE					GenerateShaderType();

	private:
		int									SaveErrorMessage(char* logFilename);

		std::string							m_shaderPath;

		ID3D11InputLayout*					m_pLayout;
		ID3D11VertexShader*					m_pVS;
		ID3D11HullShader*					m_pHS;
		ID3D11DomainShader*					m_pDS;
		ID3D11PixelShader*					m_pPS;
		ID3D11GeometryShader*				m_pGS;
		ID3D11ComputeShader*				m_pCS;

		ID3D10Blob*							m_pErrorMessage;
		CConstantBuffer**					m_pCBuffer;
		u32*								m_pBufferWidth;
		u32									m_bufferCount;
	};

	// Method for generating a shader type base on this shaders compiled components.
	inline SHADER_TYPE CShader::GenerateShaderType()
	{
		return SHADER_TYPE( (m_pVS ? SHADER_TYPE_VS : 0) |
							(m_pHS ? SHADER_TYPE_HS : 0) |
							(m_pDS ? SHADER_TYPE_DS : 0) |
							(m_pPS ? SHADER_TYPE_PS : 0) |
							(m_pGS ? SHADER_TYPE_GS : 0) );
	}
};

#endif