//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CTextureProjection.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CTEXTUREPROJECTION_H
#define CTEXTUREPROJECTION_H

#include "CConstantBuffer.h"
#include "..\Math\CMathVector2.h"
#include "..\Math\CMathMatrix4x4.h"
#include "..\Utilities\CDataQueue.h"
#include "..\Graphics\CTextureList.h"

namespace GFX
{
	// CTextureProjection class.
	// Class for handling the texture projectors in a scene.
	class CTextureProjection : public App::CDX11Object
	{
	public:
		friend class CMatPostScreen;

	public:
		static const u32 CBUFFER_INDEX							= 10;
		static const u32 PSSR_PROJ_TEXTURES_INDEX				= 20;
		static const u32 PSSR_PROJ_GRADIENTS_INDEX				= 21;

		// Structure for storing points to the projection camera's matrices for allow projected textures to be updated.
		struct MATRIX_POINTERS
		{
			Math::CMatrix4x4*						projection;
			Math::CMatrix4x4*						view;
			
			MATRIX_POINTERS()
			{
				projection		= NULL;
				view			= NULL;
			}

			MATRIX_POINTERS(Math::CMatrix4x4* View, Math::CMatrix4x4* Projection)
			{
				projection		= Projection;
				view			= View;
			}
		};

		// Structure for storing constant buffer data for projected textures.
		struct TEXTURE_PROJECTOR
		{
			Math::CMatrix4x4						projection;
			Math::CMatrix4x4						view;
			Math::CVector3							color;
			int										textureIndex;
			float									nearView;
			float									farView;
			float									padding[2];

			TEXTURE_PROJECTOR()
			{
			}

			TEXTURE_PROJECTOR(const Math::CMatrix4x4& View, const Math::CMatrix4x4& Projection, const Math::CVector3& Color, int TextureIndex, float NearView, float FarView)
			{
				projection		= Projection;
				view			= View;
				color			= Color;
				textureIndex	= TextureIndex;
				nearView		= NearView;
				farView			= FarView;
			}
		};

		// Structure for storing data about projected textures that get pushed into the queue for processing.
		struct TEXTURE_PROJECTOR_QUEUE
		{
			MATRIX_POINTERS matrices;
			TEXTURE_PROJECTOR texProjector;

			TEXTURE_PROJECTOR_QUEUE()
			{
			}

			TEXTURE_PROJECTOR_QUEUE(Math::CMatrix4x4* View, Math::CMatrix4x4* Projection, const Math::CVector3& Color, int TextureIndex, float NearView, float FarView)
			{
				matrices.view				= View;
				matrices.projection			= Projection;

				texProjector.projection		= *Projection;
				texProjector.view			= *View;
				texProjector.color			= Color;
				texProjector.textureIndex	= TextureIndex;
				texProjector.nearView		= NearView;
				texProjector.farView		= FarView;
			}
		};

	public:
		CTextureProjection();
		CTextureProjection(const CTextureProjection& other);
		virtual ~CTextureProjection();

		int									Initialize();

		int									ResetBuffer();
		int									BindBuffer();

		void								Release();

		inline CTextureArray*				TextureArray() const { return m_pTextureArray; }
		inline void							TextureArray(CTextureArray* pTextureArray) { m_pTextureArray = pTextureArray; }
		inline CTextureArray*				TextureArrayGradient() const { return m_pTextureArrayGradient; }
		inline void							TextureArrayGradient(CTextureArray* pTextureArray) { m_pTextureArrayGradient = pTextureArray; }

		inline void							AddTextureProjector(TEXTURE_PROJECTOR_QUEUE* texPrjQueueData);

	private:
		u32									m_texProjCount;
		std::string							m_texProjCountStr;

		CTextureArray*						m_pTextureArray;
		CTextureArray*						m_pTextureArrayGradient;

		TEXTURE_PROJECTOR*					m_pTextureProjArray;
		MATRIX_POINTERS*					m_pMatrixPtrArray;

		Util::CPtrQueue						m_qTextureProjQueue;

		CConstantBuffer*					m_pTextureProjCBuffer;
	};

	//
	// Texture projector inline methods.
	//
	
	inline void CTextureProjection::AddTextureProjector(TEXTURE_PROJECTOR_QUEUE* texPrjQueueData)
	{
		m_qTextureProjQueue.Push(texPrjQueueData);
		m_texProjCount++;
	}
};

#endif