//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CTextureRenderTo.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CTextureRenderTo.h"

namespace GFX
{
	CTextureRenderTo::CTextureRenderTo()
	{
		m_textureWidth						= 0;
		m_textureHeight						= 0;

		m_textureCount						= 0;

		m_pRenderTargetTexture				= NULL;
		m_pRenderTargetView					= NULL;
		m_pShaderResourceView				= NULL;
		
		m_pDepthStencilBuffer				= NULL;
		m_pDepthStencilState				= NULL;
		m_pDepthStencilView					= NULL;
		m_pDepthStencilResView				= NULL;
	}

	CTextureRenderTo::CTextureRenderTo(const CTextureRenderTo& other)
	{
	}

	CTextureRenderTo::~CTextureRenderTo()
	{
	}

	// Creates a new render to texture, texture.
	int CTextureRenderTo::Initialize(int textureWidth, int textureHeight, u32 textureCount, bool noStencil)
	{
		int retCode = Util::Err::NONE;
		
		m_textureCount						= textureCount;

		m_textureWidth						= textureWidth;
		m_textureHeight						= textureHeight;
		
		if((retCode = CreateRenderTargetView()) != Util::Err::NONE)
			return retCode;

		if((retCode = CreateDepthStencilView(noStencil)) != Util::Err::NONE)
			return retCode;

		// Create the render to texture viewport.
		m_viewport.Width = (float)textureWidth;
		m_viewport.Height = (float)textureHeight;
		m_viewport.MinDepth = 0.0f;
		m_viewport.MaxDepth = 1.0f;
		m_viewport.TopLeftX = 0.0f;
		m_viewport.TopLeftY = 0.0f;

		return retCode;
	}
	
	// Method for creating the depth stencil texture, view, and resource.
	int CTextureRenderTo::CreateRenderTargetView()
	{
		int retCode = Util::Err::NONE;

		D3D11_TEXTURE2D_DESC				td;
		D3D11_RENDER_TARGET_VIEW_DESC		rtvd;
		D3D11_SHADER_RESOURCE_VIEW_DESC		srvd;
		
		u32 i;

		// Allocate memory to array(s).
		m_pRenderTargetView					= new ID3D11RenderTargetView*[m_textureCount];

		ZeroMemory(&td, sizeof(td));

		// Create the 2d texture(s).
		td.Width							= m_textureWidth;
		td.Height							= m_textureHeight;
		td.MipLevels						= 1;
		td.ArraySize						= m_textureCount;
		td.Format							= DXGI_FORMAT_R32G32B32A32_FLOAT;
		td.SampleDesc.Count					= 1;
		td.SampleDesc.Quality				= 0;
		td.Usage							= D3D11_USAGE_DEFAULT;
		td.BindFlags						= D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;

		if(FAILED(m_pDev->CreateTexture2D(&td, NULL, &m_pRenderTargetTexture)))
		{
			return Util::Err::ERR_TEXTURE_RENDER_TO_FAILED_TO_CREATE_TEXTURE2D;
		}

		ZeroMemory(&rtvd, sizeof(rtvd));
		
		// Create the render target(s).
		rtvd.Format							= td.Format;
		rtvd.ViewDimension					= D3D11_RTV_DIMENSION_TEXTURE2DARRAY;
		rtvd.Texture2DArray.MipSlice		= 0;
		rtvd.Texture2DArray.ArraySize		= 1;
		
		for(i = 0; i < m_textureCount; i++)
		{
			rtvd.Texture2DArray.FirstArraySlice	= i;
			if(FAILED(m_pDev->CreateRenderTargetView(m_pRenderTargetTexture, &rtvd, &m_pRenderTargetView[i])))
			{
				return Util::Err::ERR_TEXTURE_RENDER_TO_FAILED_TO_CREATE_TARGET;
			}
		}

		ZeroMemory(&srvd, sizeof(srvd));
		
		// Create the shader resource view(s).
		srvd.Format							= td.Format;
		srvd.ViewDimension					= D3D11_SRV_DIMENSION_TEXTURE2DARRAY;
		srvd.Texture2DArray.MostDetailedMip	= 0;
		srvd.Texture2DArray.MipLevels		= 1;
		srvd.Texture2DArray.ArraySize		= m_textureCount;
		srvd.Texture2DArray.FirstArraySlice	= 0;
		
		if(FAILED(m_pDev->CreateShaderResourceView(m_pRenderTargetTexture, &srvd, &m_pShaderResourceView)))
		{
			return Util::Err::ERR_TEXTURE_RENDER_TO_FAILED_TO_CREATE_RESOURCE_VIEW;
		}

		return retCode;
	}

	// Method for creating the depth stencil texture, state, view, and resource.
	int CTextureRenderTo::CreateDepthStencilView(bool noStencil)
	{
		int retCode = Util::Err::NONE;
		
		D3D11_TEXTURE2D_DESC				depthBufferDesc;
		D3D11_DEPTH_STENCIL_DESC			depthStencelDesc;
		D3D11_DEPTH_STENCIL_VIEW_DESC		depthStencilViewDesc;
		D3D11_SHADER_RESOURCE_VIEW_DESC		depthStencilResourceViewDesc;

		ZeroMemory(&depthBufferDesc, sizeof(depthBufferDesc));

		depthBufferDesc.Width				= m_textureWidth;
		depthBufferDesc.Height				= m_textureHeight;
		depthBufferDesc.MipLevels			= 1;
		depthBufferDesc.ArraySize			= 1;
		depthBufferDesc.Format				= noStencil ? DXGI_FORMAT_R32_TYPELESS : DXGI_FORMAT_R24G8_TYPELESS;
		depthBufferDesc.SampleDesc.Count	= 1;
		depthBufferDesc.SampleDesc.Quality	= 0;
		depthBufferDesc.Usage				= D3D11_USAGE_DEFAULT;
		depthBufferDesc.BindFlags			= D3D11_BIND_DEPTH_STENCIL | D3D11_BIND_SHADER_RESOURCE;
		depthBufferDesc.CPUAccessFlags		= 0;
		depthBufferDesc.MiscFlags			= 0;

		if(FAILED(m_pDev->CreateTexture2D(&depthBufferDesc, NULL, &m_pDepthStencilBuffer)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_DEPTH_STENCIL_BUFFER;
		}

		ZeroMemory(&depthStencelDesc, sizeof(depthStencelDesc));

		depthStencelDesc.DepthEnable					= true;
		depthStencelDesc.DepthWriteMask					= D3D11_DEPTH_WRITE_MASK_ALL;
		depthStencelDesc.DepthFunc						= D3D11_COMPARISON_LESS;

		depthStencelDesc.StencilEnable					= noStencil ? false : true;
		depthStencelDesc.StencilWriteMask				= 0XFF;
		depthStencelDesc.StencilReadMask				= 0XFF;

		depthStencelDesc.FrontFace.StencilFailOp		= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.FrontFace.StencilDepthFailOp	= D3D11_STENCIL_OP_INCR;
		depthStencelDesc.FrontFace.StencilPassOp		= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.FrontFace.StencilFunc			= D3D11_COMPARISON_ALWAYS;

		depthStencelDesc.BackFace.StencilFailOp			= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.BackFace.StencilDepthFailOp	= D3D11_STENCIL_OP_DECR;
		depthStencelDesc.BackFace.StencilPassOp			= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.BackFace.StencilFunc			= D3D11_COMPARISON_ALWAYS;

		if(FAILED(m_pDev->CreateDepthStencilState(&depthStencelDesc, &m_pDepthStencilState)))
		{
			return Util::Err::ERR_DX11_CREATE_DEPTH_STENCIL_STATE;
		}

		ZeroMemory(&depthStencilViewDesc, sizeof(depthStencilViewDesc));

		depthStencilViewDesc.Format					= noStencil ? DXGI_FORMAT_D32_FLOAT : DXGI_FORMAT_D24_UNORM_S8_UINT;
		depthStencilViewDesc.ViewDimension			= D3D11_DSV_DIMENSION_TEXTURE2D;
		depthStencilViewDesc.Texture2D.MipSlice		= 0;
		
		if(FAILED(m_pDev->CreateDepthStencilView(m_pDepthStencilBuffer, &depthStencilViewDesc, &m_pDepthStencilView)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_DEPTH_STENCIL_VIEW;
		}

		// Depth stencil resource view.
		ZeroMemory(&depthStencilResourceViewDesc, sizeof(depthStencilResourceViewDesc));
		
		// Setup the description of the shader resource view.
		depthStencilResourceViewDesc.Format							= noStencil ? DXGI_FORMAT_R32_FLOAT : DXGI_FORMAT_R24_UNORM_X8_TYPELESS;
		depthStencilResourceViewDesc.ViewDimension					= D3D11_SRV_DIMENSION_TEXTURE2D;
		depthStencilResourceViewDesc.Texture2D.MostDetailedMip		= 0;
		depthStencilResourceViewDesc.Texture2D.MipLevels			= 1;

		// Create the shader resource view.
		if(FAILED(m_pDev->CreateShaderResourceView(m_pDepthStencilBuffer, &depthStencilResourceViewDesc, &m_pDepthStencilResView)))
		{
			return Util::Err::ERR_DX11_CREATE_DEPTH_STENCIL_RESOURCE_VIEW;
		}

		return retCode;
	}

	// Method for settings the render target to the render to texture target(s).
	int CTextureRenderTo::SetAsRenderTarget(bool setDepthBuffer)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->OMSetDepthStencilState(m_pDepthStencilState, 1);
		m_pDevCon->OMSetRenderTargets(m_textureCount, m_pRenderTargetView, setDepthBuffer ? m_pDepthStencilView : NULL);
		m_pDevCon->RSSetViewports(1, &m_viewport);

		return retCode;
	}

	// Method for settings the render target to the render to texture target(s).
	int CTextureRenderTo::SetAsRenderTarget(bool setDepthBuffer, u32 index, u32 range)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->OMSetDepthStencilState(m_pDepthStencilState, 1);
		m_pDevCon->OMSetRenderTargets(range, &m_pRenderTargetView[index], setDepthBuffer ? m_pDepthStencilView : NULL);
		m_pDevCon->RSSetViewports(1, &m_viewport);

		return retCode;
	}

	// Method for resetting the render target back to the default target.
	int CTextureRenderTo::ResetRenderTarget(bool setDepthBuffer)
	{
		int retCode = Util::Err::NONE;
		
		m_pDevCon->OMSetDepthStencilState(m_pDX11Data->pDepthStencilState, 1);
		m_pDevCon->OMSetRenderTargets(1, &m_pDX11Data->pRenderTargetView, setDepthBuffer ? m_pDX11Data->pDepthStencilView : NULL);
		m_pDX11Data->pDevCon->RSSetViewports(1, &m_pDX11Data->viewport);
		
		return retCode;
	}

	// Method for clearing render to texture. render target(s).
	int CTextureRenderTo::ClearRenderTarget(const DX11_CLEAR_MODE& clearMode, const Math::CColor&  bgColor, bool setDepthBuffer)
	{
		int retCode = Util::Err::NONE;

		if(clearMode & DX11_CLEAR_COLOR)
		{
			for(u32 i = 0; i < m_textureCount; i++)
			{
				m_pDevCon->ClearRenderTargetView(m_pRenderTargetView[i], bgColor);
			}
		}

		if(clearMode & DX11_CLEAR_DEPTH)
			m_pDevCon->ClearDepthStencilView(setDepthBuffer ? m_pDepthStencilView : NULL, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);

		return retCode;
	}

	// Method for clearing render to texture. render target(s).
	int CTextureRenderTo::ClearRenderTarget(const DX11_CLEAR_MODE& clearMode, const Math::CColor&  bgColor, bool setDepthBuffer, u32 index, u32 range)
	{
		int retCode = Util::Err::NONE;

		if(clearMode & DX11_CLEAR_COLOR)
		{
			for(u32 i = 0; i < range; i++)
				m_pDevCon->ClearRenderTargetView(m_pRenderTargetView[index + i], bgColor);
		}

		if(clearMode & DX11_CLEAR_DEPTH)
			m_pDevCon->ClearDepthStencilView(setDepthBuffer ? m_pDepthStencilView : NULL, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);

		return retCode;
	}

	// Method for clearing the default render target.
	int CTextureRenderTo::ClearBackBuffer(const DX11_CLEAR_MODE& clearMode, const Math::CColor&  bgColor, bool setDepthBuffer)
	{
		int retCode = Util::Err::NONE;
		
		if(clearMode & DX11_CLEAR_COLOR)
			m_pDevCon->ClearRenderTargetView(m_pDX11Data->pRenderTargetView, bgColor);

		if(clearMode & DX11_CLEAR_DEPTH)
			m_pDevCon->ClearDepthStencilView(setDepthBuffer ? m_pDX11Data->pDepthStencilView : NULL, D3D11_CLEAR_DEPTH, 1.0f, 0);

		return retCode;
	}

	int CTextureRenderTo::PreResize()
	{
		int retCode = Util::Err::NONE;

		Release();

		return retCode;
	}

	int CTextureRenderTo::PostResize(int textureWidth, int textureHeight)
	{
		int retCode = Util::Err::NONE;

		if((retCode = Initialize(textureWidth, textureHeight, m_textureCount)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for releasing all render to texture data.
	void CTextureRenderTo::Release()
	{
		SAFE_RELEASE(m_pDepthStencilResView);
		SAFE_RELEASE(m_pDepthStencilView);
		SAFE_RELEASE(m_pDepthStencilState);
		SAFE_RELEASE(m_pDepthStencilBuffer);

		if(m_pShaderResourceView)
		{
			SAFE_RELEASE(m_pShaderResourceView);

			for(u32 i = 0; i < m_textureCount; i++)
				SAFE_RELEASE(m_pRenderTargetView[i]);

			SAFE_RELEASE(m_pRenderTargetTexture);
		}

		SAFE_DELETE_ARRAY(m_pRenderTargetView);
	}
};