//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathCamera.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHCAMERA_H
#define CMATHCAMERA_H

#include "CMathTransform.h"
#include "CMathVector2.h"
#include "CMathFrustum.h"

namespace Math
{
	// Enumerator for the different camera modes.
	enum class CAMERA_MODE
	{
		Perspective,
		Orthographic,
	};

	// CCamera class
	// Create for create a camera with either perspective, or orthographic projection.
	class CCamera
	{
	public:
		CCamera();
		CCamera(const CCamera& other);
		virtual ~CCamera();

		void								Initialize();
		void								Initialize(float screenWidth, float screenHeight);
		void								Initialize(float aspect);
		void								Update(bool transform = true);
		
		Math::CVector2						WorldToScreen(const Math::CVector2& point);
		Math::CVector2						ScreenToWorld(const Math::CVector2& point);

		inline float						NearView() const;
		inline void							NearView(float nv);

		inline float						FarView() const;
		inline void							FarView(float fv);

		inline float						FieldOfView() const;
		inline void							FieldOfView(float fov);

		inline float						Aspect() const;

		inline float						Width() const;
		inline void							Width(float width);

		inline float						Height() const;
		inline void							Height(float height);

		inline float						ScreenWidth() const;
		inline void							ScreenWidth(float screenWidth);

		inline float						ScreenHeight() const;
		inline void							ScreenHeight(float screenHeight);

		inline CAMERA_MODE					CameraMode() const;
		inline void							CameraMode(const CAMERA_MODE& mode);
		
		inline CMatrix4x4					ViewMatrix() const;
		inline CMatrix4x4					ViewMatrixInv() const;
		inline CMatrix4x4					ProjectionMatrix() const;
		inline CMatrix4x4					ProjectionMatrixInv() const;

		inline CMatrix4x4*					ViewMatrixPtr();
		inline CMatrix4x4*					ViewMatrixInvPtr();
		inline CMatrix4x4*					ProjectionMatrixPtr();
		inline CMatrix4x4*					ProjectionMatrixInvPtr();
		
		inline Math::CTransformSharedPtr	Transform() const;
		inline void							Transform(Math::CTransformSharedPtr pTransform);

		inline CFrustum						Frustum() const;

	private:
		float								m_nearView;
		float								m_farView;

		float								m_fov;

		float								m_aspect;
		float								m_width;
		float								m_height;

		float								m_screenWidth;
		float								m_screenHeight;

		CAMERA_MODE							m_cameraMode;
		
		CMatrix4x4							m_viewMtx;
		CMatrix4x4							m_viewMtxInv;
		CMatrix4x4							m_projMtx;
		CMatrix4x4							m_projMtxInv;
		
		CTransformSharedPtr					m_pTransform;
		CFrustum							m_frustum;
	};

	// Near view getter and setter.

	inline float CCamera::NearView() const
	{
		return m_nearView;
	}

	inline void CCamera::NearView(float nv)
	{
		m_nearView = nv;
	}
	
	// Far view getter and setter.

	inline float CCamera::FarView() const
	{
		return m_farView;
	}

	inline void CCamera::FarView(float fv)
	{
		m_farView = fv;
	}

	// Field of view getter and setter.

	inline float CCamera::FieldOfView() const
	{
		return m_fov;
	}

	inline void CCamera::FieldOfView(float fov)
	{
		m_fov = fov;
	}

	// Aspect ratio getter.
	inline float CCamera::Aspect() const
	{
		return m_aspect;
	}

	// Camera width getter and setter.

	inline float CCamera::Width() const
	{
		return m_width;
	}

	inline void CCamera::Width(float width)
	{
		m_width = width;
	}

	// Camera height getter and setter.

	inline float CCamera::Height() const
	{
		return m_height;
	}

	inline void CCamera::Height(float height)
	{
		m_height = height;
	}

	// Screen width getter and setter.

	inline float CCamera::ScreenWidth() const
	{
		return m_screenWidth;
	}

	inline void CCamera::ScreenWidth(float screenWidth)
	{
		m_screenWidth = screenWidth;
	}

	// Screen height getter and setter.

	inline float CCamera::ScreenHeight() const
	{
		return m_screenHeight;
	}

	inline void CCamera::ScreenHeight(float screenHeight)
	{
		m_screenHeight = screenHeight;
	}

	// Camera mode getter and setter.

	inline CAMERA_MODE CCamera::CameraMode() const
	{
		return m_cameraMode;
	}

	inline void CCamera::CameraMode(const CAMERA_MODE& mode)
	{
		m_cameraMode = mode;
	}

	// View matrix getters.
		
	inline CMatrix4x4 CCamera::ViewMatrix() const
	{
		return m_viewMtx;
	}

	inline CMatrix4x4 CCamera::ViewMatrixInv() const
	{
		return m_viewMtxInv;
	}

	// Projection matrix getters.

	inline CMatrix4x4 CCamera::ProjectionMatrix() const
	{
		return m_projMtx;
	}

	inline CMatrix4x4 CCamera::ProjectionMatrixInv() const
	{
		return m_projMtxInv;
	}

	// View matrix pointer getters.

	inline CMatrix4x4* CCamera::ViewMatrixPtr()
	{
		return &m_viewMtx;
	}

	inline CMatrix4x4* CCamera::ViewMatrixInvPtr()
	{
		return &m_viewMtxInv;
	}

	// Projection matrix pointer getters.

	inline CMatrix4x4* CCamera::ProjectionMatrixPtr()
	{
		return &m_projMtx;
	}

	inline CMatrix4x4* CCamera::ProjectionMatrixInvPtr()
	{
		return &m_projMtxInv;
	}

	// Transform getter and setter.
		
	inline Math::CTransformSharedPtr CCamera::Transform() const
	{
		return m_pTransform;
	}

	inline void CCamera::Transform(Math::CTransformSharedPtr pTransform)
	{
		m_pTransform = pTransform;
	}

	// Frustum getter.
	inline CFrustum CCamera::Frustum() const
	{
		return m_frustum;
	}
};

#endif