//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathColor.h
//
//-------------------------------------------------------------------------------------------------

#include "CMathColor.h"

namespace Math
{
	CColor::CColor()
	{
		r = 0;
		g = 0;
		b = 0;
		a = 0;
	}

	CColor::CColor(float R, float G, float B, float A)
	{
		r = R;
		g = G;
		b = B;
		a = A;
	}

	CColor::CColor(float R, float G, float B)
	{
		r = R;
		g = G;
		b = B;
		a = 1.0f;
	}

	CColor::CColor(float f)
	{
		*this = CColor(f, f, f);
	}

	CColor::CColor(const CColor& other)
	{
		r = other.r;
		g = other.g;
		b = other.b;
		a = other.a;
	}

	CColor::CColor(CColor&& other)
	{
		r = other.r; other.r = 0.0f;
		g = other.g; other.g = 0.0f;
		b = other.b; other.b = 0.0f;
		a = other.a; other.a = 0.0f;
	}

	CColor& CColor::operator = (float f)
	{
		r = f;
		g = f;
		b = f;
		a = f;
		return *this;
	}

	CColor& CColor::operator = (const CColor& c)
	{
		r = c.r;
		g = c.g;
		b = c.b;
		a = c.a;
		return *this;
	}
	
	//
	// Addition operators.
	//
	
	CColor CColor::operator + (const CColor& c) const
	{
		return CColor(a + c.a, g + c.g, b + c.b, a + c.a);
	}

	CColor CColor::operator + (float f) const
	{
		return CColor(a + f, g + f, b + f, a + f);
	}

	CColor& CColor::operator += (const CColor& c)
	{
		*this = *this + c;
		return *this;
	}

	CColor& CColor::operator += (float f)
	{
		*this = *this + f;
		return *this;
	}
	
	//
	// Subtraction operators.
	//
	
	CColor CColor::operator - (const CColor& c) const
	{
		return CColor(a - c.a, g - c.g, b - c.b, a - c.a);
	}

	CColor CColor::operator - (float f) const
	{
		return CColor(a - f, g - f, b - f, a - f);
	}

	CColor& CColor::operator -= (const CColor& c)
	{
		*this = *this - c;
		return *this;
	}

	CColor& CColor::operator -= (float f)
	{
		*this = *this - f;
		return *this;
	}
	
	//
	// Multiplication operators.
	//

	CColor CColor::operator * (const CColor& c) const
	{
		return CColor(a * c.a, g * c.g, b * c.b, a * c.a);
	}

	CColor CColor::operator * (float f) const
	{
		return CColor(a * f, g * f, b * f, a * f);
	}

	CColor& CColor::operator *= (const CColor& c)
	{
		*this = *this * c;
		return *this;
	}

	CColor& CColor::operator *= (float f)
	{
		*this = *this * f;
		return *this;
	}
	
	//
	// Division operators.
	//

	CColor CColor::operator / (const CColor& c) const
	{
		return CColor(a / c.a, g / c.g, b / c.b, a / c.a);
	}

	CColor CColor::operator / (float f) const
	{
		return CColor(a / f, g / f, b / f, a / f);
	}

	CColor& CColor::operator /= (const CColor& c)
	{
		*this = *this / c;
		return *this;
	}

	CColor& CColor::operator /= (float f)
	{
		*this = *this / f;
		return *this;
	}
	
	//
	// Conditional operators.
	//

	bool CColor::operator == (const CColor& c) const
	{
		return (r == c.r && g == c.g && b == c.b && a == c.a);
	}

	bool CColor::operator != (const CColor& c) const
	{
		return (r != c.r || g != c.g || b != c.b || a != c.a);
	}
	
	//
	// Friend operators.
	//

	CColor operator + (float f, const CColor& c)
	{
		return c + f;
	}

	CColor operator - (float f, const CColor& c)
	{
		return c - f;
	}

	CColor operator * (float f, const CColor& c)
	{
		return c * f;
	}

	CColor operator / (float f, const CColor& c)
	{
		return c / f;
	}

	//
	// Color interpolation methods.
	//
	
	CColor MoveTowards(const CColor& from, const CColor& to, float t)
	{
		return CColor(MoveTowards(from.x, to.x, t), MoveTowards(from.y, to.y, t), MoveTowards(from.z, to.z, t), MoveTowards(from.w, to.w, t));
	}
	
	CColor MoveTowards(const CColor& from, const CColor& to, const CVector4& t)
	{
		return CColor(MoveTowards(from.x, to.x, t.x), MoveTowards(from.y, to.y, t.y), MoveTowards(from.z, to.z, t.z), MoveTowards(from.w, to.w, t.w));
	}

	CColor Lerp(const CColor& from, const CColor& to, float t)
	{
		return CColor(Lerp(from.x, to.x, t), Lerp(from.y, to.y, t), Lerp(from.z, to.z, t), Lerp(from.w, to.w, t));
	}

	CColor Lerp(const CColor& from, const CColor& to, const CVector4& t)
	{
		return CColor(Lerp(from.x, to.x, t.x), Lerp(from.y, to.y, t.y), Lerp(from.z, to.z, t.z), Lerp(from.w, to.w, t.w));
	}

	//
	// Color static methods.
	//
	
	void CColor::Saturate(CColor& c)
	{
		c = CColor(Math::Saturate(c.r), Math::Saturate(c.g), Math::Saturate(c.b), Math::Saturate(c.a));
	}

	CColor CColor::Saturated(const CColor& c)
	{
		return CColor(Math::Saturate(c.r), Math::Saturate(c.g), Math::Saturate(c.b), Math::Saturate(c.a));
	}

	float CColor::Grayscale(const CColor& c)
	{
		return (0.299f * c.r + 0.587f * c.g + 0.114f * c.b);
	}
};