//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathMatrix4x4.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHMATRIX4X4_H
#define CMATHMATRIX4X4_H

#include <cmath>
#include <string>
#include "..\Globals\CGlobals.h"
#include "CMathVector3.h"

namespace Math
{
	// CMatrix4x4 class
	// Class for creating a 4x4 matrix that includes varies methods for manipulating a matrix.
	class CMatrix4x4
	{
	public:
		union
		{
			struct
			{
				float _11, _12, _13, _14;
				float _21, _22, _23, _24;
				float _31, _32, _33, _34;
				float _41, _42, _43, _44;
			};

			struct
			{
				float _ij[4][4];
			};

			struct
			{
				float m[16];
			};
		};

		CMatrix4x4();
		CMatrix4x4(const float* M);
		CMatrix4x4(float m_11, float m_12, float m_13, float m_14,
				   float m_21, float m_22, float m_23, float m_24,
				   float m_31, float m_32, float m_33, float m_34,
				   float m_41, float m_42, float m_43, float m_44);
				  
		CMatrix4x4(const CMatrix4x4& other);
		CMatrix4x4(CMatrix4x4&& other);

		float&								operator ()	(u32 i, u32 j);
		float								operator ()	(u32 i, u32 j) const;

		CMatrix4x4&							operator =	(float f);
		CMatrix4x4&							operator =	(const CMatrix4x4& M);
		CMatrix4x4							operator -	() const;

		CMatrix4x4							operator +	(const CMatrix4x4& M) const;
		CMatrix4x4							operator -	(const CMatrix4x4& M) const;
		CMatrix4x4							operator *	(const CMatrix4x4& M) const;
		CMatrix4x4							operator *	(float f) const;
		CMatrix4x4							operator /	(float f) const;

		CMatrix4x4&							operator +=	(const CMatrix4x4& M);
		CMatrix4x4&							operator -=	(const CMatrix4x4& M);
		CMatrix4x4&							operator *=	(const CMatrix4x4& M);
		CMatrix4x4&							operator *=	(float f);
		CMatrix4x4&							operator /=	(float f);

		bool								operator ==	(const CMatrix4x4& M) const;
		bool								operator !=	(const CMatrix4x4& M) const;
		
		friend CMatrix4x4					operator *	(float f, const CMatrix4x4& m);
		friend CMatrix4x4					operator /	(float f, const CMatrix4x4& m);

		CMatrix4x4							Transpose();
		float								Determinant();
		CMatrix4x4							Inverse();

		static CMatrix4x4					Identity();

		static CMatrix4x4					Translate(const CVector3& position);
		static CMatrix4x4					Scale(const CVector3& scale);
		static CMatrix4x4					RotateX(float angle);
		static CMatrix4x4					RotateY(float angle);
		static CMatrix4x4					RotateZ(float angle);
		
		CVector3							TransformCoord(const CVector3& point);
		static CVector3						TransformCoord(const CVector3& point, const CMatrix4x4& M);

		static CMatrix4x4					LookAtLH(const CVector3& eye, const CVector3& at, const CVector3& up);
		static CMatrix4x4					LookAtRH(const CVector3& eye, const CVector3& at, const CVector3& up);
		
		static CMatrix4x4					PerspectiveFovLH(float fov, float aspect, float nv, float fv);
		static CMatrix4x4					PerspectiveFovRH(float fov, float aspect, float nv, float fv);

		static CMatrix4x4					OrthographicLH(float width, float height, float nv, float fv);
		static CMatrix4x4					OrthographicRH(float width, float height, float nv, float fv);

		static CMatrix4x4					Transpose(const CMatrix4x4& M);
		static float						Determinant(const CMatrix4x4& M);
		static CMatrix4x4					Inverse(const CMatrix4x4& M);
	};

	// CMatrix4x4 constants.
	const CMatrix4x4 MTX4X4_IDENTITY = CMatrix4x4(
		1.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 1.0f
	);
};

#endif