//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathPlane.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathPlane.h"

namespace Math
{
	CPlane::CPlane()
	{
		a = 0.0f;
		b = 0.0f;
		c = 0.0f;
		d = 0.0f;
	}

	CPlane::CPlane(float A, float B, float C, float D)
	{
		a = A;
		b = B;
		c = C;
		d = D;
	}

	CPlane::CPlane(const CVector3& point, const CVector3& normal)
	{
		this->FromPointNormal(point, normal);
	}

	CPlane::CPlane(const CVector3& a0, const CVector3& b0, const CVector3& c0)
	{
		this->FromPoints(a0, b0, c0);
	}

	CPlane::CPlane(const CPlane& other)
	{
		a = other.a;
		b = other.b;
		c = other.c;
		d = other.d;
	}

	CPlane::CPlane(CPlane&& other)
	{
		a = other.a; other.a = 0.0f;
		b = other.b; other.b = 0.0f;
		c = other.c; other.c = 0.0f;
		d = other.d; other.d = 0.0f;
	}
	
	//
	// Conditional operators.
	//

	bool CPlane::operator == (const CPlane& p) const
	{
		return a == p.a && b == p.b && c == p.c && d == p.d;
	}

	bool CPlane::operator != (const CPlane& p) const
	{
		return a != p.a || b != p.b || c != p.c || d != p.d;
	}

	//
	// General methods.
	//
		
	// Method for creating this plane from a point and a normal vector.
	void CPlane::FromPointNormal(const CVector3& point, const CVector3& normal)
	{
		a = normal.x * point.x;
		b = normal.y * point.y;
		c = normal.z * point.z;
		d = (a + b + c);
	}

	// Method for creating this plane from three points.
	void CPlane::FromPoints(const CVector3& a0, const CVector3& b0, const CVector3& c0)
	{
		CVector3 ab = b0 - a0;
		CVector3 ac = c0 - a0;
		CVector3 abXac = CVector3::Cross(ab, ac);

		float a = abXac.x;
		float b = abXac.y;
		float c = abXac.z;
		float d = abXac.x * a0.x + abXac.y * a0.y + abXac.z * a0.z;
	}

	// Method for normalizing this plane.
	void CPlane::Normalize()
	{
		float dist = sqrtf(a * a + b * b + c * c);

		a /= dist;
		b /= dist;
		c /= dist;
		d /= dist;
	}

	// Method for returning a normalized version of this plane.
	CPlane CPlane::Normalized()
	{
		CPlane result;
		float dist = sqrtf(a * a + b * b + c * c);

		result.a = a / dist;
		result.b = b / dist;
		result.c = c / dist;
		result.d = d / dist;

		return result;
	}

	//
	// Intersection methods.
	//

	bool CPlane::IntersectLine(const CVector3& v0, const CVector3& v1, CVector3& closestPoint)
	{
		CVector3 diff = v0 - v1;
		float denominator = a * diff.x + b * diff.y + c * diff.z;
		
		if(denominator == 0.0f)
		{
			return false;
		}

		float t = (a * v1.x + b * v1.y + c * v1.z + d) / denominator;
		closestPoint = v0 + (v1 - v0) * t;

		return true;
	}

	bool CPlane::IntersectPlane(CPlane p, CVector3& v0, CVector3& v1)
	{
		v0 = CVector3::Zero();
		
		float denominator = a * p.b - b * p.a;

		if(denominator != 0)
		{
			// Solve with z.
			v0.x = (b * p.d - d * p.b) / denominator;
			v0.y = (a * p.d - d * p.a) / denominator;
		}
		else
		{
			denominator = c * p.a - a * p.c;

			if(denominator != 0)
			{
				// Solve with y.
				v0.x = (c * p.d - d * p.c) / denominator;
				v0.z = (a * p.d - d * p.a) / denominator;
			}
			else
			{
				denominator = b * p.c - c * p.b;
				
				if(denominator != 0)
				{
					// Solve with x.
					v0.y = (c * p.d - d * p.c) / denominator;
					v0.z = (b * p.d - d * p.b) / denominator;
				}
				else
					return false;
			}
		}

		v1 = Math::CVector3::Cross(Normal(), p.Normal()).Normalized();
		if(v1.Length() <= Math::_Epsilon)
			return false;

		v1.Normalize();

		return true;
	}

	//
	// Plane static mathematic methods.
	//
	
	float CPlane::Dot(const CPlane& p, const CVector4& v)
	{
		return p.a * v.x + p.b * v.y + p.c * v.z + p.d * v.w;
	}

	float CPlane::DotCoord(const CPlane& p, const CVector3& v)
	{
		return p.a * v.x + p.b * v.y + p.c * v.z + p.d;
	}

	float CPlane::DotNormal(const CPlane& p, const CVector3& v)
	{
		return p.a * v.x + p.b * v.y + p.c * v.z;
	}
		
	void CPlane::FromPointNormal(CPlane& p, const CVector3& point, const CVector3& normal)
	{
		p.a = normal.x * point.x;
		p.b = normal.y * point.y;
		p.c = normal.z * point.z;
		p.d = (p.a + p.b + p.c);
	}

	void CPlane::FromPoints(CPlane& p, const CVector3& a0, const CVector3& b0, const CVector3& c0)
	{
		CVector3 ab = b0 - a0;
		CVector3 ac = c0 - a0;
		CVector3 abXac = CVector3::Cross(ab, ac);

		float a = abXac.x;
		float b = abXac.y;
		float c = abXac.z;
		float d = abXac.x * a0.x + abXac.y * a0.y + abXac.z * a0.z;
	}

	void CPlane::Normalize(CPlane& p)
	{
		float dist = sqrtf(p.a * p.a + p.b * p.b + p.c * p.c);

		p.a /= dist;
		p.b /= dist;
		p.c /= dist;
		p.d /= dist;
	}

	CPlane CPlane::Normalized(const CPlane& p)
	{
		CPlane result;
		float dist = sqrtf(p.a * p.a + p.b * p.b + p.c * p.c);

		result.a = p.a / dist;
		result.b = p.b / dist;
		result.c = p.c / dist;
		result.d = p.d / dist;

		return result;
	}
};