//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathQuaternion.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathQuaternion.h"

namespace Math
{
	CQuaternion::CQuaternion()
	{
		Identity();
	}

	CQuaternion::CQuaternion(float W, float X, float Y, float Z)
	{
		w = W;
		x = X;
		y = Y;
		z = Z;
	}

	CQuaternion::CQuaternion(const CVector3& euler)
	{
		RotateEuler(euler);
	}

	CQuaternion::CQuaternion(const CQuaternion& other)
	{
		w = other.w;
		x = other.x;
		y = other.y;
		z = other.z;
	}

	CQuaternion::CQuaternion(CQuaternion&& other)
	{
		w = other.w; other.w = 0.0f;
		x = other.x; other.x = 0.0f;
		y = other.y; other.y = 0.0f;
		z = other.z; other.z = 0.0f;
	}
	
	//
	// CQuaternion operators.
	//

	CQuaternion& CQuaternion::operator = (const CQuaternion& q)
	{
		w = q.w;
		x = q.x;
		y = q.y;
		z = q.z;

		return *this;
	}
	
	CQuaternion CQuaternion::operator * (const CQuaternion& q)
	{
		return CQuaternion(
			w * q.w - x * q.x - y * q.y - z * q.z,
			w * q.x + x * q.w + y * q.z - z * q.y,
			w * q.y - x * q.z + y * q.w + z * q.x,
			w * q.z + x * q.y + y * q.x + z * q.w
		);
	}
	
	CQuaternion& CQuaternion::operator *= (const CQuaternion& q)
	{
		*this = *this * q;
		return *this;
	}
	
	//
	// CQuaternion rotational methods.
	//

	void CQuaternion::RotateX(float theta)
	{
		float thetaOver2 = theta * 0.5f;
		
		w = cosf(thetaOver2);

		x = sinf(thetaOver2);
		y = 0.0f;
		z = 0.0f;
	}

	void CQuaternion::RotateY(float theta)
	{
		float thetaOver2 = theta * 0.5f;

		w = cosf(thetaOver2);

		x = 0.0f;
		y = sinf(thetaOver2);
		z = 0.0f;
	}

	void CQuaternion::RotateZ(float theta)
	{
		float thetaOver2 = theta * 0.5f;

		w = cosf(thetaOver2);

		x = 0.0f;
		y = 0.0f;
		z = sinf(thetaOver2);
	}

	void CQuaternion::RotateAxis(const CVector3& axis, float theta)
	{
		float thetaOver2 = theta * 0.5f;
		float sinThetaOver2 = sinf(thetaOver2);

		w = cosf(thetaOver2);
		x = axis.x * sinThetaOver2;
		y = axis.y * sinThetaOver2;
		z = axis.z * sinThetaOver2;
	}

	// Methods for rotating a quaternion to and from euler angles.

	void CQuaternion::RotateEuler(CVector3 euler)
	{
		euler = euler * 0.5f;
		float sx = sinf(euler.x); float cx = cosf(euler.x);
		float sy = sinf(euler.y); float cy = cosf(euler.y);
		float sz = sinf(euler.z); float cz = cosf(euler.z);
		
		w =  cy * cx * cz + sy * sx * sz;
		x =  cy * sx * cz + sy * cx * sz;
		y = -cy * sx * sz + sy * cx * cz;
		z = -sy * sx * cz + cy * cx * sz;
	}

	CVector3 CQuaternion::ToEuler()
	{
		CVector3 euler;
		float sx = -2.0f * (y * z - w * x);

		if(fabsf(sx) > 0.9999f)
		{
			euler.x = _PiOver2 * sx;
			euler.y = atan2f(-x * z + w * y, 0.5f - y * y - z * z);
			euler.z = 0.0f;
		}
		else
		{
			euler.x = asinf(sx);
			euler.y = atan2f(x * z + w * y, 0.5f - x * x - y * y);
			euler.z = atan2f(x * y + w * z, 0.5f - x * x - z * z);
		}

		return euler;
	}

	//
	// Quaternion normalization methods.
	//

	CQuaternion CQuaternion::Normalize()
	{
		float mag = sqrtf(w * w + x * x + y * y + z * z);
		CQuaternion q = *this;

		if(mag > 0.0f)
		{
			mag = 1.0f / mag;
			q.w *= mag;
			q.x *= mag;
			q.y *= mag;
			q.z *= mag;
		}
		else
		{
			q.Identity();
		}

		return q;
	}

	CQuaternion CQuaternion::Normalize(CQuaternion q)
	{
		float mag = sqrtf(q.w * q.w + q.x * q.x + q.y * q.y + q.z * q.z);

		if(mag > 0.0f)
		{
			mag = 1.0f / mag;
			q.w *= mag;
			q.x *= mag;
			q.y *= mag;
			q.z *= mag;
		}
		else
		{
			q.Identity();
		}

		return q;
	}

	//
	// Quaternion conjugate methods.
	//

	CQuaternion CQuaternion::Conjugate()
	{
		CQuaternion q = *this;

		q.x = -q.x;
		q.y = -q.y;
		q.z = -q.z;

		return q;
	}

	CQuaternion CQuaternion::Conjugate(const CQuaternion& q)
	{
		return CQuaternion(q.w, -q.x, -q.y, -q.z);
	}

	//
	// Quaternion inverse methods.
	//

	CQuaternion CQuaternion::Inverse()
	{
		CQuaternion q = (*this).Conjugate();
		float lenSq = (*this).LengthSq();

		q.w /= lenSq;
		q.x /= lenSq;
		q.y /= lenSq;
		q.z /= lenSq;

		return q;
	}
	
	CQuaternion CQuaternion::Inverse(CQuaternion q)
	{
		CQuaternion q0 = q.Conjugate();
		float lenSq = q.LengthSq();

		q0.w /= lenSq;
		q0.x /= lenSq;
		q0.y /= lenSq;
		q0.z /= lenSq;

		return q0;
	}

	// Quaternion spherical interpolation methods.
	CQuaternion CQuaternion::Slerp(const CQuaternion& from, const CQuaternion& to, float t)
	{
		if(t <= 0.0f) return from;
		if(t >= 1.0f) return to;
		
		float cosOmega = Dot(from, to);

		float q1x = to.x;
		float q1y = to.y;
		float q1z = to.z;
		float q1w = to.w;

		if(cosOmega < 0.0f)
		{
			q1x = -q1x;
			q1y = -q1y;
			q1z = -q1z;
			q1w = -q1w;
			cosOmega = -cosOmega;
		}

		if(cosOmega >= 1.1f)
			return from;

		float k0, k1;

		if(cosOmega > 0.9999f)
		{
			k0 = 1.0f - t;
			k1 = t;
		}
		else
		{
			float sinOmega = sqrtf(1.0f - cosOmega * cosOmega);
			float omega = atan2f(sinOmega, cosOmega);
			float oneOverSinOmega = 1.0f / sinOmega;
			
			k0 = sinf((1.0f - t) * omega) * oneOverSinOmega;
			k1 = sinf(t * omega) * oneOverSinOmega;
		}

		CQuaternion result;

		result.x = k0 * from.x + k1 * q1x;
		result.y = k0 * from.y + k1 * q1y;
		result.z = k0 * from.z + k1 * q1z;
		result.w = k0 * from.w + k1 * q1w;

		return result;
	}
	
	//
	// Quaternion dot product methods.
	//

	float CQuaternion::Dot(const CQuaternion& q)
	{
		return w * q.w + x * q.x + y * q.y + z * q.z;
	}
	
	float CQuaternion::Dot(const CQuaternion& q0, const CQuaternion& q1)
	{
		return q0.w * q1.w + q0.x * q1.x + q0.y * q1.y + q0.z * q1.z;
	}
	
	//
	// Quaternion length methods.
	//

	float CQuaternion::LengthSq(const CQuaternion& q)
	{
		return (q.w * q.w + q.x * q.x + q.y * q.y + q.z * q.z);
	}

	float CQuaternion::Length(const CQuaternion& q)
	{
		return sqrtf(LengthSq(q));
	}
};