//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector2.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHVECTOR2_H
#define CMATHVECTOR2_H

#include <cmath>
#include <minmax.h>
#include "CMathFloat.h"
#include "CMathVector3.h"
#include "CMathVector4.h"

namespace Math
{
	// CVector2 class
	// Class for create a 2d vector.
	class CVector2
	{
	public:
		float x;
		float y;

		CVector2();
		CVector2(float X, float Y);
		CVector2(const CVector2& other);
		CVector2(CVector2&& other);

		operator const float*() const		{ return (float*)this; }
		operator CVector3() const			{ return CVector3(x, y, 0.0f); }
		operator CVector4() const			{ return CVector4(x, y, 0.0f, 0.0f); }

		CVector2&							operator =	(float f);
		CVector2&							operator =	(const CVector2& v);
		CVector2							operator -	() const;
		
		CVector2							operator +	(const CVector2& v) const;
		CVector2							operator +	(float f) const;
		CVector2							operator -	(const CVector2& v) const;
		CVector2							operator -	(float f) const;
		CVector2							operator *	(const CVector2& v) const;
		CVector2							operator *	(float f) const;
		CVector2							operator /	(const CVector2& v) const;
		CVector2							operator /	(float f) const;
		
		CVector2&							operator +=	(const CVector2& v);
		CVector2&							operator +=	(float f);
		CVector2&							operator -=	(const CVector2& v);
		CVector2&							operator -=	(float f);
		CVector2&							operator *=	(const CVector2& v);
		CVector2&							operator *=	(float f);
		CVector2&							operator /=	(const CVector2& v);
		CVector2&							operator /=	(float f);

		bool								operator ==	(const CVector2& v) const;
		bool								operator !=	(const CVector2& v) const;
		
		friend CVector2						operator +	(float f, const CVector2& v);
		friend CVector2						operator -	(float f, const CVector2& v);
		friend CVector2						operator *	(float f, const CVector2& v);
		friend CVector2						operator /	(float f, const CVector2& v);

		// Inline methods.
		inline float						LengthSq();
		inline float						Length();
		inline void							Normalize();
		inline CVector2						Normalized();
		inline float						Dot(const CVector2& v);

		// Static methods.
		static float						LengthSq(const CVector2& v);
		static float						Length(CVector2 v);
		static void							Normalize(CVector2& v);
		static CVector2						Normalized(const CVector2& v);
		static float						Dot(const CVector2& v0, const CVector2& v1);
		static CVector2						Min(const CVector2& v0, const CVector2& v1);
		static CVector2						Max(const CVector2& v0, const CVector2& v1);

		friend CVector2						MoveTowards(const CVector2& from, const CVector2& to, float t);
		friend CVector2						MoveTowards(const CVector2& from, const CVector2& to, const CVector2& t);
		friend CVector2						Lerp(const CVector2& from, const CVector2& to, float t);
		friend CVector2						Lerp(const CVector2& from, const CVector2& to, const CVector2& t);

		static CVector2						Zero();
		static CVector2						One();

		static CVector2						Right();
		static CVector2						Left();
		static CVector2						Up();
		static CVector2						Down();
	};

	// Length getters.

	inline float CVector2::LengthSq()
	{
		return (x * x + y * y);
	}

	inline float CVector2::Length()
	{
		return sqrtf(LengthSq());
	}

	// Normalization getter and setter.

	inline void CVector2::Normalize()
	{
		*this /= Length();
	}

	inline CVector2 CVector2::Normalized()
	{
		return *this / Length();
	}

	// Dot product method.
	inline float CVector2::Dot(const CVector2& v)
	{
		return x * v.x + y * v.y;
	}

	// CVector2 constants.
	const CVector2												VEC2_ZERO								= CVector2(0.0f, 0.0f);
	const CVector2												VEC2_ONE								= CVector2(1.0f, 1.0f);

	const CVector2												VEC2_RIGHT								= CVector2(1.0f, 0.0f);
	const CVector2												VEC2_LEFT								= CVector2(-1.0f, 0.0f);
	const CVector2												VEC2_UP									= CVector2(0.0f, 1.0f);
	const CVector2												VEC2_DOWN								= CVector2(0.0f, -1.0f);
};

#endif
