//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector3.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHVECTOR3_H
#define CMATHVECTOR3_H

#include <cmath>
#include <minmax.h>
#include "CMathFloat.h"
#include "CMathVector4.h"

namespace Math
{
	// CVector3 class
	// Class for create a 3d vector.
	class CVector3
	{
	public:
		float x;
		float y;
		float z;

		CVector3();
		CVector3(float X, float Y, float Z);
		CVector3(const CVector3& other);
		CVector3(CVector3&& other);

		operator const float*() const		{ return (float*)this; }
		operator CVector4() const			{ return CVector4(x, y, z, 0.0f); }

		CVector3&							operator =	(float f);
		CVector3&							operator =	(const CVector3& v);
		CVector3							operator -	() const;
		
		CVector3							operator +	(const CVector3& v) const;
		CVector3							operator +	(float f) const;
		CVector3							operator -	(const CVector3& v) const;
		CVector3							operator -	(float f) const;
		CVector3							operator *	(const CVector3& v) const;
		CVector3							operator *	(float f) const;
		CVector3							operator /	(const CVector3& v) const;
		CVector3							operator /	(float f) const;
		
		CVector3&							operator +=	(const CVector3& v);
		CVector3&							operator +=	(float f);
		CVector3&							operator -=	(const CVector3& v);
		CVector3&							operator -=	(float f);
		CVector3&							operator *=	(const CVector3& v);
		CVector3&							operator *=	(float f);
		CVector3&							operator /=	(const CVector3& v);
		CVector3&							operator /=	(float f);

		bool								operator ==	(const CVector3& v) const;
		bool								operator !=	(const CVector3& v) const;
		
		friend CVector3						operator +	(float f, const CVector3& v);
		friend CVector3						operator -	(float f, const CVector3& v);
		friend CVector3						operator *	(float f, const CVector3& v);
		friend CVector3						operator /	(float f, const CVector3& v);

		// Inline methods.
		inline float						LengthSq();
		inline float						Length();
		inline void							Normalize();
		inline CVector3						Normalized();
		inline float						Dot(const CVector3& v);
		inline CVector3						Cross(const CVector3& v);
		
		// Static methods.
		static float						LengthSq(const CVector3& v);
		static float						Length(CVector3 v);
		static void							Normalize(CVector3 v);
		static CVector3						Normalized(const CVector3& v);
		static float						Dot(const CVector3& v0, const CVector3& v1);
		static CVector3						Cross(const CVector3& v0, const CVector3& v1);
		static CVector3						Min(const CVector3& v0, const CVector3& v1);
		static CVector3						Max(const CVector3& v0, const CVector3& v1);
		
		friend CVector3						MoveTowards(const CVector3& from, const CVector3& to, float t);
		friend CVector3						MoveTowards(const CVector3& from, const CVector3& to, const CVector3& t);
		friend CVector3						Lerp(const CVector3& from, const CVector3& to, float t);
		friend CVector3						Lerp(const CVector3& from, const CVector3& to, const CVector3& t);

		static CVector3						Zero();
		static CVector3						One();

		static CVector3						Right();
		static CVector3						Left();
		static CVector3						Up();
		static CVector3						Down();
		static CVector3						Forward();
		static CVector3						Back();
	};

	// Length getters.

	inline float CVector3::LengthSq()
	{
		return (x * x + y * y + z * z);
	}

	inline float CVector3::Length()
	{
		return sqrtf(LengthSq());
	}

	// Normalization getter and setter.

	inline void CVector3::Normalize()
	{
		*this /= Length();
	}

	inline CVector3 CVector3::Normalized()
	{
		return *this / Length();
	}
	
	// Dot product method.
	inline float CVector3::Dot(const CVector3& v)
	{
		return x * v.x + y * v.y + z * v.z;
	}
		
	// Cross product method.
	inline CVector3 CVector3::Cross(const CVector3& v)
	{
		return CVector3(
			(y * v.z - z * v.y), 
			(z * v.x - x * v.z), 
			(x * v.y - y * v.x)
		); 
	}
	
	// CVector3 constants.
	const CVector3												VEC3_ZERO								= CVector3(0.0f, 0.0f, 0.0f);
	const CVector3												VEC3_ONE								= CVector3(1.0f, 1.0f, 1.0f);

	const CVector3												VEC3_RIGHT								= CVector3(1.0f, 0.0f, 0.0f);
	const CVector3												VEC3_LEFT								= CVector3(-1.0f, 0.0f, 0.0f);
	const CVector3												VEC3_UP									= CVector3(0.0f, 1.0f, 0.0f);
	const CVector3												VEC3_DOWN								= CVector3(0.0f, -1.0f, 0.0f);
	const CVector3												VEC3_FORWARD							= CVector3(0.0f, 0.0f, 1.0f);
	const CVector3												VEC3_BACK								= CVector3(0.0f, 0.0f, -1.0f);
};

#endif