//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector4.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathVector4.h"

namespace Math
{
	CVector4::CVector4()
	{
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
		w = 0.0f;
	}

	CVector4::CVector4(float X, float Y, float Z, float W)
	{
		x = X;
		y = Y;
		z = Z;
		w = W;
	}

	CVector4::CVector4(const CVector4& other)
	{
		x = other.x;
		y = other.y;
		z = other.z;
		w = other.w;
	}

	CVector4::CVector4(CVector4&& other)
	{
		x = other.x; other.x = 0.0f;
		y = other.y; other.y = 0.0f;
		z = other.z; other.z = 0.0f;
		w = other.w; other.w = 0.0f;
	}

	CVector4& CVector4::operator = (float f)
	{
		x = f;
		y = f;
		z = f;
		w = f;

		return *this;
	}

	CVector4& CVector4::operator = (const CVector4& v)
	{
		x = v.x;
		y = v.y;
		z = v.z;
		w = v.w;

		return *this;
	}

	CVector4 CVector4::operator - () const
	{
		return CVector4(-x, -y, -z, -w);
	}

	//
	// Addition operators.
	//
	
	CVector4 CVector4::operator + (const CVector4& v) const
	{
		return CVector4(x + v.x, y + v.y, z + v.z, w + v.w);
	}

	CVector4 CVector4::operator + (float f) const
	{
		return CVector4(x + f, y + f, z + f, w + f);
	}

	CVector4& CVector4::operator += (const CVector4& v)
	{
		*this = *this + v;
		return *this;
	}

	CVector4& CVector4::operator += (float f)
	{
		*this = *this + f;
		return *this;
	}

	//
	// Subtraction operators.
	//
		
	CVector4 CVector4::operator - (const CVector4& v) const
	{
		return CVector4(x - v.x, y - v.y, z - v.z, w - v.w);
	}

	CVector4 CVector4::operator - (float f) const
	{
		return CVector4(x - f, y - f, z - f, w - f);
	}

	CVector4& CVector4::operator -= (const CVector4& v)
	{
		*this = *this - v;
		return *this;
	}

	CVector4& CVector4::operator -= (float f)
	{
		*this = *this - f;
		return *this;
	}

	//
	// Multiplication operators.
	//

	CVector4 CVector4::operator * (const CVector4& v) const
	{
		return CVector4(x * v.x, y * v.y, z * v.z, w * v.w);
	}

	CVector4 CVector4::operator * (float f) const
	{
		return CVector4(x *f, y * f, z * f, w * f);
	}

	CVector4& CVector4::operator *= (const CVector4& v)
	{
		*this = *this * v;
		return *this;
	}

	CVector4& CVector4::operator *= (float f)
	{
		*this = *this * f;
		return *this;
	}

	//
	// Division operators.
	//

	CVector4 CVector4::operator / (const CVector4& v) const
	{
		return CVector4(x / v.x, y / v.y, z / v.z, w / v.w);
	}

	CVector4 CVector4::operator / (float f) const
	{
		return CVector4(x / f, y / f, z / f, w / f);
	}

	CVector4& CVector4::operator /= (const CVector4& v)
	{
		*this = *this / v;
		return *this;
	}

	CVector4& CVector4::operator /= (float f)
	{
		*this = *this / f;
		return *this;
	}

	//
	// Conditional operators.
	//

	bool CVector4::operator == (const CVector4& v) const
	{
		return (x == v.x && y == v.y && z == v.z && w == v.w);
	}

	bool CVector4::operator != (const CVector4& v) const
	{
		return (x != v.x || y != v.y || z != v.z || w != v.w);
	}
	
	//
	// Friend operators.
	//

	CVector4 operator + (float f, const CVector4& v)
	{
		return v + f;
	}

	CVector4 operator - (float f, const CVector4& v)
	{
		return v - f;
	}

	CVector4 operator * (float f, const CVector4& v)
	{
		return v * f;
	}

	CVector4 operator / (float f, const CVector4& v)
	{
		return v / f;
	}
	
	//
	// Vector4 static mathematic methods.
	//

	float CVector4::LengthSq(const CVector4& v)
	{
		return (v.x * v.x + v.y * v.y + v.z * v.z + v.w * v.w);
	}

	float CVector4::Length(CVector4 v)
	{
		return sqrtf(v.LengthSq());
	}

	void CVector4::Normalize(CVector4& v)
	{
		v /= Length(v);
	}

	CVector4 CVector4::Normalized(const CVector4& v)
	{
		return v / Length(v);
	}

	float CVector4::Dot(const CVector4& v0, const CVector4& v1)
	{
		return v0.x * v1.x + v0.y * v1.y + v0.z * v1.z + v0.w * v1.w;
	}
	
	CVector4 CVector4::Min(const CVector4& v0, const CVector4& v1)
	{
		return CVector4(min(v0.x, v1.x), min(v0.y, v1.y), min(v0.z, v1.z), min(v0.w, v1.w));
	}

	CVector4 CVector4::Max(const CVector4& v0, const CVector4& v1)
	{
		return CVector4(max(v0.x, v1.x), max(v0.y, v1.y), max(v0.z, v1.z), max(v0.w, v1.w));
	}
	
	//
	// Vector4 interpolation methods.
	//
	
	CVector4 MoveTowards(const CVector4& from, const CVector4& to, float t)
	{
		return CVector4(MoveTowards(from.x, to.x, t), MoveTowards(from.y, to.y, t), MoveTowards(from.z, to.z, t), MoveTowards(from.w, to.w, t));
	}
	
	CVector4 MoveTowards(const CVector4& from, const CVector4& to, const CVector4& t)
	{
		return CVector4(MoveTowards(from.x, to.x, t.x), MoveTowards(from.y, to.y, t.y), MoveTowards(from.z, to.z, t.z), MoveTowards(from.w, to.w, t.w));
	}

	CVector4 Lerp(const CVector4& from, const CVector4& to, float t)
	{
		return CVector4(Lerp(from.x, to.x, t), Lerp(from.y, to.y, t), Lerp(from.z, to.z, t), Lerp(from.w, to.w, t));
	}

	CVector4 Lerp(const CVector4& from, const CVector4& to, const CVector4& t)
	{
		return CVector4(Lerp(from.x, to.x, t.x), Lerp(from.y, to.y, t.y), Lerp(from.z, to.z, t.z), Lerp(from.w, to.w, t.w));
	}

	//
	// Vector4 variable methods.
	//
	
	CVector4 CVector4::Zero()
	{
		return VEC4_ZERO;
	}

	CVector4 CVector4::One()
	{
		return VEC4_ONE;
	}
};