//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector4.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHVECTOR4_H
#define CMATHVECTOR4_H

#include <cmath>
#include <minmax.h>
#include "CMathFloat.h"

namespace Math
{
	// CVector4 class
	// Class for create a 4d vector.
	class CVector4
	{
	public:
		float x;
		float y;
		float z;
		float w;

		CVector4();
		CVector4(float X, float Y, float Z, float W);
		CVector4(const CVector4& other);
		CVector4(CVector4&& other);

		operator const float*() const		{ return (float*)this; }

		CVector4&							operator =	(float f);
		CVector4&							operator =	(const CVector4& v);
		CVector4							operator -	() const;
		
		CVector4							operator +	(const CVector4& v) const;
		CVector4							operator +	(float f) const;
		CVector4							operator -	(const CVector4& v) const;
		CVector4							operator -	(float f) const;
		CVector4							operator *	(const CVector4& v) const;
		CVector4							operator *	(float f) const;
		CVector4							operator /	(const CVector4& v) const;
		CVector4							operator /	(float f) const;
		
		CVector4&							operator +=	(const CVector4& v);
		CVector4&							operator +=	(float f);
		CVector4&							operator -=	(const CVector4& v);
		CVector4&							operator -=	(float f);
		CVector4&							operator *=	(const CVector4& v);
		CVector4&							operator *=	(float f);
		CVector4&							operator /=	(const CVector4& v);
		CVector4&							operator /=	(float f);

		bool								operator ==	(const CVector4& v) const;
		bool								operator !=	(const CVector4& v) const;
		
		friend CVector4						operator +	(float f, const CVector4& v);
		friend CVector4						operator -	(float f, const CVector4& v);
		friend CVector4						operator *	(float f, const CVector4& v);
		friend CVector4						operator /	(float f, const CVector4& v);

		// Inline methods.
		inline float						LengthSq();
		inline float						Length();
		inline void							Normalize();
		inline CVector4						Normalized();
		inline float						Dot(const CVector4& v);
		
		// Static methods.
		static float						LengthSq(const CVector4& v);
		static float						Length(CVector4 v);
		static void							Normalize(CVector4& v);
		static CVector4						Normalized(const CVector4& v);
		static float						Dot(const CVector4& v0, const CVector4& v1);
		static CVector4						Min(const CVector4& v0, const CVector4& v1);
		static CVector4						Max(const CVector4& v0, const CVector4& v1);
		
		friend CVector4						MoveTowards(const CVector4& from, const CVector4& to, float t);
		friend CVector4						MoveTowards(const CVector4& from, const CVector4& to, const CVector4& t);
		friend CVector4						Lerp(const CVector4& from, const CVector4& to, float t);
		friend CVector4						Lerp(const CVector4& from, const CVector4& to, const CVector4& t);

		static CVector4						Zero();
		static CVector4						One();
	};
	
	// Length getters.

	inline float CVector4::LengthSq()
	{
		return (x * x + y * y + z * z + w * w);
	}

	inline float CVector4::Length()
	{
		return sqrtf(LengthSq());
	}

	// Normalization getter and setter.

	inline void CVector4::Normalize()
	{
		*this /= Length();
	}

	inline CVector4 CVector4::Normalized()
	{
		return *this / Length();
	}

	// Dot product method.
	inline float CVector4::Dot(const CVector4& v)
	{
		return x * v.x + y * v.y + z * v.z + w * v.w;
	}
		
	// CVector4 constants.
	const CVector4												VEC4_ZERO								= CVector4(0.0f, 0.0f, 0.0f, 0.0f);
	const CVector4												VEC4_ONE								= CVector4(1.0f, 1.0f, 1.0f, 1.0f);
};

#endif