//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Physics\CPhysics.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CPhysics.h"

namespace Physics
{
	//
	// Physics general.
	//

	// Function for testing a ray against a plane.
	bool RayIntersectPlane(const Math::CRay& ray, const Math::CPlane& plane, float& t, Math::CVector3* intersectionPoint)
	{
		float projRayLength = ray.Direction().Dot(plane.Normal());

		if(projRayLength > -Math::_Epsilon)
			return false;

		float planeDistance = ray.Origin().Dot(plane.Normal()) + plane.d;
		
		t = -(planeDistance / projRayLength);

		if(t < 0.0f)
			return false;

		if(intersectionPoint) *intersectionPoint = ray.Origin() + (ray.Direction() * t);

		return true;
	}

	// Function for finding the closest two points between two lines.
	bool LineLineDistance(const Math::CVector3& p0, const Math::CVector3& p1, const Math::CVector3& p2, const Math::CVector3& p3, Math::CVector3& point0, Math::CVector3& point1)
	{
		Math::CVector3 u = p1 - p0;
		Math::CVector3 v = p3 - p2;
		Math::CVector3 w = p0 - p2;

		float a = Math::CVector3::Dot(u, u);
		float b = Math::CVector3::Dot(u, v);
		float c = Math::CVector3::Dot(v, v);
		float d = Math::CVector3::Dot(u, w);
		float e = Math::CVector3::Dot(v, w);
		float D = a * c - b * b;
		float sc, sN, sD = D;
		float tc, tN, tD = D;

		if(D < Math::_Epsilon)
		{
			sN = 0.0f;
			sD = 1.0f;
			tN = e;
			tD = c;
		}
		else
		{
			sN = (b * e - c * d);
			tN = (a * e - b * d);
			if(sN < 0.0f)
			{
				sN = 0.0f;
				tN = e;
				tD = c;
			}
			else if(sN > sD)
			{
				sN = sD;
				tN = e + b;
				tD = c;
			}
		}

		if(tN < 0.0f)
		{
			tN = 0.0f;
			if(-d < 0.0f)
				sN = sD;
			else if(-d > a)
				sN = sD;
			else
			{
				sN = -d;
				sD = a;
			}
		}
		else if(tN > tD)
		{
			tN = tD;
			if((-d + b) < 0.0f)
				sN = 0.0f;
			else if((-d + b) > a)
				sN = sD;
			else
			{
				sN = (-d + b);
				sD = a;
			}
		}

		sc = (fabsf(sN) < Math::_Epsilon ? 0.0f : sN / sD);
		tc = (fabsf(tN) < Math::_Epsilon ? 0.0f : tN / tD);

		Math::CVector3 dP = w + ((u * sc) - (v * tc));
		point0 = p0 + (u * sc);
		point1 = p2 + (v * tc);

		return true;
	}

	// Function for finding the closest point between two lines.
	bool LineLineIntersect(const Math::CVector3& p0, const Math::CVector3& p1, const Math::CVector3& p2, const Math::CVector3& p3, Math::CVector3& pa, Math::CVector3& pb, float& mua, float& mub)
	{
		float numer, denom;
		
		Math::CVector3 u = p1 - p0;
		Math::CVector3 v = p3 - p2;
		Math::CVector3 w = p0 - p2;
		
		float a = Math::CVector3::Dot(u, u);
		float b = Math::CVector3::Dot(u, v);
		float c = Math::CVector3::Dot(v, v);
		float d = Math::CVector3::Dot(u, w);
		float e = Math::CVector3::Dot(v, w);

		denom = a * c - b * b;

		if (fabsf(denom) < Math::_Epsilon)
			return false;

		numer = e * b - d * c;

		mua = numer / denom;
		mub = (e + b * (mua)) / c;

		pa.x = p0.x + mua * u.x;
		pa.y = p0.y + mua * u.y;
		pa.z = p0.z + mua * u.z;
		pb.x = p2.x + mub * v.x;
		pb.y = p2.y + mub * v.y;
		pb.z = p2.z + mub * v.z;

		return true;
	}

	// Function for line vs point intersection.
	bool LinePointIntersect(const Math::CVector3& point, const Math::CVector3& p0, const Math::CVector3& p1, Math::CVector3& closestPoint, float& distance)
	{
		Math::CVector3 segVec = p1 - p0;
		Math::CVector3 segVecNorm = segVec.Normalized();
		Math::CVector3 ptVec = point - p0;

		float projection = ptVec.Dot(segVecNorm);
		
		if(projection < 0)
		{
			closestPoint = p0;
			distance = (point - p0).Length();
			return true;
		}
		else if(projection > segVec.Length())
		{
			closestPoint = p1;
			distance = (point - p1).Length();
			return true;
		}

		Math::CVector3 projVec = segVecNorm * projection;
		closestPoint = p0 + projVec;

		return true;
	}

	// Function for checking if a point is in a triangle.
	bool PointInTriangle(const Math::CVector3& point, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, const Math::CVector3& triNormal)
	{
		Math::CVector3 edge, edgeNormal, direction;

		// Edge 0.
		edge = v1 - v0;
		direction = v0 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		// Edge 1.
		edge = v2 - v1;
		direction = v1 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		// Edge 2.
		edge = v0 - v2;
		direction = v2 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		return true;
	}

	// Function for checking if a point is in a quad.
	bool PointInQuad(const Math::CVector3& point, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, const Math::CVector3& v3, const Math::CVector3& triNormal)
	{
		Math::CVector3 edge, edgeNormal, direction;

		// Edge 0.
		edge = v1 - v0;
		direction = v0 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		// Edge 1.
		edge = v2 - v1;
		direction = v1 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		// Edge 2.
		edge = v3 - v2;
		direction = v2 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		// Edge 3.
		edge = v0 - v3;
		direction = v3 - point;
		edgeNormal = edge.Cross(triNormal);

		if(direction.Dot(edgeNormal) < 0.0f)
			return false;

		return true;
	}

	// Function for checking if a point is in an n-gon.
	bool PointInNGon(const Math::CVector3& point, Math::CVector3*& vertices, u32 vertexCount, const Math::CVector3& triNormal)
	{
		if(vertexCount <= 1)
			return false;

		Math::CVector3 edge, edgeNormal, direction;

		// Check all of the edges in the n-gon.
		for(u32 i = 0; i < vertexCount; i++)
		{
			if(i < vertexCount - 1)
			{
				edge = vertices[i + 1] - vertices[i];
				direction = vertices[i] - point;
			}
			else
			{
				edge = vertices[0] - vertices[i];
				direction = vertices[i] - point;
			}

			edgeNormal = edge.Cross(triNormal);

			if(direction.Dot(edgeNormal) < 0.0f)
				return false;
		}

		return true;
	}

	// Function for solving collision quadratic equations.
	bool CollistionSolver(float a, float b, float c, float& t)
	{
		float d = b * b - 4 * a * c;

		if(d < 0 || fabsf(a) <= Math::_Epsilon)
			return false;

		d = sqrtf(d);

		float oneOverTwoA = 1.0f / (2.0f * a);

		float t0 = (-b - d) * oneOverTwoA;
		float t1 = (-b + d) * oneOverTwoA;

		if(t1 < t0)
		{
			float temp = t0;
			t0 = t1;
			t1 = temp;
		}

		if(t1 <= 0)
			return false;

		if(t0 < 0.0f)
			t = t1;
		else
			t = t0;

		return true;
	}

	//
	// Physics sphere.
	//

	// Function for checking if a sphere intersects a plane.
	bool SphereIntersectPlane(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& planeNormal, const Math::CVector3& planePoint, float& tMax)
	{
		float numer = Math::CVector3::Dot(center - planePoint, planeNormal) - radius;
		float denom = Math::CVector3::Dot(velocity, planeNormal);
		
		if(numer < 0.0f || denom > -Math::_Epsilon)
		{
			if(denom > -Math::_Epsilon)
				return false;
			if(numer < -radius)
				return false;

			tMax = numer;

			return true;
		}

		float t = -(numer / denom);

		if(t < 0.0f || t > tMax)
			return false;

		tMax = t;

		return true;
	}

	// Function for checking if a sphere intersects a triangle.
	bool SphereIntersectTriangle(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		float t = tMax;

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, v0, t))
			return false;

		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInTriangle(collisionCenter, v0, v1, v2, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;

			return true;
		}
		
		bool bCollided = false;

		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v0, v1, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v1, v2, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v2, v0, tMax, collisionNormal);

		return bCollided;
	}

	// Function for checking if a sphere intersects a quad.
	bool SphereIntersectQuad(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, const Math::CVector3& v3, const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		float t = tMax;

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, v0, t))
			return false;

		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInQuad(collisionCenter, v0, v1, v2, v3, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;

			return true;
		}
		
		bool bCollided = false;

		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v0, v1, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v1, v2, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v2, v3, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v3, v0, tMax, collisionNormal);

		return bCollided;
	}

	// Function for checking if sphere intersects a n-gon.
	bool SphereIntersectNGon(const Math::CVector3& center, float radius, const Math::CVector3& velocity, Math::CVector3*& vertices, u32 vertexCount, const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		if(vertexCount <= 1)
			return false;

		float t = tMax;

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, vertices[0], t))
			return false;

		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInNGon(collisionCenter, vertices, vertexCount, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;

			return true;
		}
		
		bool bCollided = false;

		for(u32 i = 0; i < vertexCount; i++)
		{
			if(i < vertexCount - 1)
				bCollided |= SphereIntersectLineSegment(center, radius, velocity, vertices[i], vertices[i + 1], tMax, collisionNormal);
			else
				bCollided |= SphereIntersectLineSegment(center, radius, velocity, vertices[i], vertices[0], tMax, collisionNormal);
		}

		return bCollided;
	}

	// Function for checking if a sphere intersects a line segment.
	bool SphereIntersectLineSegment(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& v0, const Math::CVector3& v1, float& tMax, Math::CVector3& collisionNormal)
	{
		Math::CVector3 E = v1 - v0;
		float e = E.Length();

		if(e < Math::_Epsilon)
			return false;

		E /= e;

		Math::CVector3 D = center - v0;

		Math::CVector3 X = Math::CVector3::Cross(D, E);
		Math::CVector3 Y = Math::CVector3::Cross(velocity, E);
		
		float c = X.LengthSq() - (radius * radius);

		if(c < 0.0f)
		{
			float d = Math::CVector3::Dot(D, E);

			if(d < 0.0f)
			{
				return SphereIntersectPoint(center, radius, velocity, v0, tMax, collisionNormal);
			}
			else if(d > e)
			{
				return SphereIntersectPoint(center, radius, velocity, v1, tMax, collisionNormal);
			}
			else
			{
				Math::CVector3 pointOnEdge = v0 + E * d;

				collisionNormal = center - pointOnEdge;
				float n = Math::CVector3::Length(collisionNormal);
				collisionNormal /= n;

				float t = n - radius;
				
				if(t > tMax)
					return false;

				tMax = t;

				return true;
			}
		}

		if(tMax < 0.0f)
			return false;

		float a = Math::CVector3::LengthSq(Y);
		float b = 2.0f * Math::CVector3::Dot(X, Y);
		float t;

		if(!CollistionSolver(a, b, c, t))
			return false;

		if(t > tMax)
			return false;

		Math::CVector3 collisionCenter = center + velocity * t;
		float d = Math::CVector3::Dot(collisionCenter - v0, E);
		
		if(d < 0.0f)
		{
			return SphereIntersectPoint(center, radius, velocity, v0, tMax, collisionNormal);
		}
		else if(d > e)
		{
			return SphereIntersectPoint(center, radius, velocity, v1, tMax, collisionNormal);
		}

		Math::CVector3 pointOnEdge = v0 + E * d;
		collisionNormal = Math::CVector3::Normalized(collisionCenter - pointOnEdge);

		tMax = t;

		return true;
	}
	
	// Function for sphere vs point intersection.
	bool SphereIntersectPoint(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& point, float& tMax, Math::CVector3& collisionNormal)
	{
		Math::CVector3 difference = center - point;
		float lenSq = difference.LengthSq();

		float c = lenSq - (radius * radius);
		
		if(c < 0.0f)
		{
			float len = sqrtf(lenSq);
			float t = len - radius;

			if(tMax < t)
				return false;

			collisionNormal = difference / len;

			tMax = t;

			return true;
		}

		if(tMax < 0.0f)
			return false;

		float a = Math::CVector3::LengthSq(velocity);
		float b = 2.0f * Math::CVector3::Dot(velocity, difference);
		float t;

		if(!CollistionSolver(a, b, c, t))
			return false;

		if(t > tMax)
			return false;

		Math::CVector3 collisionCenter = center + velocity * t;
		collisionNormal = Math::CVector3::Normalized(collisionCenter - point);

		tMax = t;

		return true;
	}
	
	// Function for sphere vs sphere intersection.
	bool SphereIntersectSphere(const Math::CVector3& center, float radius, const Math::CVector3& velocity, const Math::CVector3& point, float pointRadius, float& tMax, Math::CVector3& collisionNormal)
	{
		float twoRadii = (radius + pointRadius);

		Math::CVector3 difference = center - point;
		float lenSq = difference.LengthSq();

		if(lenSq <= Math::_Epsilon)
			return false;

		float c = lenSq - twoRadii * twoRadii;
		
		if(c < 0.0f)
		{
			float len = sqrtf(lenSq);
			float t = len - twoRadii;

			if(tMax < t)
				return false;

			collisionNormal = difference / len;

			tMax = t;

			return true;
		}

		if(tMax < 0.0f)
			return false;

		float a = Math::CVector3::LengthSq(velocity);
		float b = 2.0f * Math::CVector3::Dot(velocity, difference);
		float t;

		if(!CollistionSolver(a, b, c, t))
			return false;

		if(tMax < t)
			return false;

		Math::CVector3 collisionCenter = center + velocity * t;
		collisionNormal = Math::CVector3::Normalized(collisionCenter - point);

		tMax = t;

		return true;
	}

	//
	// Physics capsule.
	//
	
	bool CapsuleIntersectTriangle(const Math::CVector3& point0, const Math::CVector3& point1, 
		float radius, const Math::CVector3& velocity, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, 
		const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		float t = tMax;

		Math::CVector3 center;

		float minDist = FLT_MAX;
		float tempDist = 0;
		Math::CVector3 tempPoint;
		Math::CVector3 edgePoint;
		bool intersection = false;
		
		// Fing the capsule's closest point to triangle line segments.
		if(LineLineDistance(point0, point1, v0, v1, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
		}
		if(LineLineDistance(point0, point1, v1, v2, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
		}
		if(LineLineDistance(point0, point1, v2, v0, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
		}

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, v0, t))
			return false;
		
		// Calculate the sphere's center point.
		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInTriangle(collisionCenter, v0, v1, v2, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;
			return true;
		}
		
		bool bCollided = false;
		
		// Test capsule closest point sphere against triangle line segments.
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v0, v1, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v1, v2, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v2, v0, tMax, collisionNormal);
		
		return bCollided;
	}

	bool CapsuleIntersectQuad(const Math::CVector3& point0, const Math::CVector3& point1, float radius, 
		const Math::CVector3& velocity, const Math::CVector3& v0, const Math::CVector3& v1, const Math::CVector3& v2, 
		const Math::CVector3& v3, const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		float t = tMax;

		Math::CVector3 center;

		float minDist = FLT_MAX;
		float tempDist = 0;
		Math::CVector3 tempPoint;
		bool intersection = false;
		Math::CVector3 edgePoint;
		int counter = 0;
		
		// Fing the capsule's closest point to quad line segments.
		if(LineLineDistance(point0, point1, v0, v1, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
			counter++;
		}
		if(LineLineDistance(point0, point1, v1, v2, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
			counter++;
		}
		if(LineLineDistance(point0, point1, v2, v3, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
			counter++;
		}
		if(LineLineDistance(point0, point1, v3, v0, tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
		{
			center = tempPoint;
			minDist = tempDist;
			counter++;
		}

		if(counter <= 0)
			return false;

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, v0, t))
			return false;
		
		// Calculate the sphere's center point.
		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInQuad(collisionCenter, v0, v1, v2, v3, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;
			return true;
		}
		
		bool bCollided = false;
		
		// Test capsule closest point sphere against quad line segments.
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v0, v1, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v1, v2, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v2, v3, tMax, collisionNormal);
		bCollided |= SphereIntersectLineSegment(center, radius, velocity, v3, v0, tMax, collisionNormal);
		
		return bCollided;
	}
	
	bool CapsuleIntersectNGon(const Math::CVector3& point0, const Math::CVector3& point1, float radius, const Math::CVector3& velocity, 
		Math::CVector3*& vertices, u32 vertexCount, const Math::CVector3& triNormal, float& tMax, Math::CVector3& collisionNormal)
	{
		if(vertexCount <= 1)
			return false;
			
		u32 i;
		float t = tMax;

		Math::CVector3 center;

		float minDist = FLT_MAX;
		float tempDist = 0;
		Math::CVector3 tempPoint;
		Math::CVector3 edgePoint;
		bool intersection = false;

		// Find the closest point on the capsule line to a vertex.
		for(i = 0; i < vertexCount; i++)
		{
			if(i < vertexCount - 1)
			{
				if(LineLineDistance(point0, point1, vertices[i], vertices[i + 1], tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
				{
					center = tempPoint;
					minDist = tempDist;
				}
			}
			else
			{
				if(LineLineDistance(point0, point1, vertices[i], vertices[0], tempPoint, edgePoint) && (tempDist = (edgePoint - tempPoint).LengthSq()) < minDist)
				{
					center = tempPoint;
					minDist = tempDist;
				}
			}
		}

		if(!SphereIntersectPlane(center, radius, velocity, triNormal, vertices[0], t))
			return false;
		
		// Calculate the sphere's center point.
		Math::CVector3 collisionCenter;

		if(t < 0)
			collisionCenter = center + (triNormal * -t);
		else
			collisionCenter = center + (velocity * t);

		if(PointInNGon(collisionCenter, vertices, vertexCount, triNormal))
		{
			collisionNormal = triNormal;
			tMax = t;
			return true;
		}
		
		bool bCollided = false;
		
		// Check capsule closest point sphere againts line segments.
		for(i = 0; i < vertexCount; i++)
		{
			if(i < vertexCount - 1)
			{
				bCollided |= SphereIntersectLineSegment(center, radius, velocity, vertices[i], vertices[i + 1], tMax, collisionNormal);
			}
			else
			{
				bCollided |= SphereIntersectLineSegment(center, radius, velocity, vertices[i], vertices[0], tMax, collisionNormal);
			}
		}
		
		return bCollided;
	}

	// Function for checking is a sapsule is intersecting a sphere.
	bool CapsuleIntersectSphere(const Math::CVector3& point0, const Math::CVector3& point1, float radius, const Math::CVector3& velocity, const Math::CVector3& point, float pointRadius, float& tMax, Math::CVector3& collisionNormal)
	{
		Math::CVector3 closestPoint;
		float distance = 0.0f;

		LinePointIntersect(point, point0, point1, closestPoint, distance);

		return SphereIntersectSphere(closestPoint, radius, velocity, point, pointRadius, tMax, collisionNormal);
	}
};