//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Resources\Shaders\FT2Font.hlsl
//
//-------------------------------------------------------------------------------------------------

#include "RenderUtil.hlsli"

// Matrix constant buffer.
cbuffer MatrixBuffer : register(b0)
{
	float4x4	World;
	float4x4	View;
	float4x4	Projection;
};

// Data constant buffer.
cbuffer ConstantBuffer : register(b1)
{
	float4		Color;
	bool		Textured;
	bool		PointFilter;
	float		FontSize;
};

// Vertex input structure.
struct VIn
{
	float4 Position			: POSITION;
	float2 TexCoord			: TEXCOORD;
	float4 InstancePos		: INSTPOS;
	float4 InstanceScale	: INSTSCALE;
	float4 InstanceUV		: INSTUV;
	float4 InstanceColor	: INSTCOLOR;
};

// Vertex output structure.
struct VOut
{
	float4 Position			: SV_POSITION;
	float4 Scale			: POSITION0;
	float4 WorldPos			: POSITION1;
	float4 TexCoord			: TEXCOORD0;
	float Size				: TEXCOORD1;
	float4 Color			: COLOR;
};

// Pixel output structure.
struct POut
{
	float4		Color			: SV_TARGET0;
	float4		Diffuse			: SV_TARGET1;
};

// Vertex shader function.
VOut VShader(VIn input)
{
	VOut output = (VOut)0;
	
	output.Scale			=	input.InstanceScale;
	output.WorldPos			=	float4(World._14, World._24, World._34, 1.0f) + float4(input.InstancePos.xyz, 0.0f);
	
	output.Color			=	Color * input.InstanceColor;
	output.TexCoord			=	input.InstanceUV;
	output.Size				=	input.InstanceScale.z;
	
	return output;
}

// Geometry shader function.
[maxvertexcount(4)]
void GShader(point VOut input[1], inout TriangleStream<VOut> outputStream)
{
	float3 vert[4];

	vert[0] = input[0].WorldPos.xyz + float3( 1,  0, 0) * input[0].Scale.xyz;
	vert[1] = input[0].WorldPos.xyz + float3( 0,  0, 0) * input[0].Scale.xyz;
	vert[2] = input[0].WorldPos.xyz + float3( 1,  1, 0) * input[0].Scale.xyz;
	vert[3] = input[0].WorldPos.xyz + float3( 0,  1, 0) * input[0].Scale.xyz;

	float2 texCoord[4];

	texCoord[0] = float2(1, 1);
	texCoord[1] = float2(0, 1);
	texCoord[2] = float2(1, 0);
	texCoord[3] = float2(0, 0);

	// Build the quad in the geometry shader.
	VOut output = input[0];
	for(int i = 0; i < 4; i++)
	{
		output.Position						=	mul(float4(vert[i], 1.0f), World);
		output.Position						=	mul(output.Position, View);
		output.Position						=	mul(output.Position, Projection);

		output.TexCoord.xy					=	texCoord[i] * input[0].TexCoord.zw + input[0].TexCoord.xy;

		outputStream.Append(output);
	}

	outputStream.RestartStrip();
}

Texture2D		Texture;

// Pixel shader function.
POut PShader(VOut input) : SV_TARGET
{
	if(input.Color.a <= 0.01f)
		discard;
		
	POut output;
	output.Diffuse = float4(1, 1, 1, 0.0f);

	if(!Textured)
	{
		output.Color = input.Color;
		return output;
	}
	else
	{
		float4 color;

		// Display font with signed distance field.
		float GlyphBuffer = 0.8f;
		float GlyphGamma = 0.055f / (input.Size);

		if(PointFilter)
		{
			color = Texture.Sample(samplerPointClamp, input.TexCoord);

			float distance = color.a;
			float alpha = smoothstep(GlyphBuffer - GlyphGamma, GlyphBuffer + GlyphGamma, distance);

			color.rgb = input.Color.rgb;
			color.a = alpha;
		}
		else
		{
			color = Texture.Sample(samplerLinearClamp, input.TexCoord);

			float distance = color.a;
			float alpha = smoothstep(GlyphBuffer - GlyphGamma, GlyphBuffer + GlyphGamma, distance);

			color.rgb = input.Color.rgb;
			color.a = alpha;
		}

		color.a *= input.Color.a;
		output.Color = color;

		return output;
	}
}