//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Resources\Shaders\PostLighting.hlsli
//
//-------------------------------------------------------------------------------------------------

// Light count defines.
#ifndef DIRLIGHT_COUNT
#define DIRLIGHT_COUNT 0
#endif

#ifndef POINTLIGHT_COUNT
#define POINTLIGHT_COUNT 0
#endif

#ifndef SPOTLIGHT_COUNT
#define SPOTLIGHT_COUNT 0
#endif

//
// Directional lighting.
//

// If directional light count is greater than zero, create an array of directional light data.
#if DIRLIGHT_COUNT

// Directional light data structure.
struct DirectionalLight
{
	float4x4			ViewProj;
	float4				Color;
	float4				Direction;
	float				Strength;
	int					ShadowsOn;
	float				ShadowStrength;
	float				Padding0;
};

cbuffer DirLightBuffer : register(b11)
{
	DirectionalLight	DIR_LIGHT[DIRLIGHT_COUNT];
};

Texture2DArray dirShadowMapTextures		: register(t30);

// Direction light shadow mapping function.
float DirectionalShadow(float3 worldPoint, DirectionalLight dirLight, int index)
{
	float4 projPos = mul(float4(worldPoint, 1.0f), dirLight.ViewProj);

	if(projPos.z < 0.0f)
		return 1.0f;

	projPos.xyz /= projPos.w;
	
	float bias = 0.0035f;
	float2 projectTexCoord;

	projectTexCoord.x = projPos.x * 0.5f + 0.5f;
	projectTexCoord.y = -projPos.y * 0.5f + 0.5f;

	// Check for projection.
	if(projPos.z > 0.0f && saturate(projectTexCoord.x) == projectTexCoord.x && saturate(projectTexCoord.y) == projectTexCoord.y)
	{
		float depth = (dirShadowMapTextures.Sample(samplerAnisoClamp, float3(projectTexCoord.xy, index))).r;
		
		float lightDepthValue = projPos.z - bias;
		if(lightDepthValue > depth)
			return 0.0f;
		else
			return 1.0f;
	}

	return 1.0f;
}

// Function for applying a directional light to the deferred buffer.
void DirectionalLighting(float3 diffuseColor, float4 specularColor, float3 normal, float3 cameraPoint, float3 worldPoint, 
	DirectionalLight dirLight, int index, inout float4 LightingColor)
{
	float shadowIntensity = 1.0f;

	if(dirLight.ShadowsOn)
	{
		shadowIntensity = DirectionalShadow(worldPoint, dirLight, index);
		shadowIntensity = saturate(shadowIntensity + dirLight.ShadowStrength);
	}

	dirLight.Direction.xyz = -dirLight.Direction.xyz;
	float diffuse = saturate(dot(normal, dirLight.Direction.xyz)) * shadowIntensity;
	
	// If the diffuse values is greater than zero, calculate apply the lighting values.
	if(diffuse > 0.0f)
	{
		float3 reflection = normalize(2 * normal * diffuse + dirLight.Direction.xyz);
		float specular = pow(saturate(dot(reflection, normalize(cameraPoint - worldPoint))), specularColor.a * 128) * shadowIntensity;

		dirLight.Color.rgb *= dirLight.Strength;
	
		LightingColor.rgb += (diffuseColor.rgb * diffuse * dirLight.Color.rgb + 
			specularColor.rgb * specular * dirLight.Color.rgb);
	}
}

#endif

//
// Point lighting.
//

// If point light count is greater than zero, create an array of point light data.
#if POINTLIGHT_COUNT

// Point light data structure.
struct PointLight
{
	float4x4			ViewProj[6];
	float4				Color;
	float4				Position;
	float4				Attenuation;
	float				Range;
	float				Strength;
	int					ShadowsOn;
	float				ShadowStrength;
};

cbuffer PointLightBuffer : register(b12)
{
	PointLight			POINT_LIGHT[POINTLIGHT_COUNT];
};

Texture2DArray pointShadowMapTextures	: register(t31);

// Point light shadow mapping function.
float PointShadow(float3 worldPoint, PointLight pointLight, int viewProjIndex, int index)
{
	float4 projPos = mul(float4(worldPoint, 1.0f), pointLight.ViewProj[viewProjIndex]);

	if(projPos.z < 0.0f)
		return 0.0f;

	projPos.xyz /= projPos.w;
	
	float bias = 0.0005f;
	float2 projectTexCoord;

	projectTexCoord.x = projPos.x * 0.5f + 0.5f;
	projectTexCoord.y = -projPos.y * 0.5f + 0.5f;

	// Check for projection.
	if(projPos.z > 0.0f && saturate(projectTexCoord.x) == projectTexCoord.x && saturate(projectTexCoord.y) == projectTexCoord.y)
	{
		float depth = (pointShadowMapTextures.Sample(samplerAnisoClamp, float3(projectTexCoord.xy, index))).r;
		float lightDepthValue = projPos.z - bias;
		if(lightDepthValue > depth)
			return 0.0f;
		else
			return 1.0f;
	}

	return 0.0f;
}

// Function for applying a point light to the deferred buffer.
void PointLighting(float3 diffuseColor, float4 specularColor, float3 normal, float3 cameraPoint, float3 worldPoint, 
	PointLight pointLight, int index, inout float4 LightingColor)
{
	float shadowIntensity = 1.0f;

	if(pointLight.ShadowsOn)
	{
		// Calculate this point light's shadow intensity.
		shadowIntensity = 0.0f;

		shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 0, index * 6));
		if(shadowIntensity < 0.99f) shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 1, index * 6 + 1));
		if(shadowIntensity < 0.99f) shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 2, index * 6 + 2));
		if(shadowIntensity < 0.99f) shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 3, index * 6 + 3));
		if(shadowIntensity < 0.99f) shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 4, index * 6 + 4));
		if(shadowIntensity < 0.99f) shadowIntensity = max(shadowIntensity, PointShadow(worldPoint, pointLight, 5, index * 6 + 5));
		
		shadowIntensity = saturate(shadowIntensity + pointLight.ShadowStrength);
	}

	float3 difference = (pointLight.Position.xyz - worldPoint.xyz);
	float3 direction = normalize(difference);
	
	float diffuse = saturate(dot(normal, direction)) * shadowIntensity;

	// If the diffuse values is greater than zero, calculate apply the lighting values.
	if(diffuse > 0.0f)
	{
		float distance = saturate(length(difference / pointLight.Range));
		float intensity = (1.0f - distance) * (1.0f / (pointLight.Attenuation.x + pointLight.Attenuation.y * distance + pointLight.Attenuation.z * distance * distance));
		diffuse *= intensity;
		
		float3 h = normalize(normalize(cameraPoint - worldPoint) + direction);
		float specular = pow(saturate(dot(h, normal.xyz)), specularColor.a * 128) * intensity * shadowIntensity;

		pointLight.Color.rgb *= pointLight.Strength;
	
		LightingColor.rgb += diffuseColor.rgb * diffuse * pointLight.Color.rgb + 
			specularColor.rgb * specular * pointLight.Color.rgb;
	}
}

#endif

//
// Spot lighting.
//

// If spotlight count is greater than zero, create an array of spotlight data.
#if SPOTLIGHT_COUNT

// Spotlight data structure.
struct Spotlight
{
	float4x4			ViewProj;
	float4				Color;
	float4				Direction;
	float4				Position;
	float4				Attenuation;
	float				Range;
	float				Strength;
	float				Falloff;
	float				Phi;
	float				Theta;
	int					ShadowsOn;
	float				ShadowStrength;
	float				Padding0;
};

cbuffer SpotlightBuffer : register(b13)
{
	Spotlight			SPOTLIGHT[SPOTLIGHT_COUNT];
};

Texture2DArray spotShadowMapTextures	: register(t32);

// Spotlight shadow mapping function.
float SpotShadow(float3 worldPoint, Spotlight spotlight, int index)
{
	float4 projPos = mul(float4(worldPoint, 1.0f), spotlight.ViewProj);
	if(projPos.z < 0.0f)
		return 0.0f;

	projPos.xyz /= projPos.w;
	
	float bias = 0.0005f;
	float2 projectTexCoord;

	projectTexCoord.x = projPos.x * 0.5f + 0.5f;
	projectTexCoord.y = -projPos.y * 0.5f + 0.5f;

	// Check for projection.
	if(projPos.z > 0.0f && saturate(projectTexCoord.x) == projectTexCoord.x && saturate(projectTexCoord.y) == projectTexCoord.y)
	{
		float depth = (spotShadowMapTextures.Sample(samplerAnisoClamp, float3(projectTexCoord.xy, index))).r;
		
		float lightDepthValue = projPos.z - bias;
		if(lightDepthValue > depth)
			return 0.0f;
		else
			return 1.0f;
	}

	return 1.0f;
}

// Function for applying a spotlight to the deferred buffer.
void SpotLighting(float3 diffuseColor, float4 specularColor, float3 normal, float3 cameraPoint, float3 worldPoint, 
	Spotlight spotlight, int index, inout float4 LightingColor)
{
	float shadowIntensity = 1.0f;

	if(spotlight.ShadowsOn)
	{
		shadowIntensity = SpotShadow(worldPoint, spotlight, index);
		shadowIntensity = saturate(shadowIntensity + spotlight.ShadowStrength);
	}

	float3 difference = (spotlight.Position.xyz - worldPoint.xyz);
	float3 direction = normalize(difference);
	float diffuse = saturate(dot(normal, direction)) * shadowIntensity;
	
	// If the diffuse values is greater than zero, calculate apply the lighting values.
	if(diffuse > 0.0f)
	{
		float distance = saturate(length(difference / spotlight.Range));
		float intensity = (1.0f - distance) * (1.0f / (spotlight.Attenuation.x + spotlight.Attenuation.y * distance + spotlight.Attenuation.z * distance * distance));
		diffuse *= intensity;

		float3 h = normalize(normalize(cameraPoint - worldPoint) + direction);
		float specular = pow(saturate(dot(h, normal.xyz)), specularColor.a * 128) * intensity * shadowIntensity;

		// Calculate the falloff the the spotline cone angles.
		float rho = saturate(dot(-spotlight.Direction.xyz, direction));
		float den = spotlight.Theta - spotlight.Phi;
		float num = saturate(rho - spotlight.Phi);
		float spotlightFactor = saturate(pow(num / den, spotlight.Falloff));

		spotlight.Color.rgb *= spotlight.Strength * spotlightFactor;

		LightingColor.rgb += diffuseColor.rgb * diffuse * spotlight.Color.rgb + 
			specularColor.rgb * specular * spotlight.Color.rgb;
	}
}

#endif