//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Utilities\CSystemInfo.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CSystemInfo.h"

namespace Util
{
	CSysInfo::CSysInfo()
	{
		m_totalVMemory						= 0;
		m_totalVMemoryUsed					= 0;
		m_totalPMemory						= 0;
		m_totalPMemoryUsed					= 0;

		m_processVMemoryUsage				= 0;
		m_processPMemoryUsage				= 0;

		m_totalCPUUsage						= 0;
		m_processCPUUsage					= 0;

		m_lastCpuSampleTime					= 0;
		m_processorCount					= 0;

		m_currentProcess					= GetCurrentProcess();

		m_lastMemSampleTime					= GetTickCount();
		m_lastCpuSampleTime					= GetTickCount();

	}

	CSysInfo::~CSysInfo()
	{
		Release();
	}

	// Method for calculating global system memory information.
	int CSysInfo::CalculateMemoryInfo()
	{
		int retCode = Util::Err::NONE;

		MEMORYSTATUSEX memoryInfo;
		memoryInfo.dwLength = sizeof(MEMORYSTATUSEX);
		if(!GlobalMemoryStatusEx(&memoryInfo))
		{
			return Util::Err::ERR_SYSINFO_FAILED_TO_CALC_GLOBAL_MEM_STATUS;
		}

		m_totalVMemory			= (memoryInfo.ullTotalPageFile) / BYTES_TO_MBS;
		m_totalVMemoryUsed		= (memoryInfo.ullTotalPageFile - memoryInfo.ullAvailPageFile) / BYTES_TO_MBS;
		m_totalPMemory			= (memoryInfo.ullTotalPhys) / BYTES_TO_MBS;
		m_totalPMemoryUsed		= (memoryInfo.ullTotalPhys - memoryInfo.ullAvailPhys) / BYTES_TO_MBS;

		return retCode;
	}
	
	// Method for calculating the current process virtual memory usage.
	int CSysInfo::CalculateMemoryUsage(u32* vmemory, u32* pmemory)
	{
		int retCode = Util::Err::NONE;
		
		if((m_lastMemSampleTime + 1000) < GetTickCount())
		{
			CalculateMemoryInfo();

			m_lastMemSampleTime = GetTickCount();

			if(!vmemory && !pmemory)
				return retCode;

			PROCESS_MEMORY_COUNTERS_EX pmc;
			pmc.cb = sizeof(pmc);
		
			if(!GetProcessMemoryInfo(m_currentProcess, (PROCESS_MEMORY_COUNTERS*)&pmc, pmc.cb))
			{
				return Util::Err::ERR_SYSINFO_FAILED_TO_CALC_PROCESS_MEM_USAGE;
			}
		
			if(vmemory) *vmemory = pmc.PrivateUsage / BYTES_TO_MBS;
			if(pmemory) *pmemory = pmc.WorkingSetSize / BYTES_TO_MBS;
		}

		return retCode;
	}

	// Method for calculating the total cpu.
	int CSysInfo::InitCPUCalculations()
	{
		int retCode = Util::Err::NONE;
		
		// Initialize cpu calculations for total cpu usage.
		if(PdhOpenQuery(NULL, 0, &m_cpuQuery) != ERROR_SUCCESS)
		{
			return Util::Err::ERR_SYSINFO_FAILED_TO_OPEN_CPU_QUERY;
		}

		if(PdhAddCounter(m_cpuQuery, TEXT("\\Processor(_Total)\\% processor time"), 0, &m_cpuCounter) != ERROR_SUCCESS)
		{
			return Util::Err::ERR_SYSINFO_FAILED_TO_ADD_CPU_COUNTER;
		}

		PDH_STATUS status;
		if((status = PdhCollectQueryData(m_cpuQuery)) != ERROR_SUCCESS)
		{
			return Util::Err::ERR_SYSINFO_FAILED_TO_COLLECT_CPU_DATA;
		}

		// Initialize cpu calculations for process cpu usage.
		SYSTEM_INFO sysInfo;
		FILETIME fTime, fSys, fUser;
		
		GetSystemInfo(&sysInfo);
		m_processorCount = sysInfo.dwNumberOfProcessors;

		GetSystemTimeAsFileTime(&fTime);
		memcpy(&m_lastCPU, &fTime, sizeof(FILETIME));

		if(!GetProcessTimes(m_currentProcess, &fTime, &fTime, &fSys, &fUser))
		{
			return Util::Err::ERR_SYSINFO_FAILED_GET_PROCESS_TIMES;
		}
		
		memcpy(&m_lastSysCPU, &fSys, sizeof(FILETIME));
		memcpy(&m_lastUserCPU, &fUser, sizeof(FILETIME));

		return retCode;
	}

	// Method for calculating the current CPU usage.
	int CSysInfo::CalculateCPUUsage()
	{
		int retCode = Util::Err::NONE;

		if((m_lastCpuSampleTime + 1000) < GetTickCount())
		{
			PDH_FMT_COUNTERVALUE counterValue;

			PdhCollectQueryData(m_cpuQuery);
			PdhGetFormattedCounterValue(m_cpuCounter, PDH_FMT_LONG, NULL, &counterValue);

			m_totalCPUUsage = (float)counterValue.longValue;

			CalculateProcessCPUUsage();

			m_lastCpuSampleTime = GetTickCount();
		}

		return retCode;
	}
	
	// Method for calculating the current process CPU usage.
	int CSysInfo::CalculateProcessCPUUsage()
	{
		int retCode = Util::Err::NONE;

		FILETIME fTime, fSys, fUser;
		ULARGE_INTEGER now, sys, user;

		GetSystemTimeAsFileTime(&fTime);
		memcpy(&now, &fTime, sizeof(FILETIME));

		if(!GetProcessTimes(m_currentProcess, &fTime, &fTime, &fSys, &fUser))
		{
			return Util::Err::ERR_SYSINFO_FAILED_GET_PROCESS_TIMES;
		}

		memcpy(&sys, &fSys, sizeof(FILETIME));
		memcpy(&user, &fUser, sizeof(FILETIME));

		m_processCPUUsage = (float)((sys.QuadPart - m_lastSysCPU.QuadPart) +
			(user.QuadPart - m_lastUserCPU.QuadPart));

		m_processCPUUsage /= now.QuadPart - m_lastCPU.QuadPart;
		m_processCPUUsage /= m_processorCount;

		m_lastCPU			= now;
		m_lastSysCPU		= sys;
		m_lastUserCPU		= user;

		m_processCPUUsage	*= 100.0f;

		return retCode;
	}

	void CSysInfo::Release()
	{
		PdhCloseQuery(m_cpuQuery);
	}
};