//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Project II
//
// File: Resources\Shaders\FireEffect.hlsl
//
//-------------------------------------------------------------------------------------------------

#include "RenderUtil.hlsli"

// Matrix constant buffer.
cbuffer MatrixBuffer : register(b0)
{
	float4x4	World;
	float4x4	WorldInv;
	float4x4	View;
	float4x4	ViewInv;
	float4x4	Projection;
	float4x4	ProjectionInv;
};

// Data constant buffer.
cbuffer ConstantBuffer : register(b1)
{
	float4		Color;
	float4		CameraPosition;
	float4		Speeds;
	float4		Scales;
	float		Softness;
	float		Time;
	float		ScreenWidth;
	float		ScreenHeight;
	int			BillboardType; // 0x0 = None, 0x1 = X, 0x2 = Y, 0x4 = Z, 0x5 = Flat, 0x7 = All
	float3		padding0;
};

// Distortion constant buffer.
cbuffer DistortionBuffer : register(b2)
{
	float2		Distortion0;
	float2		Distortion1;
	float2		Distortion2;
	float		DistortionScale;
	float		DistortionBias;
};

// Vertex input structure.
struct VIn
{
	float4	Position		: POSITION;
	float4	InstancePos		: INSTPOS;
	float4	InstanceScale	: INSTSCALE;
	float4	InstanceUV		: INSTUV;
	float4	InstanceColor	: INSTCOLOR;
	float4	InstanceData0	: INSTDATA0;
};

// Vertex output structure.
struct VOut
{
	float4	Position		: SV_POSITION;

	float4	Color			: COLOR;
	float4	Data0			: INSTDATA0;

	float4	Scale			: POSITION0;
	float4	WorldPos		: POSITION1;

	float4	Depth			: TEXCOORD0;

	float2	TexCoord		: TEXCOORD1;
	float4	TexCoord0		: TEXCOORD2;
	float2	TexCoord1		: TEXCOORD3;
	float2	TexCoord2		: TEXCOORD4;
};

// Pixel output structure.
struct POut
{
	float4	Color			: SV_TARGET0;
	float4	Diffuse			: SV_TARGET1;
};

bool ParticleAlive(float4 data)
{
	return (Time - data.x < data.y || data.y <= 1e-7f);
}

// Vertex shader function.
VOut VShader(VIn input)
{
	VOut output = (VOut)0;
	
	output.Scale				=	input.InstanceScale;
	output.WorldPos				=	float4(World._14, World._24, World._34, 1.0f) + float4(input.InstancePos.xyz, 0.0f);
	
	output.Color				=	Color * input.InstanceColor;
	output.TexCoord0			=	input.InstanceUV;
	output.Data0				=	input.InstanceData0;
	
	return output;
}

// Geometry shader function.
[maxvertexcount(4)]
void GShader(point VOut input[1], inout TriangleStream<VOut> outputStream)
{
	float halfSize = 0.5f;

	float3 planeNormal = (input[0].WorldPos.xyz) - CameraPosition.xyz;

	float3 upVec = float3(0, 1, 0);
	float3 rightVec = float3(1, 0, 0);
	
	if(!(BillboardType & 0x1)) // Not X.
	{
		planeNormal.x = 0;
	}
	if(!(BillboardType & 0x2)) // Not Y.
	{
		planeNormal.y = 0;
	}
	if(!(BillboardType & 0x4)) // Not Z.
	{
		planeNormal.z = 0;
	}

	planeNormal = normalize(planeNormal);
	rightVec = normalize(cross(planeNormal, upVec));
	upVec = normalize(cross(rightVec, planeNormal));

	upVec *= input[0].Scale.xyz * halfSize;
	rightVec *= input[0].Scale.xyz * halfSize;

	float3 vert[4];

	vert[0] = input[0].WorldPos.xyz - rightVec - upVec;
	vert[1] = input[0].WorldPos.xyz + rightVec - upVec;
	vert[2] = input[0].WorldPos.xyz - rightVec + upVec;
	vert[3] = input[0].WorldPos.xyz + rightVec + upVec;

	float2 texCoord[4];

	texCoord[0] = float2(0, 1);
	texCoord[1] = float2(1, 1);
	texCoord[2] = float2(0, 0);
	texCoord[3] = float2(1, 0);

	// Create the particle quad in the geometry shader.
	VOut output = input[0];
	for(int i = 0; i < 4; i++)
	{
		output.Position				=	mul(float4(vert[i], 1.0f), World);
		output.Position				=	mul(output.Position, View);
		output.Depth				=	output.Position;
		output.Position				=	mul(output.Position, Projection);
		
		output.TexCoord2.xy			=	(texCoord[i] * input[0].TexCoord0.zw + input[0].TexCoord0.xy) * Scales.z;
		output.TexCoord1.xy			=	(texCoord[i] * input[0].TexCoord0.zw + input[0].TexCoord0.xy) * Scales.y;
		output.TexCoord0.xy			=	(texCoord[i] * input[0].TexCoord0.zw + input[0].TexCoord0.xy);
		output.TexCoord.xy			=	output.TexCoord0.xy;
		output.TexCoord0.xy			*=	Scales.x;
		
		output.TexCoord0.y			=	output.TexCoord0.y + (Time * Speeds.x);
		output.TexCoord1.y			=	output.TexCoord1.y + (Time * Speeds.y);
		output.TexCoord2.y			=	output.TexCoord2.y + (Time * Speeds.z);

		outputStream.Append(output);
	}

	outputStream.RestartStrip();
}

Texture2D		fireColorTex		: register(t0);
Texture2D		fireNoiseTex		: register(t1);
Texture2D		fireAlphaTex		: register(t2);
Texture2D		DepthStencilTex		: register(t121);

// Pixel shader function.
POut PShader(VOut input) : SV_TARGET
{
	if(!ParticleAlive(input.Data0) || input.Color.a < 0.01f)
		discard;

	float2 texCoord = 0.5f * float2(input.Position.x / ScreenWidth, -input.Position.y / ScreenHeight) + 0.5f;
	float4 depthStencil = DepthStencilTex.Load(int3(input.Position.x, input.Position.y, 0));
	
	// Calculate soft particle fade.
	float z = depthStencil.r;
	float x = (texCoord.x) * 2 - 1;
	float y = (1.0f - texCoord.y) * 2 - 1;
	
	float4 projectPos = float4(x, y, z, 1.0f);
	float4 viewPosition = mul(projectPos, ProjectionInv);
	viewPosition.xyz /= viewPosition.w;
	
	input.Depth.xyz /= input.Depth.w;

	// Soften the particle.
	float fade = saturate((viewPosition.z - input.Depth.z) * Softness);
	
	POut output;
	
	float4 noise0;
	float4 noise1;
	float4 noise2;
	float4 finalNoise;
	float perturb;
	float2 noiseCoords;
	float4 fireColor;
	float4 alphaColor;

	// Add distortion to the flames.
	noise0 = fireNoiseTex.Sample(samplerAniso, input.TexCoord0.xy);
	noise1 = fireNoiseTex.Sample(samplerAniso, input.TexCoord1);
	noise2 = fireNoiseTex.Sample(samplerAniso, input.TexCoord2);

	noise0 = (noise0 - 0.5f) * 2.0f;
	noise1 = (noise1 - 0.5f) * 2.0f;
	noise2 = (noise2 - 0.5f) * 2.0f;

	noise0.xy *= Distortion0.xy;
	noise1.xy *= Distortion1.xy;
	noise2.xy *= Distortion2.xy;

	finalNoise = noise0 + noise1 + noise2;

	perturb = ((1.0f - input.TexCoord.y) * DistortionScale) + DistortionBias;

	noiseCoords.xy = (finalNoise.xy * perturb) + input.TexCoord.xy;

	// Extract color and alpha values for the fire, with the noise coordinates.
	fireColor = fireColorTex.Sample(samplerAnisoClamp, noiseCoords.xy);
	alphaColor = fireAlphaTex.Sample(samplerAnisoClamp, noiseCoords.xy);

	fireColor.a = alphaColor.r;
	
	output.Diffuse = float4(0, 0, 0, 0);
	output.Color = fireColor;
	output.Color.a *= fade;
	
	return output;
}